//==============================================================================
//  Verilog model for AdestoTech Devices
//  AT45DQXXX  parameterizable
//  Copyright 2013 Adesto Technologies Corporation, Inc.
//
//  Jun, 2013
//
//  For model support, please contact verilog_models@adestotech.com
//
//  Revision History:
//
//  0.1    : Copied from Atmel AT45DBxxxD, revision 2.1.
//  0.3    : Started changes for AT45DQ series.
//  1.0    : AT45DQ series Alpha Release.
//  1.1    : Changes after internal review:
//              a) Sectors 0a and 0b treated separately for resume/suspend.
//              b) Block Erase - all eight pages simultaneously.
//              c) Sector Erase - one block at a time.
//              d) Reset suspend status bit at the start of resume.
//              e) 45DQ161 timing numbers.
//  1.2    : Change time scale for logging to milli-seconds.
//  1.3    : Behavior and timing changes to match structural netlist (bug fixes):
//              a) Change timing for byte programming to (bytes*tBP) instead of (tP).
//              b) No Software Reset during OTP (Security Register) programming.
//              c) No Suspend/Resume of Auto Page Rewrite.
//              d) No Read from buffer while using buffer in background operation.
//              e) Add busy time for aborted operations - tABORT.
//              f) Write to buffer for aborted program through buffer operations.
//              g) Add simulation option to disable Quad-bit operation (no QEN).
//              h) Softreset only if device is busy or suspended.
//  1.4    : More bug-fixes:
//              a) Inadvertant reset when QE_reg not properly tested in quad-bit read.
//              b) Changed the RESETB related timing checks to eliminate false errors.
//              c) Changed timing of PRR, LRR, SRR to skip one clock cycle.
//              d) Corrected number of dummy bytes for legacy opcodes 52h and 68h.
//              e) Set SLE status bit on sector lockdown freeze.
//  1.5    : AT45DQ series Beta Release.
//  1.6    : Post-Beta changes:
//              a) Changed return type for check_lockdown and check_protect to integer.
//              b) Removed unused function check_busy.
//              c) Corrected timing numbers tV, tDIS, tWH and tWL.
//              d) Added error message #1008 for starting an operation while the same
//                 opcode is already running in the background.
//              e) Changed duplicate error #1215 to #1218.
//  1.7    : AT45DQ series Updated Beta Release.
//  1.8    : Post-Updated Beta changes:
//              a) Corrected disable MMPR_ to disable SRR_ in always @(SRR).
//              b) False report of "Incomplete byte in serial output" #2150 when CSB
//                 remains asserted after tV past the falling edge of SCK.
//              c) Corrected reading of MEMORY_FILE when BINARY_OPT == 1.
//              d) Added missing address reset on page changes for task cont_read_out.
//  1.9    : AT45DQ series Beta Release 2nd Update.
//  2.0    : Post-Beta (2nd Update) fixes and support for AT45DBxxxE series:
//              a) Added errors #1022, #1102, #1202, #1212, #1342, #1347, #1357, #1362,
//                 #2027, #2047 for suspend/resume not available in AT45DB021E and
//                 dual/quad bit operations not available in AT45DBxxxE series.
//              b) Added tWPE and tWPD timing for asserting WPB protect.
//              c) No tABORT delay for blocking program/erase when WPB is asserted.
//              d) Moved timing data to an include file, selected by a define.
//              e) Turn on output drivers at tV instead of tHO.
//              f) Removed disabling of event handlers on posedge CSB, as it prevented
//                 the event handlers from posting error messages on premature ending.
//              g) Added parameter WARN_SOUT to enable "Incomplete byte in serial output"
//                 error messages #2150, #2160, #2180, #2205 and #2235.  Default is to
//                 not print these messages (WARN_SOUT=0), as serial output may be
//                 interrupted by posedge CSB at any time, with no ill effect.
//              h) Added support for Read-Modify-Write (opcodes 58h and 59h) for some
//                 inventory of AT45DB and AT45DQ series (Contact sales for availability).
//              i) Changed read_out* tasks to wrap-around addresses before first output.
//  2.1    : AT45DQ and AT45DB series Beta Release 3nd Update.
//  2.2    : Post-Beta (3rd Update) fixes:
//              a) Changed behavior for out-of-range addresses to match structural
//                 netlist behavior.  For DataFlash page size, where the address space
//                 has gaps between the end of one page and the start of the next, the
//                 output for read operations will be XX when an invalid address is
//                 supplied (controlled by BADADDR_OPT switch).  The same BADADDR_OPT
//                 switch will also control the corruption of the relevant buffer and
//                 memory page for write operations with invalid byte addresses.  Note
//                 that an invalid address is an error and the result of issuing an
//                 invalid address is not specified or guaranteed.
//              b) The error messages for out-of-range addresses now report the actual
//                 address, from the customer view.  (Previously the reported address
//                 was for the internal model representation of the flash memory.)  The
//                 debug messages, if enabled, still report addresses with respect to
//                 the model representation of the flash memory.
//              c) Change flash write operations from regular assignments to bit-AND,
//                 to more accurately reflect the result of writing to memory that was
//                 not erased prior.  Errors will still be posted when writing to
//                 unerased memory and the EPE status bit will be asserted.  Any operation
//                 that programs or erases non-volatile memory will set or reset the EPE
//                 bit, including security register program, lockdown register program or
//                 freeze, protection register erase/program, Binary/DataFlash page size
//                 configuration, Quad-bit mode configuration and general program and
//                 erase operations.  Aborted operations will not modify the EPE bit.
//              d) Change parameter name AT45DQ_RMW_OVERRIDE to RMW_OVERRIDE.
//              e) Reset all volatile registers for Ultra-Deep Power-Down (similar
//                 to a power cycle).  The compare bit in the status register and the
//                 software sector protection state are affected.
//              f) Added the WPB==1'bz case to allow opcode software sector protection
//                 disable (SPD, 3D_2A_7F_9A) and WPB posedge transition handling to
//                 proceed when WPB is floating.  This is for the lack of a pull-up
//                 (tri1) on WPB in this model.
//              g) Disabling Quad-bit mode (QDIS, opcode 3D_2A_81_67) while WPB is low
//                 will enable hardware sector protection.
//              h) Compare status bit is reset by some non-volatile program operations.
//              i) Exclude tri-stated SO from data-in setup time check (dual/quad bit).
//  2.3    : AT45DQ and AT45DB series first General Release.
//==============================================================================
//  Select the device to use by parameter DEVICE.
//==============================================================================

`timescale 1ns / 10ps

`ifndef AT45DQXXX_TIMING
`define AT45DQXXX_TIMING "AT45DB021E_1v65_timing.v"
`endif

module AT45DQXXX (CSB, SCK, SI, WPB, RESETB, SO);

parameter DEVICE = "AT45DB021E";

// --- Select only one ---
// parameter DEVICE = "AT45DB021E";	// 8789B-DFLASH-11/2012
// parameter DEVICE = "AT45DB041E";	// 8783B-DFLASH-11/2012
// parameter DEVICE = "AT45DB161E";	// 8782D-DFLASH-11/2012
// parameter DEVICE = "AT45DB321E";	// 8784B-DFLASH-11/2012
// parameter DEVICE = "AT45DQ161";	// 8790A-DFLASH-11/2012

//==============================================================================
//  Module Port Declarations
//==============================================================================

input	CSB, SCK;
inout	SI, WPB, RESETB, SO;

wire	CSB, SCK;
wire	SI, RESETB, SO;
//tri1	WPB;			// Pull-up not modeled, to match structural netlist
wire	WPB;

//==============================================================================
//  Setup initial device configuration
//==============================================================================

parameter BINARY_OPT = 0;	// 0: Dataflash page size; 1: Binary page size
parameter QUADEN_OPT = 1;	// 1: Initial Quad-bit read-write enabled
parameter QUADEN_DIS = 0;	// 1: Fuse option to disable QEN (Quad-bit mode)

// Some AT45DB and AT45DQ series inventory include support for Read-Modify-Write
// operations (available by request).  To allow Read-Modify-Write support for
// devices where some of the inventory supports it, set RMW_OVERRIDE to 1.
parameter RMW_OVERRIDE = 0;

//==============================================================================
//  Setup simulation parameters
//==============================================================================

parameter TIMESCALE  = 1.0e-6;	// TimeScale for logging output in ms
parameter WARN_SOUT  = 0;	// 1: Enable "Incomplete byte in serial output" message
parameter BADADDR_OPT = 1;	// 1: Output XX when invalid DataFlash address supplied
// The BADADDR_OPT controls the output of read operations when a byte address is specified
// outside the DataFlash in-page byte address range.  The default (BADADDR_OPT=1) is to
// output XX.  With BADADDR_OPT=0 the model will mirror the serial output of the structural
// netlist for invalid DataFlash byte addresses.  The BADADDR_OPT=0 option is intended for
// model verification and not for normal model use.  Note that for either BADADDR_OPT option
// an invalid address is an error condition and the model will output error messages.
// For write operations with an invalid DataFlash in-page byte address, the BADADDR_OPT=1
// option will result in the corruption of any buffers or memory pages written to.

parameter DEBUG_OPT0 = 0;	// 1: Log main events
parameter DEBUG_OPT1 = 0;	// 1: Log all memory reads/writes
parameter DEBUG_OPT2 = 0;	// 1: Log address and dummies
parameter DEBUG_OPT3 = 0;	// 1: Log minor events
parameter DEBUG_OPT4 = 0;	// 1: Log timing checks
parameter DEBUG_OPT5 = 0;	// 1: Log detail for code development
parameter DEBUG_OPT6 = 0;	// 1: Log data moves for BINARY_OPT==1 MEMORY_FILE read
parameter DEBUG_OPT7 = 0;	// 1: Log detail of serial_in and serial_out tasks

//==============================================================================
//  Memory & Registers PreLoading Parameters:
/*==============================================================================

These parameters are related to Memory and Registers Preloading.
Memory, Sector Protection Register, Sector Lock-down Register and
Security Register can be preloaded, in Hex format.

To pre-load Memory (in Hex format), define parameter
MEMORY_FILE = <filename>, where <filename> is the name of
the pre-load file.
If MEMORY_FILE = "", the Memory is initialized to Erased state (all data = FF).
If the memory is initialized, the status of all pages will be Not-Erased.

In this model the flash memory is represented as a single continuous array,
with each DataFlash sized page directly following the previous DataFlash sized
page.  Specifying addresses in the MEMORY_FILE when using the DataFlash page
size option will result in an incorrect read of the MEMORY_FILE due to the
address mapping needed when addressing a memory location during normal device
operation.  (See task parse_address for details on this mapping.)

When using the Binary page size option (parameter BINARY_OPT=1) the
MEMORY_FILE is read into the same single continous array memory representation.
The model will then move the MEMORY_FILE data to the appropriate addresses for
Binary mode operation (See INITIALIZE_ block).  For BINARY_OPT=1 addresses
may be used in the MEMORY_FILE.

To pre-load Sector Protection Register (in Hex format), define
parameter PROTECTION = <filename>, where <filename> is the name of
the pre-load file.
If PROTECTION = "", all sectors are initially in the Protected state
(all data = FF).

To pre-load Sector Lock-down Register (in Hex format), define parameter
LOCKDOWN = <filename>, where <filename> is the name of the
pre-load file.
If LOCKDOWN = "", all sectors are initially in the Unlocked state
(all data = 00).

To pre-load Security Register (only the User Programmable Bytes 0 to 63),
define parameter SECURITY = <filename>, where <filename> is the name
of the pre-load file.
If SECURITY = "", the register is initialized to erased state (all data = FF).

The Factory Programmed Bytes 64 to 127 are always initialized by defining
parameter FACTORY = "factory.txt". As the Factory Programmed Bytes are
accessible to the user for read, a sample of "factory.txt" file
needs to be included in the Verilog Model directory.

==============================================================================*/

parameter MEMORY_FILE = "";			// Memory pre-load
parameter PROTECTION = "";			// Sector Protection pre-load
parameter LOCKDOWN = "";			// Sector Lock-down pre-load
parameter SECURITY = "";			// Security Register Bytes[0:63]
parameter FACTORY = "AT45DQXXX_factory.txt";	// Security Register Bytes[64:127]

//==============================================================================
//  Status Register 1, Density Code
//  1M = 0011, 2M = 0101, 4M = 0111, 8M = 1001, 16M = 1011, 32M = 1101, 64M = 1111
//==============================================================================

localparam STATUS2    =	1;

localparam STATUS3    =	(DEVICE == "AT45DB011E") ? 1 :
			(DEVICE == "AT45DB021E") ? 0 :
			(DEVICE == "AT45DB041E") ? 1 :
			(DEVICE == "AT45DB081E") ? 0 :
			(DEVICE == "AT45DB161E") ? 1 :
			(DEVICE == "AT45DB321E") ? 0 :
			(DEVICE == "AT45DB642E") ? 1 :
			(DEVICE == "AT45DQ011")  ? 1 :
			(DEVICE == "AT45DQ021")  ? 0 :
			(DEVICE == "AT45DQ041")  ? 1 :
			(DEVICE == "AT45DQ081")  ? 0 :
			(DEVICE == "AT45DQ161")  ? 1 :
			(DEVICE == "AT45DQ321")  ? 0 :
			(DEVICE == "AT45DQ642")  ? 1 : 1;

localparam STATUS4    =	(DEVICE == "AT45DB011E") ? 0 :
			(DEVICE == "AT45DB021E") ? 1 :
			(DEVICE == "AT45DB041E") ? 1 :
			(DEVICE == "AT45DB081E") ? 0 :
			(DEVICE == "AT45DB161E") ? 0 :
			(DEVICE == "AT45DB321E") ? 1 :
			(DEVICE == "AT45DB642E") ? 1 :
			(DEVICE == "AT45DQ011")  ? 0 :
			(DEVICE == "AT45DQ021")  ? 1 :
			(DEVICE == "AT45DQ041")  ? 1 :
			(DEVICE == "AT45DQ081")  ? 0 :
			(DEVICE == "AT45DQ161")  ? 0 :
			(DEVICE == "AT45DQ321")  ? 1 :
			(DEVICE == "AT45DQ642")  ? 1 : 1;

localparam STATUS5    =	(DEVICE == "AT45DB011E") ? 0 :
			(DEVICE == "AT45DB021E") ? 0 :
			(DEVICE == "AT45DB041E") ? 0 :
			(DEVICE == "AT45DB081E") ? 1 :
			(DEVICE == "AT45DB161E") ? 1 :
			(DEVICE == "AT45DB321E") ? 1 :
			(DEVICE == "AT45DB642E") ? 1 :
			(DEVICE == "AT45DQ011")  ? 0 :
			(DEVICE == "AT45DQ021")  ? 0 :
			(DEVICE == "AT45DQ041")  ? 0 :
			(DEVICE == "AT45DQ081")  ? 1 :
			(DEVICE == "AT45DQ161")  ? 1 :
			(DEVICE == "AT45DQ321")  ? 1 :
			(DEVICE == "AT45DQ642")  ? 1 : 1;

//==============================================================================
//  Total number of memory pages in the device
//==============================================================================

localparam PAGES    =	(DEVICE == "AT45DB011E") ?  512 :
			(DEVICE == "AT45DB021E") ? 1024 :
			(DEVICE == "AT45DB041E") ? 2048 :
			(DEVICE == "AT45DB081E") ? 4096 :
			(DEVICE == "AT45DB161E") ? 4096 :
			(DEVICE == "AT45DB321E") ? 8192 :
			(DEVICE == "AT45DB642E") ? 8192 :
			(DEVICE == "AT45DQ011")  ?  512 :
			(DEVICE == "AT45DQ021")  ? 1024 :
			(DEVICE == "AT45DQ041")  ? 2048 :
			(DEVICE == "AT45DQ081")  ? 4096 :
			(DEVICE == "AT45DQ161")  ? 4096 :
			(DEVICE == "AT45DQ321")  ? 8192 :
			(DEVICE == "AT45DQ642")  ? 8192 : 32768 /*16384*/;

//==============================================================================
//  Number of memory pages per sector
//==============================================================================

localparam PAGES_PER_SECTOR =
			(DEVICE == "AT45DB011E") ? 128 :
			(DEVICE == "AT45DB021E") ? 128 :
			(DEVICE == "AT45DB041E") ? 256 :
			(DEVICE == "AT45DB081E") ? 256 :
			(DEVICE == "AT45DB161E") ? 256 :
			(DEVICE == "AT45DB321E") ? 128 :
			(DEVICE == "AT45DB642E") ? 256 :
			(DEVICE == "AT45DQ011")  ? 128 :
			(DEVICE == "AT45DQ021")  ? 128 :
			(DEVICE == "AT45DQ041")  ? 256 :
			(DEVICE == "AT45DQ081")  ? 256 :
			(DEVICE == "AT45DQ161")  ? 256 :
			(DEVICE == "AT45DQ321")  ? 128 :
			(DEVICE == "AT45DQ642")  ? 256 : 1024;

//==============================================================================
//  Total number of sectors of memory in the device (0a + 0b counts as one)
//==============================================================================

localparam SECTORS    =	(DEVICE == "AT45DB011E") ?  4 :
			(DEVICE == "AT45DB021E") ?  8 :
			(DEVICE == "AT45DB041E") ?  8 :
			(DEVICE == "AT45DB081E") ? 16 :
			(DEVICE == "AT45DB161E") ? 16 :
			(DEVICE == "AT45DB321E") ? 64 :
			(DEVICE == "AT45DB642E") ? 32 :
			(DEVICE == "AT45DQ011")  ?  4 :
			(DEVICE == "AT45DQ021")  ?  8 :
			(DEVICE == "AT45DQ041")  ?  8 :
			(DEVICE == "AT45DQ081")  ? 16 :
			(DEVICE == "AT45DQ161")  ? 16 :
			(DEVICE == "AT45DQ321")  ? 64 :
			(DEVICE == "AT45DQ642")  ? 32 : 16;

//==============================================================================
//  Number of bytes per DataFlash memory page
//==============================================================================

localparam DF_PAGESIZE =
			(DEVICE == "AT45DB011E") ?  264 :
			(DEVICE == "AT45DB021E") ?  264 :
			(DEVICE == "AT45DB041E") ?  264 :
			(DEVICE == "AT45DB081E") ?  264 :
			(DEVICE == "AT45DB161E") ?  528 :
			(DEVICE == "AT45DB321E") ?  528 :
			(DEVICE == "AT45DB642E") ? 1056 :
			(DEVICE == "AT45DQ011")  ?  264 :
			(DEVICE == "AT45DQ021")  ?  264 :
			(DEVICE == "AT45DQ041")  ?  264 :
			(DEVICE == "AT45DQ081")  ?  264 :
			(DEVICE == "AT45DQ161")  ?  528 :
			(DEVICE == "AT45DQ321")  ?  528 :
			(DEVICE == "AT45DQ642")  ? 1056 : 264;

//==============================================================================
//  Number of additional bytes per page for DataFlash page size beyond the
//  Binary page size
//==============================================================================

localparam DF_EXTRA_BYTES_PER_PAGE =
			(DEVICE == "AT45DB011E") ?  8 :
			(DEVICE == "AT45DB021E") ?  8 :
			(DEVICE == "AT45DB041E") ?  8 :
			(DEVICE == "AT45DB081E") ?  8 :
			(DEVICE == "AT45DB161E") ? 16 :
			(DEVICE == "AT45DB321E") ? 16 :
			(DEVICE == "AT45DB642E") ? 32 :
			(DEVICE == "AT45DQ011")  ?  8 :
			(DEVICE == "AT45DQ021")  ?  8 :
			(DEVICE == "AT45DQ041")  ?  8 :
			(DEVICE == "AT45DQ081")  ?  8 :
			(DEVICE == "AT45DQ161")  ? 16 :
			(DEVICE == "AT45DQ321")  ? 16 :
			(DEVICE == "AT45DQ642")  ? 32 : 8;

//==============================================================================
//  Total size of device storage, in bytes
//==============================================================================

localparam MEMSIZE	= DF_PAGESIZE * PAGES;

//==============================================================================
//  Page size for Binary page size option (BINARY_OPT==1)
//==============================================================================

localparam BIN_PAGESIZE	= DF_PAGESIZE - DF_EXTRA_BYTES_PER_PAGE;

//==============================================================================
//  Number of SRAM buffers, each capable of storing one page worth
//==============================================================================

localparam BUFFERS    =	(DEVICE == "AT45DB011E") ? 1 :
			(DEVICE == "AT45DB021E") ? 1 :
			(DEVICE == "AT45DB041E") ? 2 :
			(DEVICE == "AT45DB081E") ? 2 :
			(DEVICE == "AT45DB161E") ? 2 :
			(DEVICE == "AT45DB321E") ? 2 :
			(DEVICE == "AT45DB642E") ? 2 :
			(DEVICE == "AT45DQ011")  ? 1 :
			(DEVICE == "AT45DQ021")  ? 1 :
			(DEVICE == "AT45DQ041")  ? 2 :
			(DEVICE == "AT45DQ081")  ? 2 :
			(DEVICE == "AT45DQ161")  ? 2 :
			(DEVICE == "AT45DQ321")  ? 2 :
			(DEVICE == "AT45DQ642")  ? 2 : 2;

//==============================================================================
//  Number of address bits to access a byte within one DataFlash size page
//  For Binary page size it is one bit less.  The SRAM buffers are the same
//  size as one page and has the same number of address bits.
//  Note that the device can switch between DataFlash and Binary page sizes.
//==============================================================================

localparam DF_INPAGE_ADDR_BITS =
			(DEVICE == "AT45DB011E") ?  9 :
			(DEVICE == "AT45DB021E") ?  9 :
			(DEVICE == "AT45DB041E") ?  9 :
			(DEVICE == "AT45DB081E") ?  9 :
			(DEVICE == "AT45DB161E") ? 10 :
			(DEVICE == "AT45DB321E") ? 10 :
			(DEVICE == "AT45DB642E") ? 11 :
			(DEVICE == "AT45DQ011")  ?  9 :
			(DEVICE == "AT45DQ021")  ?  9 :
			(DEVICE == "AT45DQ041")  ?  9 :
			(DEVICE == "AT45DQ081")  ?  9 :
			(DEVICE == "AT45DQ161")  ? 10 :
			(DEVICE == "AT45DQ321")  ? 10 :
			(DEVICE == "AT45DQ642")  ? 11 : 9;

//==============================================================================
//  Manufacturer and Device ID
//==============================================================================

localparam MAN_ID     =	(DEVICE == "AT45DB011E") ? 40'h1F_22_00_01_00 :
			(DEVICE == "AT45DB021E") ? 40'h1F_23_00_01_00 :
			(DEVICE == "AT45DB041E") ? 40'h1F_24_00_01_00 :
			(DEVICE == "AT45DB081E") ? 40'h1F_25_00_01_00 :
			(DEVICE == "AT45DB161E") ? 40'h1F_26_00_01_00 :
			(DEVICE == "AT45DB321E") ? 40'h1F_27_00_01_00 :
			(DEVICE == "AT45DB642E") ? 40'h1F_28_00_01_00 :
			(DEVICE == "AT45DQ011")  ? 40'h1F_22_00_01_00 :
			(DEVICE == "AT45DQ021")  ? 40'h1F_23_00_01_00 :
			(DEVICE == "AT45DQ041")  ? 40'h1F_24_00_01_00 :
			(DEVICE == "AT45DQ081")  ? 40'h1F_25_00_01_00 :
			(DEVICE == "AT45DQ161")  ? 40'h1F_26_00_01_00 :
			(DEVICE == "AT45DQ321")  ? 40'h1F_27_00_01_00 :
			(DEVICE == "AT45DQ642")  ? 40'h1F_28_00_01_00 :
						   40'h1F_28_00_01_00 ;

//==============================================================================
//  Support Suspend/Resume of program or erase operations (1 = supported)
//==============================================================================

localparam SUPPORT_SUSP =
			(DEVICE == "AT45DB011E") ? 0 :
			(DEVICE == "AT45DB021E") ? 0 :
			(DEVICE == "AT45DB041E") ? 1 :
			(DEVICE == "AT45DB081E") ? 1 :
			(DEVICE == "AT45DB161E") ? 1 :
			(DEVICE == "AT45DB321E") ? 1 :
			(DEVICE == "AT45DB642E") ? 1 :
			(DEVICE == "AT45DQ011")  ? 0 :
			(DEVICE == "AT45DQ021")  ? 0 :
			(DEVICE == "AT45DQ041")  ? 1 :
			(DEVICE == "AT45DQ081")  ? 1 :
			(DEVICE == "AT45DQ161")  ? 1 :
			(DEVICE == "AT45DQ321")  ? 1 :
			(DEVICE == "AT45DQ642")  ? 1 : 1;

//==============================================================================
//  Support Dual/Quad bit read/write mode (1 = supported)
//==============================================================================

localparam SUPPORT_DQBIT =
			(DEVICE == "AT45DB011E") ? 0 :
			(DEVICE == "AT45DB021E") ? 0 :
			(DEVICE == "AT45DB041E") ? 0 :
			(DEVICE == "AT45DB081E") ? 0 :
			(DEVICE == "AT45DB161E") ? 0 :
			(DEVICE == "AT45DB321E") ? 0 :
			(DEVICE == "AT45DB642E") ? 0 :
			(DEVICE == "AT45DQ011")  ? 1 :
			(DEVICE == "AT45DQ021")  ? 1 :
			(DEVICE == "AT45DQ041")  ? 1 :
			(DEVICE == "AT45DQ081")  ? 1 :
			(DEVICE == "AT45DQ161")  ? 1 :
			(DEVICE == "AT45DQ321")  ? 1 :
			(DEVICE == "AT45DQ642")  ? 1 : 1;

//==============================================================================
//  Support Read-Modify-Write opcodes 58h and 59h (1 = supported)
//==============================================================================

localparam SUPPORT_RMW =
			(DEVICE == "AT45DB011E") ? RMW_OVERRIDE :
			(DEVICE == "AT45DB021E") ? 1 :
			(DEVICE == "AT45DB041E") ? 1 :
			(DEVICE == "AT45DB081E") ? RMW_OVERRIDE :
			(DEVICE == "AT45DB161E") ? RMW_OVERRIDE :
			(DEVICE == "AT45DB321E") ? RMW_OVERRIDE :
			(DEVICE == "AT45DB642E") ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ011")  ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ021")  ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ041")  ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ081")  ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ161")  ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ321")  ? RMW_OVERRIDE :
			(DEVICE == "AT45DQ642")  ? RMW_OVERRIDE : RMW_OVERRIDE;

//##############################################################################

`protect

//==============================================================================
//  Status Register Bits
//==============================================================================

localparam ST_RDY = 7;  // status1_reg[7] = RDY/nBUSY (device is ready)
localparam ST_CMP = 6;  // status1_reg[6] = COMP      (match compare result)
localparam ST_PRT = 1;  // status1_reg[1] = PROTECT   (software sector protection enabled)
localparam ST_BIN = 0;  // status1_reg[0] = PAGE SIZE (binary page size)

localparam ST_EPE = 5;  // status2_reg[5] = E|P ERROR (erase or program error detected)
localparam ST_SLE = 3;  // status2_reg[3] = LOCKDOWN  (sector lockdown command enabled)
localparam ST_PS2 = 2;  // status2_reg[2] = P SUSP B2 (program from buffer 2 suspended)
localparam ST_PS1 = 1;  // status2_reg[1] = P SUSP B1 (program from buffer 1 suspended)
localparam ST_ES  = 0;  // status2_reg[0] = E SUSPEND (erase suspended)

//==============================================================================
//  Other local constants
//==============================================================================

//--- Serial-in/out mode
localparam S_SNGL = 1;	// Single line serial in on SI or out on SO
localparam S_DUAL = 2;	// Dual bit serial in/out on {SO, SI}
localparam S_QUAD = 4;	// Quad bit serial in/out on {RESETB, WPB, SO, SI}

//--- Memory selector for read-out or other memory operations
localparam MEM_MM = 0;	// Main memory selector
localparam MEM_B1 = 1;	// Buffer 1 selector
localparam MEM_B2 = 2;	// Buffer 2 selector

//--- Register selector for read-out
localparam R_PRT =  1;	// Protection Register
localparam R_LCK =  2;	// Lockdown Register
localparam R_SEC =  3;	// Security Register

//--- Indicator for scope of in-progress erase operation
localparam ERS_P =  1;	// In-progress page erase
localparam ERS_B =  2;	// In-progress block erase
localparam ERS_S =  3;	// In-progress sector erase
localparam ERS_C =  4;	// In-progress chip erase

//==============================================================================
//  Memory and other registers (Module scope)
//==============================================================================

reg [7:0] memory [MEMSIZE-1:0];		// Main memory (non-volatile flash)

reg [7:0] buffer1 [DF_PAGESIZE-1:0];	// SRAM Buffer 1
reg [7:0] buffer2 [DF_PAGESIZE-1:0];	// SRAM Buffer 2

reg [PAGES-1:0] page_status;		// 0: Page is erased, otherwise not erased

reg [7:0] status1_reg;			// Status Register Byte 1
reg [7:0] status2_reg;			// Status Register Byte 2
reg [7:0] config_reg;			// Store QE bit

reg [7:0] factory_reg[63:0];		// factory programmed security register
reg [7:0] security_reg[63:0];		// Security (OTP) register
reg security_flag;			// 0 if security register has not been programmed yet

reg [7:0] prot_reg [SECTORS-1:0];	// Sector protection register
reg [7:0] lock_reg [SECTORS-1:0];	// Sector lock-down register
reg [SECTORS:0] prot_status;		// 0: Sector protected; 1: unprotected
reg [SECTORS:0] lock_status;		// 0: Sector unlocked; 1: locked-down

// prot_status and lock_status register width is SECTORS + 1, so sector 0a status is in
// reg[0], sector 0b status is in reg[1], and last sector status is in reg[SECTORS]

//==============================================================================
//  Flags to track the current operation of the device
//==============================================================================

reg deep_power_down;	// Device is in Deep Power-Down state
reg ultra_power_down;	// Device is in Ultra-Deep Power-Down state
reg updating_buffer1;	// SRAM buffer 1 is being written
reg updating_buffer2;	// SRAM buffer 2 is being written
reg inuse_buffer1;	// SRAM buffer 1 is used by a background operation
reg inuse_buffer2;	// SRAM buffer 2 is used by a background operation
reg updating_memory;	// Main memory is being written (at inprog_pgm_page)
reg updating_pagesize;	// Page size (DataFlash or Binary) is being updated
reg updating_cfg_reg;	// Config register is being updated
reg updating_sec_reg;	// Security register is being updated
reg updating_fac_reg;	// Factory register is being updated
reg updating_lock_reg;	// Sector lockdown is being updated
reg updating_prot_reg;	// Sector protection is being updated
reg updating_freeze;	// Sector lockdown freeze state is being updated
reg comparing;		// Comparing buffer to main memory page (compare status bit)
reg erasing_page;	// Main memory page is being erased (at inprog_ers_page)
reg erasing_block;	// Main memory block is being erased (at inprog_ers_page)
reg erasing_sector;	// Main memory sector is being erased (at inprog_ers_page)
reg erasing_chip;	// Entire main memory is being erased
reg suspending;		// A program or erase is being suspended
reg resuming;		// A program or erase is being resumed
reg reset_active;	// 1'b1 while RESETB is low and not Quad mode, else 1'b0
reg hw_protect_active;	// 1'b1 while WPB is low and not Quad mode, else 1'b0
reg hw_protect_pending;	// 1'b1 while a WPB event is setting up or disabling protection

//==============================================================================
//  Global variables related to program/erase suspend
//==============================================================================

reg	allow_pgm_suspend;	// This operation may be suspended if 1'b1
integer	inprog_pgm_buffer;	// Source buffer for program operation (MEM_B1 or MEM_B2)
integer	inprog_pgm_page;	// Memory page being programmed
time	inprog_pgm_start;	// Start time of program operation
time	inprog_pgm_delay;	// Time needed to complete operation
reg	inprog_pgm_epe;		// EPE status bit for this program operation

reg	allow_ers_suspend;	// This operation may be suspended if 1'b1
integer	inprog_ers_scope;	// ERS_P (page), ERS_B (block), ERS_S (sector), ERS_C (chip)
integer	inprog_ers_page;	// Memory page being erased
time	inprog_ers_start;	// Start time of erase operation
time	inprog_ers_delay;	// Time needed to complete operation
reg	inprog_ers_epe;		// EPE status bit for this erase operation

//==============================================================================
//  Events to trigger some task based on the opcode received
//==============================================================================

event  EDPD;		// Deep Power-down (enable)
event  RDPD;		// Resume from Deep Power-down
event  UDPD;		// Ultra-Deep Power-down (enable)
event  XUDPD;		// Exit from Ultra-Deep Power-down
event  MMPR;		// Main Memory Page Read
event  MMCAR;		// Continuous Array Read
event  DORA;		// Dual-Output Read Array
event  QORA;		// Quad-Output Read Array
event  MIR;		// Manufacturer ID Read
event  B1R;		// Buffer 1 Read
event  B2R;		// Buffer 2 Read
event  MMPTB1T;		// Main Memory Page To Buffer 1 Transfer
event  MMPTB2T;		// Main Memory Page To Buffer 2 Transfer
event  MMPTB1C;		// Main Memory Page To Buffer 1 Compare
event  MMPTB2C;		// Main Memory Page To Buffer 2 Compare
event  B1W;		// Buffer 1 Write
event  B2W;		// Buffer 2 Write
event  DB1W;		// Dual-input Buffer 1 Write
event  DB2W;		// Dual-input Buffer 2 Write
event  QB1W;		// Quad-input Buffer 1 Write
event  QB2W;		// Quad-input Buffer 2 Write
event  B1TMMPPWBIE;	// Buffer 1 To Main Memory Page Prog With Built-In Erase
event  B2TMMPPWBIE;	// Buffer 2 To Main Memory Page Prog With Built-In Erase
event  B1TMMPPWOBIE;	// Buffer 1 To Main Memory Page Prog WithOut Built-In Erase
event  B2TMMPPWOBIE;	// Buffer 2 To Main Memory Page Prog WithOut Built-In Erase
event  PE;		// Page Erase
event  BE;		// Block Erase
event  SE;		// Sector Erase
event  OPC7;		// 4-byte command starting with C7
event  CE;		// Chip Erase
event  MMPPB1;		// Main Memory Page Prog. Through Buffer 1 With Erase
event  MMPPB2;		// Main Memory Page Prog. Through Buffer 2 With Erase
event  MMPPB1NE;	// Main Memory Page Prog. Through Buffer 1 Without Erase
event  APRB1;		// Auto Page Rewrite (or Read-Modify-Write) Through Buffer 1
event  APRB2;		// Auto Page Rewrite (or Read-Modify-Write) Through Buffer 2
event  SR;		// Status Register Read
event  CSR;		// Configuration Register Read
event  OP3D;		// 4-byte command starting with 3D
event  BPS;		// Binary Page Size Program
event  DPS;		// DataFlash Page Size Program
event  QEN;		// Enable quad-bit input/output
event  QDIS;		// Disable quad-bit input/output
event  PRE;		// Sector Protection Register Erase
event  PRP;		// Sector Protection Register Program
event  LRP;		// Sector Lock-down Register Program
event  SPE;		// Sector Protection Enable
event  SPD;		// Sector Protection Disable
event  PRR;		// Sector Protection Register Read
event  LRR;		// Sector Lock-down Register Read
event  OP9B;		// 4-byte command starting with 9B
event  SRP;		// Security Register Program
event  SRR;		// Security Register Read
event  OP34;		// 4-byte command starting with 34
event  FREEZE;		// Freeze sector lockdown and disable LRP
event  SUSPEND;		// Program/Erase Suspend
event  RESUME;		// Program/Erase Resume
event  OPF0;		// 4-byte command starting with F0
event  SOFTRESET;	// Software Reset

//==============================================================================
//  Other events, for debugging
//==============================================================================

event  BYTEBOUND;	// Debug event to indicate byte boundary on input
event  DEBUG_SO;	// Debug marker for SPI (single bit) output
event  DEBUG_DO;	// Debug marker for dual-bit output
event  DEBUG_QO;	// Debug marker for quad-bit output

//==============================================================================
//  Other global variables and registers
//==============================================================================

reg [7:0] debug_reg;	// register to copy input byte, for debug.

reg [7:0] opcode_reg;   // Store current opcode
reg [7:0] bg_opcode;	// Opcode of in-progress background operation

reg binary_page;	// 0: DataFlash page size; 1: Binary page size
integer	page_size;	// Number of bytes per page (for DF or Bin size)
integer	memory_top;	// Last addressable byte in memory (for DF or Bin size)

reg foreground_op_enable;	// Device is ready to do read and buffer operations
reg background_op_enable;	// Device is ready to do flash program/erase

integer arr_rd_dummybyte;	// Dummy bytes to skip in current memory read
integer buf_rd_dummybyte;	// Dummy bytes to skip in current buffer read

// The 'fast_mode' flag indicates that the device will change to ready status
// after just the page program time for pages that are already erased.
reg fast_mode;

// The 'skip' flag indicates that an extra clock needs to be skipped.
// This skipping is needed only for Inactive Clock Low.
reg skip;

// 0: SPI Mode 0 (SCK low @ negedge CSB); 3: SPI Mode 3 (SCK high @ negedge CSB)
integer spi_mode;

wire QE_reg;		// Mirrors Quad Enable bit state
assign QE_reg = config_reg[7];

//==============================================================================
//  For convenient simulation waveform debugging, not used elsewhere in model
//==============================================================================

wire status_RDY_bit;		assign status_RDY_bit	    = status1_reg[ST_RDY];
wire status_COMP_bit;		assign status_COMP_bit	    = status1_reg[ST_CMP];
wire status_PROTECT_bit;	assign status_PROTECT_bit   = status1_reg[ST_PRT];
wire status_PAGESIZE_bit;	assign status_PAGESIZE_bit  = status1_reg[ST_BIN];
wire status_EPE_bit;		assign status_EPE_bit	    = status2_reg[ST_EPE];
wire status_LOCKDOWN_bit;	assign status_LOCKDOWN_bit  = status2_reg[ST_SLE];
wire status_PS2_bit;		assign status_PS2_bit	    = status2_reg[ST_PS2];
wire status_PS1_bit;		assign status_PS1_bit	    = status2_reg[ST_PS1];
wire status_ES_bit;		assign status_ES_bit	    = status2_reg[ST_ES];

wire behav_BUSY;		assign behav_BUSY	    = ~status_RDY_bit;

// Define RDYBSY_reg, used as a timing control in some earlier testbenches
wire RDYBSY_reg;		assign RDYBSY_reg	    = status_RDY_bit;

//==============================================================================
//  Variables and flags to determine and test clock period
//==============================================================================

reg clk_valid;          // To verify clock timing for array read
time t_clk_prev_edge;   // Time of the previous clock rising edge
time t_clk_period;      // Time of the previous clock rising edge

reg sck_CAR1_valid;     // 1'b1 if SCK freq is less than fCAR1
reg sck_CAR2_valid;     // 1'b1 if SCK freq is less than fCAR2
reg sck_CAR3_valid;     // 1'b1 if SCK freq is less than fCAR3
reg sck_SCK_valid;      // 1'b1 if SCK freq is less than fSCK
reg sck_MAX_valid;      // 1'b1 if SCK freq is less than fMAX

//==============================================================================
//  Registers and continuous assignments for serial output
//==============================================================================

reg SI_reg;             // Signal out reg for IO<0>
reg SO_reg;             // Signal out reg for IO<1>
reg WPB_reg;            // Signal out reg for IO<2>
reg RESETB_reg;         // Signal out reg for IO<3>

reg SI_on;              // Signal out enable for IO<0>
reg SO_on;              // Signal out enable for IO<1>
reg WPB_on;             // Signal out enable for IO<2>
reg RESETB_on;          // Signal out enable for IO<3>

bufif1( SI, SI_reg, SI_on );
bufif1( SO, SO_reg, SO_on );
bufif1( WPB, WPB_reg, WPB_on );
bufif1( RESETB, RESETB_reg, RESETB_on );

// Conditions for setup/hold checks on inputs
assign SI_datain = !(CSB || SI_on);
assign SO_datain = (SO !== 1'bz) && !(CSB || SO_on);
assign WPB_datain = (QE_reg === 1'b1) && (CSB === 1'b0) && (WPB_on === 1'b0);
assign RESETB_datain = (QE_reg === 1'b1) && (CSB === 1'b0) && (RESETB_on === 1'b0);
assign RESETB_active = (QE_reg === 1'b0) && (RESETB === 1'b0);

//##############################################################################
//  SPECIFY BLOCK (Timing Parameters)
//##############################################################################

specify  /* all timing checks */

`include `AT45DQXXX_TIMING

    // SCK related
    //$period( posedge SCK, tSCK );	// SCK period is checked between
    //$period( negedge SCK, tSCK );	// rise-to-rise, as well as fall-to-fall
    $width( posedge SCK, tWH );		// High PulseWidth
    $width( negedge SCK, tWL );		// Low PulseWidth

    // CSB related
    $width( posedge CSB, tCS );		// CSB Min. High
    $setup( CSB,posedge SCK, tCSS );	// CSB setup time
    $hold( posedge SCK, CSB, tCSH );	// CSB hold time

    // Data In related. Checked only when CSB is low
    $setup( SI, posedge SCK &&& SI_datain, tSU );	// SI setup time
    $hold( posedge SCK &&& SI_datain, SI, tH );		// SI hold time
    $setup( SO, posedge SCK &&& SO_datain, tSU );	// SO setup time
    $hold( posedge SCK &&& SO_datain, SO, tH );		// SO hold time
    $setup( WPB, posedge SCK &&& WPB_datain, tSU );	// WPB setup time
    $hold( posedge SCK &&& WPB_datain, WPB, tH );	// WPB hold time
    $setup( RESETB, posedge SCK &&& RESETB_datain, tSU );// RESETB setup time
    $hold( posedge SCK &&& RESETB_datain, RESETB, tH );	// RESETB hold time

    // RESETB related
    $width( edge [10,x0] RESETB &&& RESETB_active, tRST );	// RESETB Low Width

    // Minimum CSB low pulse to exit Ultra-Deep Power-Down
    $width( negedge CSB &&& ultra_power_down, tCSLU );	// CSB Low Width

endspecify

//##############################################################################
//  Initialize
//##############################################################################

initial
begin : INITIALIZE_
    integer i;
    integer j;
    reg [7:0] temp_reg1;	// Temporary register for protection/lockdown
    reg mem_initialized;	// Memory pre-loaded flag
    integer bin_pagestart;	// Start address for a page in Binary memory organization
    integer df_pagestart;	// Start address for a page in DataFlash memory organization

    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Device %s initialization", TIMESCALE*$time, DEVICE );

    t_clk_prev_edge = $time;

    // Memory Initialization - start with erased state.  Pre-initialization to
    // erased state is useful if a user wants to initialize just a few locations
    for (i=0; i<MEMSIZE; i=i+1)
    begin
	memory[i] = 8'hFF;
    end
    mem_initialized = 0;

    // Preload main memory if needed
    if (MEMORY_FILE != ""  &&  BINARY_OPT == 0)
    begin
	$readmemh( MEMORY_FILE, memory );
	mem_initialized = 1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Reading %s into the DataFlash page size Memory Array", TIMESCALE*$time, MEMORY_FILE );
    end
    else if (MEMORY_FILE != ""  &&  BINARY_OPT == 1)
    begin
	$readmemh( MEMORY_FILE, memory );
	mem_initialized = 1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Reading %s into the Binary page size Memory Array", TIMESCALE*$time, MEMORY_FILE );
	// From last page to page 0
	for (j=PAGES-1; j>=0; j=j-1)
	begin
	    bin_pagestart = BIN_PAGESIZE * j;
	    df_pagestart = DF_PAGESIZE * j;
	    if (DEBUG_OPT6) $display( "DBG: %m, page = %04h, bin_pagestart = %06h, df_pagestart = %06h", j, bin_pagestart, df_pagestart );
	    // Fill upper (non-Binary mode) remainder of the DataFlash page with 0xFF
	    for (i=DF_PAGESIZE-1; i>=BIN_PAGESIZE; i=i-1)
	    begin
	        memory[df_pagestart+i] = 8'hFF;
		if (DEBUG_OPT6) $display( "DBG: df memory[%06h] := ff", df_pagestart+i );
	    end
	    // Move contents of Binary page to DataFlash memory organization
	    for (i=BIN_PAGESIZE-1; i>=0; i=i-1)
	    begin
	        memory[df_pagestart+i] = memory[bin_pagestart+i];
		if (DEBUG_OPT6) $display( "DBG: df memory[%06h] := bin memory[%06h] = %02h", df_pagestart+i, bin_pagestart+i, memory[bin_pagestart+i] );
	    end
	end
    end

    // If memory was initialized set pages to 'Not Erased' else set to 'Erased' state
    if (mem_initialized == 1'b1)
	for (j=0; j<PAGES; j=j+1) page_status[j] = 1'b1;
    else
	for (j=0; j<PAGES; j=j+1) page_status[j] = 1'b0;

    // Initialize status registers
    status1_reg[7] = 1'b1;	// ST_RDY, Device is ready at power up
    status1_reg[6] = 1'b0;	// ST_CMP, Compare bit is 0 at power up
    status1_reg[5] = STATUS5;	// Density code
    status1_reg[4] = STATUS4;	// Density code
    status1_reg[3] = STATUS3;	// Density code
    status1_reg[2] = STATUS2;	// Density code
    status1_reg[1] = 1'b0;	// ST_PRT, Protection enable is 0 at power up
    status1_reg[0] = BINARY_OPT;// ST_BIN, Binary page size option

    // status2_reg[7] mirrors status1_reg[7]
    status2_reg[6] = 1'b0;	// Reserved
    status2_reg[5] = 1'b0;	// ST_EPE, No erase|program error at power up
    status2_reg[4] = 1'b0;	// Reserved
    status2_reg[3] = 1'b1;	// ST_SLE, Sector Lockdown command is enabled at power up
    status2_reg[2] = 1'b0;	// ST_PS2, No program operation suspended at power up
    status2_reg[1] = 1'b0;	// ST_PS1, No program operation suspended at power up
    status2_reg[0] = 1'b0;	// ST_ES,  No erase operation suspended at power up

    config_reg = 8'h08;		// Bit 3 set to match structural netlist
    if (QUADEN_DIS == 1  ||  SUPPORT_DQBIT == 0)
	config_reg[7] = 0;	// Set QE bit to 0
    else
	config_reg[7] = QUADEN_OPT;

    arr_rd_dummybyte = 0;
    buf_rd_dummybyte = 0;
    fast_mode = 1'b0;
    opcode_reg = 8'hXX;
    bg_opcode = 8'hXX;

    binary_page = BINARY_OPT;
    page_size = DF_PAGESIZE - BINARY_OPT*DF_EXTRA_BYTES_PER_PAGE;
    memory_top = DF_PAGESIZE*(PAGES - 1) + page_size - 1;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) binary_page = %1b, page_size = %5d, memory_top = %06h", TIMESCALE*$time, binary_page, page_size, memory_top );

    // Initialize security register, protection and lock-down registers

    for (j=0; j<SECTORS; j=j+1)
    begin
	prot_reg[j] = 8'hFF;
	lock_reg[j] = 8'h00;
    end

    for (j=0; j<64; j=j+1)
	security_reg[j] = 8'hFF;
    security_flag = 0;

    // Pre-load security register, protection and lock-down registers, if needed
    if (PROTECTION != "")
    begin
	$readmemh( PROTECTION, prot_reg );
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Reading %s into the Protection Register", TIMESCALE*$time, PROTECTION );
    end

    if (LOCKDOWN !="")
    begin
	$readmemh( LOCKDOWN, lock_reg );
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Reading %s into the Lockdown Register", TIMESCALE*$time, LOCKDOWN );
    end

    $readmemh( FACTORY, factory_reg );
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Reading %s into the Factory Register", TIMESCALE*$time, FACTORY );

    if (SECURITY != "")
    begin
	$readmemh( SECURITY, security_reg );
	security_flag = 1'b1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Reading %s into the Security Register", TIMESCALE*$time, SECURITY );
    end

    // Initialize prot_status
    temp_reg1[7:0] = prot_reg[0];
    prot_status[0] = &(temp_reg1[7:6]);
    prot_status[1] = &(temp_reg1[5:4]);
    for (j=1; j< SECTORS; j= j+1)
	prot_status[j+1] = &(prot_reg[j]);

    // Initialize lock_status
    temp_reg1[7:0] = lock_reg[0];
    lock_status[0] = &(temp_reg1[7:6]);
    lock_status[1] = &(temp_reg1[5:4]);
    for (j=1; j< SECTORS; j= j+1)
	lock_status[j+1] = &(lock_reg[j]);

    // There is no activity at this time, chip is in stand-by mode
    deep_power_down = 1'b0;
    ultra_power_down = 1'b0;
    updating_buffer1 = 1'b0;
    updating_buffer2 = 1'b0;
    inuse_buffer1 = 1'b0;
    inuse_buffer2 = 1'b0;
    updating_memory = 1'b0;
    updating_pagesize = 1'b0;
    updating_cfg_reg = 1'b0;
    updating_sec_reg = 1'b0;
    updating_fac_reg = 1'b0;
    updating_lock_reg = 1'b0;
    updating_prot_reg = 1'b0;
    updating_freeze = 1'b0;
    comparing = 1'b0;
    erasing_page = 1'b0;
    erasing_block = 1'b0;
    erasing_sector = 1'b0;
    erasing_chip = 1'b0;
    suspending = 1'b0;
    resuming = 1'b0;
    reset_active = 1'b0;
    hw_protect_active = 1'b0;
    hw_protect_pending = 1'b0;

    // All outputs are High-impedance
    SI_on = 1'b0;
    SO_on = 1'b0;
    WPB_on    = 1'b0;
    RESETB_on = 1'b0;

    status1_reg[ST_RDY] = 1'b1;	// Device is ready

    // Power-up Timing Restrictions
    foreground_op_enable = 1'b0;
    background_op_enable = 1'b0;
    #tVCSL;
    foreground_op_enable = 1'b1; // Enable foreground op_codes
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Foreground OpCodes enabled", TIMESCALE*$time );
    #(tPUW - tVCSL);
    background_op_enable = 1'b1; // Enable background op_codes
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Background (Write) OpCodes enabled", TIMESCALE*$time );
end //initial

//==============================================================================
//  CSB asserted, start of op-code
//==============================================================================

always @( negedge CSB )		// The device will now become active
begin : GET_OPCODE_
    reg valid;			// Local flag to test serial input

    // Ignore CSB when RESETB is asserted
    if (reset_active == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1005}, %m (@ %.3fms), CSB falling edge ignored, device still in reset", TIMESCALE*$time );
	disable GET_OPCODE_;
    end

    // Requires only a pulse on CSB to exit Ultra-Deep Power-Down state
    if (ultra_power_down == 1'b1)
    begin
	-> XUDPD;
	disable GET_OPCODE_;
    end

    if (SCK === 1'b0)
    begin
	skip = 1'b1;
	spi_mode = 0;
    end
    else
    begin
	skip = 1'b0;
	spi_mode = 3;
	// If the opcode is related to SPI Mode 0/3, no skipping is needed. So, skip
	// will be reset to "0".
	// If opcode is related to Inactive Clock Low/high, skipping might or might
	// not be needed, depending on the value of SCK at negedge of CSB. So, in
	// such situations, skip will retain its value.
    end

    // Set defaults for op-codes that need these values, override per op-code below
    fast_mode = 1'b0;
    arr_rd_dummybyte = 0;
    buf_rd_dummybyte = 0;
    clk_valid = 1'b1;

    serial_in( S_SNGL, opcode_reg, valid );
    if (!valid) disable GET_OPCODE_;

    if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) opcode_reg = %02h", TIMESCALE*$time, opcode_reg );
    if (DEBUG_OPT4) $display( "DBG: %m (@ %.3fms) t_clk_period = %f", TIMESCALE*$time, t_clk_period );

    if (DEBUG_OPT5)
    begin
	$display( "DBG: %m (@ %.3fms) opcode = %02h, background opcode = %02h, RDY/BSY = %1d, PS1 = %1d, PS2 = %1d, ES = %1d",
	    TIMESCALE*$time, opcode_reg, bg_opcode, status1_reg[ST_RDY], status2_reg[ST_PS1], status2_reg[ST_PS2], status2_reg[ST_ES] );
	$display( "DBG: %m (@ %.3fms) allow_pgm_suspend = %1d, inprog_pgm_buffer = %1d, inprog_pgm_page = %04h, inprog_pgm_start = %.3fms, inprog_pgm_delay = %.3fms",
	    TIMESCALE*$time, allow_pgm_suspend, inprog_pgm_buffer, inprog_pgm_page, TIMESCALE*inprog_pgm_start, TIMESCALE*inprog_pgm_delay );
	$display( "DBG: %m (@ %.3fms) allow_ers_suspend = %1d, inprog_ers_scope = %1d, inprog_ers_page = %04h, inprog_ers_start = %.3fms, inprog_ers_delay = %.3fms",
	    TIMESCALE*$time, allow_ers_suspend, inprog_ers_scope, inprog_ers_page, TIMESCALE*inprog_ers_start, TIMESCALE*inprog_ers_delay );
	$display( "DBG: %m (@ %.3fms) updating_memory = %1d, erasing_page = %1d, erasing_block = %1d, erasing_sector = %1d, erasing_chip = %1d, suspending = %1d, resuming = %1d",
	    TIMESCALE*$time, updating_memory, erasing_page, erasing_block, erasing_sector, erasing_chip, suspending, resuming );
    end

    if (opcode_reg === bg_opcode)
	$display( "ERROR: AT45DQXXX{#1008}, %m (@ %.3fms), Opcode %02h is already running as a background operation.", TIMESCALE*$time, opcode_reg );

    if (foreground_op_enable == 1'b0)	// No foreground or background opcode accepted
	$display( "ERROR: AT45DQXXX{#1010}, %m (@ %.3fms), Opcode %02h is not allowed, device may only be selected after time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tVCSL );
    else if (deep_power_down == 1'b1)	// Can be only after background has been enabled
    case (opcode_reg)
	8'hAB : -> RDPD;		//Resume from Deep Power-down
	default : $display( "ERROR: AT45DQXXX{#1015}, %m (@ %.3fms), Opcode %02h is not allowed while device is in Deep Power-down.", TIMESCALE*$time, opcode_reg );
    endcase
    else		// Foreground is enabled; Background can be either enabled or disabled
    case (opcode_reg)	// based on opcode, trigger an action
	8'hE8 : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 4;
		    clk_valid = sck_CAR1_valid;
		    -> MMCAR;	// Main Memory Continuous Array Read (Legacy Command) [5.1]
		end
	8'h1B : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 2;
		    clk_valid = sck_CAR1_valid;
		    -> MMCAR;	// Main Memory Continuous Array Read (High Frequency Mode) [5.2]
		end
	8'h0B : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 1;
		    clk_valid = sck_CAR1_valid;
		    -> MMCAR;	// Main Memory Continuous Array Read (High Frequency Mode) [5.3]
		end
	8'h03 : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 0;
		    clk_valid = sck_CAR2_valid;
		    -> MMCAR;	// Main Memory Continuous Array Read (Low Frequency Mode) [5.4]
		end
	8'h01 : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 0;
		    clk_valid = sck_CAR3_valid;
		    -> MMCAR;	// Main Memory Continuous Array Read (Low Power Mode) [5.5]
		end
	8'hD2 : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 4;
		    clk_valid = sck_SCK_valid;
		    -> MMPR;	// Main Memory Page Read [5.6]
		end
	8'hD4 : begin
		    skip = 1'b0;
		    buf_rd_dummybyte = 1;
		    clk_valid = sck_MAX_valid;
		    -> B1R;	// Buffer 1 Read [5.7]
		end
	8'hD6 : begin
		    skip = 1'b0;
		    buf_rd_dummybyte = 1;
		    clk_valid = sck_MAX_valid;
		    -> B2R;	// Buffer 2 Read [5.7]
		end
	8'hD1 : begin
		    skip = 1'b0;
		    buf_rd_dummybyte = 0;
		    clk_valid = sck_CAR2_valid;
		    -> B1R;	// Buffer 1 Read [5.7]
		end
	8'hD3 : begin
		    skip = 1'b0;
		    buf_rd_dummybyte = 0;
		    clk_valid = sck_CAR2_valid;
		    -> B2R;	// Buffer 2 Read [5.7]
		end
	8'h3B : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 1;
		    clk_valid = sck_SCK_valid;
		    -> DORA;	// Dual-Output Read Array [5.8]
		end
	8'h6B : begin
		    skip = 1'b0;
		    arr_rd_dummybyte = 1;
		    clk_valid = sck_SCK_valid;
		    -> QORA;	// Quad-Output Read Array [5.9]
		end
	8'h84 : -> B1W;		// Buffer 1 Write [6.1]
	8'h87 : -> B2W;		// Buffer 2 Write [6.1]
	8'h24 : -> DB1W;	// Dual-input Buffer 1 Write [6.2]
	8'h27 : -> DB2W;	// Dual-input Buffer 2 Write [6.2]
	8'h44 : -> QB1W;	// Quad-input Buffer 1 Write [6.3]
	8'h47 : -> QB2W;	// Quad-input Buffer 2 Write [6.3]
	8'h54 : begin
		    buf_rd_dummybyte = 1;
		    clk_valid = 1'b1;
		    -> B1R;	// Buffer 1 Read [Legacy?]
		end
	8'h56 : begin
		    buf_rd_dummybyte = 1;
		    clk_valid = 1'b1;
		    -> B2R;	// Buffer 2 Read [Legacy?]
		end
	8'h83 : -> B1TMMPPWBIE;	// Buffer 1 To Main Memory Page Prog With Built-In Erase [6.4]
	8'h86 : -> B2TMMPPWBIE;	// Buffer 2 To Main Memory Page Prog With Built-In Erase [6.4]
	8'h88 : -> B1TMMPPWOBIE;// Buffer 1 To Main Memory Page Prog Without Built-In Erase [6.5]
	8'h89 : -> B2TMMPPWOBIE;// Buffer 2 To Main Memory Page Prog Without Built-In Erase [6.5]
	8'h82 : -> MMPPB1;	// Main Memory Page Prog Through Buffer 1 With Built-In Erase [6.6]
	8'h85 : -> MMPPB2;	// Main Memory Page Prog Through Buffer 2 With Built-In Erase [6.6]
	8'h02 : -> MMPPB1NE;	// Main Memory Page Prog Through Buffer 1 Without Built-In Erase [6.7]
	8'h81 : -> PE;		// Page Erase [6.8]
	8'h50 : -> BE;		// Block Erase [6.9]
	8'h7C : -> SE;		// Sector Erase [6.10]
	8'hC7 : -> OPC7;	// 4-byte command starting with C7 (Chip Erase [6.11])
	8'hB0 : begin
		    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) SUSPEND event", TIMESCALE*$time );
		    -> SUSPEND;	// Suspend in-progress program or erase operation [6.12]
		end
	8'hD0 : begin
		    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) RESUME event", TIMESCALE*$time );
		    -> RESUME;	// Resume suspended program or erase operation [6.13]
		end
	8'hF0 : -> OPF0;	// 4-byte command starting with F0 (Software Reset [13])
	8'h53 : -> MMPTB1T;	// Main Memory Page To Buffer 1 Transfer [9.1]
	8'h55 : -> MMPTB2T;	// Main Memory Page To Buffer 2 Transfer [9.1]
	8'h60 : -> MMPTB1C;	// Main Memory Page To Buffer 1 Compare [9.2]
	8'h61 : -> MMPTB2C;	// Main Memory Page To Buffer 2 Compare [9.2]
	8'h93 : begin
		    fast_mode = 1'b1;
		    -> B1TMMPPWBIE;	// Buffer 1 To Main Memory Page Prog With Built-In Erase [?]
		end
	8'h96 : begin
		    fast_mode = 1'b1;
		    -> B2TMMPPWBIE;	// Buffer 2 To Main Memory Page Prog With Built-In Erase [?]
		end
	8'h98 : begin
		    fast_mode = 1'b1;
		    -> B1TMMPPWOBIE;	// Buffer 1 To Main Memory Page Prog Without Built-In Erase [?]
		end
	8'h99 : begin
		    fast_mode = 1'b1;
		    -> B2TMMPPWOBIE;	// Buffer 2 To Main Memory Page Prog Without Built-In Erase [?]
		end
	8'h92 : begin
		    fast_mode = 1'b1;
		    -> MMPPB1;	// Main Memory Page Prog. Through Buffer 1 [?]
		end
	8'h95 : begin
		    fast_mode = 1'b1;
		    -> MMPPB2;	// Main Memory Page Prog. Through Buffer 2 [?]
		end
	8'h58 : -> APRB1;	// Auto Page Rewrite (or Read-Modify-Write) Through Buffer 1 [9.3]/[6.6]
	8'h59 : -> APRB2;	// Auto Page Rewrite (or Read-Modify-Write) Through Buffer 2 [9.3]/[6.6]
	8'h57 : begin
		    -> SR;	// Status Register Read [legacy?]
		end
	8'hD7 : begin
		    skip = 1'b0;
		    -> SR;	// Status Register Read [9.4]
		end
	8'h3F : begin
		    skip = 1'b0;
		    clk_valid = sck_SCK_valid;
		    -> CSR;	// Configuration Register Read [9.5]
		end
	8'h9F : begin
		    skip = 1'b0;
		    -> MIR;	// Manufacturer ID Read [12]
		end
	8'h3D : -> OP3D;	// 4-byte command starting with 3D
	8'h32 : begin
		    skip = 1'b0;
		    -> PRR;	// Protection Register Read [7.3.3]
		end
	8'h35 : begin
		    skip = 1'b0;
		    -> LRR;	// Lock_down Register Read [8.1.1]
		end
	8'h77 : begin
		    skip = 1'b0;
		    -> SRR;	// Security Register Read [8.2.2]
		end
	8'h9B : -> OP9B;	// 4-byte command starting with 9B (Program Security Register [8.2.1])
	8'hB9 : -> EDPD;	// Deep Power-down (enable) [10]
	8'hAB : -> RDPD;	// Resume from Deep Power-down [10.1]
	8'h52 : begin
		    arr_rd_dummybyte = 4;
		    clk_valid = 1'b1;
		    -> MMPR;	// Main Memory Page Read [Legacy?]
		end
	8'h68 : begin
		    arr_rd_dummybyte = 4;
		    clk_valid = 1'b1;
		    -> MMCAR;	// Main Memory Continuous Array Read [Legacy?]
		end
	8'h34 : -> OP34;	// 4-byte command starting with 34 (Freeze Sector Lockdown [8.1.2])
	8'h79 : -> UDPD;	// Enter Ultra-Deep Power-Down [10.2]
	default : $display( "ERROR: AT45DQXXX{#1020}, %m (@ %.3fms), Unrecognized opcode %02h.", TIMESCALE*$time, opcode_reg );
    endcase
end // negedge CSB

//==============================================================================
//  Event: Program/Erase Suspend
//==============================================================================

always @(SUSPEND)
begin : SUSPEND_
    if (SUPPORT_SUSP == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1022}, %m (@ %.3fms), Opcode %02h (SUSPEND) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable SUSPEND_;
    end

    if (suspending == 1'b1  ||  resuming == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1025}, %m (@ %.3fms), Opcode %02h (SUSPEND) is not allowed, suspend or resume in progress.", TIMESCALE*$time, opcode_reg );
	disable SUSPEND_;
    end

    @( posedge CSB );

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1)  &&  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1030}, %m (@ %.3fms), Opcode %02h (SUSPEND) is not allowed, already program and erase suspended.", TIMESCALE*$time, opcode_reg );
	disable SUSPEND_;
    end

    if (status1_reg[ST_RDY] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1035}, %m (@ %.3fms), Opcode %02h (SUSPEND) ignored, device is ready, nothing to suspend.", TIMESCALE*$time, opcode_reg );
	disable SUSPEND_;
    end
    else if (updating_memory == 1'b1  &&  allow_pgm_suspend == 1'b1)
    begin
	suspending = 1'b1;
	bg_opcode = opcode_reg;
	#tSUSP
	disable B1TMMPPWBIE_;
	disable B2TMMPPWBIE_;
	disable B1TMMPPWOBIE_;
	disable B2TMMPPWOBIE_;
	disable MMPPB1_;
	disable MMPPB2_;
	disable MMPPB1NE_;
	disable RESUME_;
	if (inprog_pgm_buffer == MEM_B1)
	    status2_reg[ST_PS1] = 1'b1;
	else if (inprog_pgm_buffer == MEM_B2)
	    status2_reg[ST_PS2] = 1'b1;
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1040}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, model error, no buffer selected.", TIMESCALE*$time, opcode_reg );
	    suspending = 1'b0;
	    bg_opcode = 8'hXX;
	    disable SUSPEND_;
	end
	// Subtract elapsed time from expected program time
	inprog_pgm_delay = inprog_pgm_delay - ($time - inprog_pgm_start);
	status1_reg[ST_RDY] = 1'b1;	// Device is Ready
	updating_memory = 1'b0;
	case (inprog_pgm_buffer)
	    MEM_B1 :	inuse_buffer1 = 1'b0;
	    MEM_B2 :	inuse_buffer2 = 1'b0;
	endcase
	allow_pgm_suspend = 1'b0;
	if (DEBUG_OPT0)
	    $display( "DBG: %m (@ %.3fms) Program from buffer %1d Suspended at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_pgm_buffer, inprog_pgm_page, TIMESCALE*inprog_pgm_delay );
	suspending = 1'b0;
	bg_opcode = 8'hXX;
    end
    else if ((erasing_page == 1'b1  ||  erasing_block == 1'b1  ||  erasing_sector == 1'b1  ||  erasing_chip == 1'b1)  &&  allow_ers_suspend == 1'b1)
    begin
	suspending = 1'b1;
	bg_opcode = opcode_reg;
	#tSUSE
	disable PE_;
	disable BE_;
	disable SE_;
	disable CE_;
	disable RESUME_;
	// Subtract elapsed time from expected erase time
	inprog_ers_delay = inprog_ers_delay - ($time - inprog_ers_start);
	if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Suspend Erase, start = %.3fms, delay = %.3fms", TIMESCALE*$time, TIMESCALE*inprog_ers_start, TIMESCALE*inprog_ers_delay );
	status2_reg[ST_ES] = 1'b1;	// Device is Erase Suspended
	status1_reg[ST_RDY] = 1'b1;	// Device is Ready
	erasing_page = 1'b0;
	erasing_block = 1'b0;
	erasing_sector = 1'b0;
	erasing_chip = 1'b0;
	allow_ers_suspend = 1'b0;
	if (DEBUG_OPT0)
	    case (inprog_ers_scope)
		ERS_P : $display( "DBG: %m (@ %.3fms) Page Erase Suspended at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
		ERS_B : $display( "DBG: %m (@ %.3fms) Block Erase Suspended at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
		ERS_S : $display( "DBG: %m (@ %.3fms) Sector Erase Suspended at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
		ERS_C : $display( "DBG: %m (@ %.3fms) Chip Erase Suspended at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
	    endcase
	suspending = 1'b0;
	bg_opcode = 8'hXX;
    end
    else
    begin
	if (deep_power_down == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1045}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, deep power down (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_memory == 1'b1  &&  (updating_buffer1 == 1'b1  ||  updating_buffer2 == 1'b1))
	    $display( "ERROR: AT45DQXXX{#1050}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, auto page rewrite (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_buffer1 == 1'b1  ||  updating_buffer2 == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1055}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, transfer to buffer (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_pagesize == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1060}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, page size program (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_cfg_reg == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1065}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, config reg program (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_sec_reg == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1070}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, security reg program (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_fac_reg == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1075}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, factory reg program (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_lock_reg == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1080}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, lockdown reg program (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_prot_reg == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1085}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, protection reg program (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (updating_freeze == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1090}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, freeze lockdown (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else if (comparing == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1095}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, comparison (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	else
	    $display( "ERROR: AT45DQXXX{#1100}, %m (@ %.3fms), Opcode %02h (SUSPEND) aborted, operation (%02h) cannot be suspended.", TIMESCALE*$time, opcode_reg, bg_opcode );
	disable SUSPEND_;
    end
end // SUSPEND

//==============================================================================
//  Event: Program/Erase Resume
//==============================================================================

always @(RESUME)
begin : RESUME_
    integer page;		// Local page counter
    integer block;		// Block containing page
    integer sector;		// Sector containing page
    integer pagecount;		// Number of pages erased during erase resume
    integer blockcount;		// Number of blocks (8 pages) erased
    reg skip_page;		// Flag to skip locked or protected page

    if (SUPPORT_SUSP == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1102}, %m (@ %.3fms), Opcode %02h (RESUME) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable RESUME_;
    end

    if (suspending == 1'b1  ||  resuming == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1105}, %m (@ %.3fms), Opcode %02h (RESUME) is not allowed, suspend or resume in progress.", TIMESCALE*$time, opcode_reg );
	disable RESUME_;
    end

    if (status2_reg[ST_PS2] === 1'b0  &&  status2_reg[ST_PS1] === 1'b0  &&  status2_reg[ST_ES] === 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1110}, %m (@ %.3fms), Opcode %02h (RESUME) is not allowed, device is not program or erase suspended.", TIMESCALE*$time, opcode_reg );
	disable RESUME_;
    end

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1115}, %m (@ %.3fms), Opcode %02h (RESUME) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable RESUME_;
    end

    // allow_pgm_suspend;	// This operation may be suspended if 1'b1
    // inprog_pgm_buffer;	// Source buffer for program operation (MEM_B1 or MEM_B2)
    // inprog_pgm_page;		// Memory page being programmed
    // inprog_pgm_start;	// Start time of program operation
    // inprog_pgm_delay;	// Time needed to complete operation
    // inprog_pgm_epe;		// EPE status bit for this program operation

    // allow_ers_suspend;	// This operation may be suspended if 1'b1
    // inprog_ers_scope;	// ERS_P (page), ERS_B (block), ERS_S (sector), ERS_C (chip)
    // inprog_ers_page;		// Memory page being programmed
    // inprog_ers_start;	// Start time of program operation
    // inprog_ers_delay;	// Time needed to complete operation
    // inprog_ers_epe;		// EPE status bit for this erase operation

    @( posedge CSB );
    // Resume suspended program operation before resuming suspended erase operation
    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1)
    begin
	resuming = 1'b1;
	bg_opcode = opcode_reg;
	status1_reg[ST_RDY] = 1'b0;
	#tRESP
	if ((status2_reg[ST_PS1] == 1'b1  &&  inprog_pgm_buffer != MEM_B1) || (status2_reg[ST_PS2] == 1'b1  &&  inprog_pgm_buffer != MEM_B2))
	begin
	    $display( "ERROR: AT45DQXXX{#1120}, %m (@ %.3fms), Opcode %02h (RESUME) aborted, model error, buffer to status bit mismatch.", TIMESCALE*$time, opcode_reg );
	    resuming = 1'b0;
	    bg_opcode = 8'hXX;
	    disable RESUME_;
	end
	if (DEBUG_OPT0)
	    $display( "DBG: %m (@ %.3fms) Program from buffer %1d Resumed at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_pgm_buffer, inprog_pgm_page, TIMESCALE*inprog_pgm_delay );
	status2_reg[ST_PS2] = 1'b0;	// Reset program suspend bits
	status2_reg[ST_PS1] = 1'b0;
	updating_memory = 1'b1;
	case (inprog_pgm_buffer)
	    MEM_B1 :	inuse_buffer1 = 1'b1;
	    MEM_B2 :	inuse_buffer2 = 1'b1;
	endcase
	allow_pgm_suspend = 1'b1;
	inprog_pgm_start = $time;	// Reset start time, used by SUSPEND to update delay
	// In the model the page programming is instantaneous, so it is already completed by the
	// time we suspend the program operation.  We just need to wait out the remaining time.
	// The programming may be suspended again during this time.
	resuming = 1'b0;
	#(inprog_pgm_delay);		// Wait remaining delay time
	updating_memory = 1'b0;
	inuse_buffer1 = 1'b0;
	inuse_buffer2 = 1'b0;
	allow_pgm_suspend = 1'b0;
	inprog_pgm_buffer = 0;
	status1_reg[ST_RDY] = 1'b1;
	// status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
	status2_reg[ST_EPE] = inprog_pgm_epe;
	bg_opcode = 8'hXX;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Resumed program completed", TIMESCALE*$time );
    end
    else if (status2_reg[ST_ES] == 1'b1)
    begin
	resuming = 1'b1;
	bg_opcode = opcode_reg;
	status1_reg[ST_RDY] = 1'b0;
	#tRESE
	inprog_ers_start = $time;	// Reset start time, used by SUSPEND to update delay
	if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Erase, start = %.3fms, delay = %.3fms", TIMESCALE*$time, TIMESCALE*inprog_ers_start, TIMESCALE*inprog_ers_delay );
	status2_reg[ST_ES] = 1'b0;	// Reset erase suspend bit
	pagecount = 0;
	blockcount = 0;
	case (inprog_ers_scope)
	    ERS_P : begin
			if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Page Erase Resumed at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
			erasing_page = 1'b1;
			allow_ers_suspend = 1'b1;
			resuming = 1'b0;
			// Model already erased the page.  Nothing further to do.
		    end
	    ERS_B : begin
			if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Block Erase Resumed at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
			erasing_block = 1'b1;
			allow_ers_suspend = 1'b1;
			resuming = 1'b0;
			// Model already erased the block.  Nothing further to do.
		    end
	    ERS_S : begin
			if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Sector Erase Resumed at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
			erasing_sector = 1'b1;
			allow_ers_suspend = 1'b1;
			resuming = 1'b0;
			sector = inprog_ers_page / PAGES_PER_SECTOR;
			if (inprog_ers_page < 8)
			begin
			    // Model already erased the first block (sector 0).
			end
			else if (sector == 0)
			begin
			    for (page=inprog_ers_page+1; page<PAGES_PER_SECTOR; page=page+1)
			    begin
				inprog_ers_page = page;	// Memory page being erased
				erase_page( page );
				pagecount = pagecount + 1;
				if ((page % 8) == 7)
				begin
				    blockcount = blockcount + 1;
				    #tBEmin;		// Wait one assumed block erase time after each eight pages
				end
			    end
			end
			else
			begin
			    for (page=inprog_ers_page+1; page<(PAGES_PER_SECTOR*(sector+1)); page=page+1)
			    begin
				inprog_ers_page = page;	// Memory page being erased
				erase_page( page );
				pagecount = pagecount + 1;
				if ((page % 8) == 7)
				begin
				    blockcount = blockcount + 1;
				    #tBEmin;		// Wait one assumed block erase time after each eight pages
				end
			    end
			end
		    end
	    ERS_C : begin
			if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Chip Erase Resumed at page %04h, time remaining %.3fms", TIMESCALE*$time, inprog_ers_page, TIMESCALE*inprog_ers_delay );
			erasing_chip = 1'b1;
			allow_ers_suspend = 1'b1;
			resuming = 1'b0;
			for (page=inprog_ers_page+1; page<PAGES; page=page+1)
			begin
			    inprog_ers_page = page;		// Memory page being erased
			    sector = page / PAGES_PER_SECTOR;
			    skip_page = (check_lockdown( page ) || check_protect( page ));

			    if ((page == 0  ||  page == 8  ||  (page % PAGES_PER_SECTOR) == 0)  &&  skip_page)
			    begin
				if (page == 0)
				    $display( "ERROR: AT45DQXXX{#1125}, %m (@ %.3fms), Opcode %02h (RESUME:CE), sector 0a is protected/locked.", TIMESCALE*$time, opcode_reg );
				else if (page == 8)
				    $display( "ERROR: AT45DQXXX{#1130}, %m (@ %.3fms), Opcode %02h (RESUME:CE), sector 0b is protected/locked.", TIMESCALE*$time, opcode_reg );
				else if ((page % PAGES_PER_SECTOR) == 0)
				    $display( "ERROR: AT45DQXXX{#1135}, %m (@ %.3fms), Opcode %02h (RESUME:CE), sector %04h is protected/locked.", TIMESCALE*$time, opcode_reg, sector );
			    end

			    if (!skip_page)
			    begin
				erase_page( page );
				pagecount = pagecount + 1;
				if ((page % 8) == 7)
				begin
				    blockcount = blockcount + 1;
				    #tBEmin;		// Wait one assumed block erase time after each eight pages
				end
			    end
			end // for
		    end
	endcase

	if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Resume Erase, inprog_ers_delay = %.3fms, pagecount = %1d, blockcount = %1d, tBEmin = %.3fms, blockcount*tBEmin = %.3fms, remaining = %.3fms",
			TIMESCALE*$time, TIMESCALE*inprog_ers_delay, pagecount, blockcount, TIMESCALE*tBEmin, TIMESCALE*blockcount*tBEmin, TIMESCALE*(inprog_ers_delay - blockcount*tBEmin) );
	#(inprog_ers_delay - blockcount*tBEmin);	// Wait remaining delay time
	erasing_page = 1'b0;
	erasing_block = 1'b0;
	erasing_sector = 1'b0;
	erasing_chip = 1'b0;
	allow_ers_suspend = 1'b0;
	inprog_ers_scope = 0;
	status1_reg[ST_RDY] = 1'b1;
	// status1_reg[ST_CMP] = 1'b0;			// Reset Compare status bit
	status2_reg[ST_EPE] = inprog_ers_epe;
	bg_opcode = 8'hXX;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Resumed erase completed", TIMESCALE*$time );
    end
    else
    begin
	// It should never even get here
	$display( "ERROR: AT45DQXXX{#1140}, %m (@ %.3fms), Opcode %02h (RESUME) aborted, no suspend flags are set.", TIMESCALE*$time, opcode_reg );
	disable RESUME_;
    end
end // RESUME

//==============================================================================
//  Event: Deep Power-down
//==============================================================================

always @(EDPD)
begin : EDPD_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1145}, %m (@ %.3fms), Opcode %02h (EDPD) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable EDPD_;
    end

    @( posedge CSB );

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1150}, %m (@ %.3fms), Opcode %02h (EDPD) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	disable EDPD_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1155}, %m (@ %.3fms), Opcode %02h (EDPD) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	disable EDPD_;
    end

    bg_opcode = opcode_reg;
    #tEDPD deep_power_down = 1'b1;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Entered Deep Power-Down", TIMESCALE*$time );
end // EDPD

//==============================================================================
//  Event: Resume from Deep Power-down
//==============================================================================

always @(RDPD)
begin : RDPD_
    @( posedge CSB );
    bg_opcode = opcode_reg;
    #tRDPD deep_power_down = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Exited Deep Power-Down", TIMESCALE*$time );
end // RDPD

//==============================================================================
//  Event: Ultra-Deep Power-down
//==============================================================================

always @(UDPD)
begin : UDPD_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1160}, %m (@ %.3fms), Opcode %02h (UDPD) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable UDPD_;
    end

    @( posedge CSB );

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1165}, %m (@ %.3fms), Opcode %02h (UDPD) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	disable UDPD_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1170}, %m (@ %.3fms), Opcode %02h (UDPD) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	disable UDPD_;
    end

    bg_opcode = opcode_reg;
    corrupt_buffer( MEM_B1 );		// Destroy contents of SRAM buffer
    corrupt_buffer( MEM_B2 );		// Destroy contents of SRAM buffer

    // Ultra-Deep Power-Down is similar to a power cycle.  All volatile registers are reset.
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status1_reg[ST_PRT] = 1'b0;		// Reset Sector Protection Enabled bit

    #tEUDPD ultra_power_down = 1'b1;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Entered Ultra-Deep Power-Down", TIMESCALE*$time );
end // UDPD

//==============================================================================
//  Event: Exit Ultra-Deep Power-down
//==============================================================================

always @(XUDPD)
begin : XUDPD_
    @( posedge CSB );
    bg_opcode = opcode_reg;
    #tXUDPD ultra_power_down = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Exited Ultra-Deep Power-Down", TIMESCALE*$time );
end // XUDPD

//==============================================================================
//  Event: Main Memory Page Read
//==============================================================================

always @(MMPR)
begin : MMPR_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1175}, %m (@ %.3fms), Opcode %02h (MMPR) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPR_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPR_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    get_dummy_bytes( arr_rd_dummybyte, valid );	// Skip a number of dummy bytes
    if (!valid) disable MMPR_;

    if (skip == 1'b1) @( posedge SCK );		// Skip one SCK

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1180}, %m (@ %.3fms), Opcode %02h (MMPR) read from sector with suspended program operation.", TIMESCALE*$time, opcode_reg );
	read_out_x( S_SNGL, MEM_MM, addr, page_start, page_end );
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1185}, %m (@ %.3fms), Opcode %02h (MMPR) read from sector with suspended erase operation.", TIMESCALE*$time, opcode_reg );
	read_out_x( S_SNGL, MEM_MM, addr, page_start, page_end );
    end
    else if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1187}, %m (@ %.3fms), Opcode %02h (MMPR) byte address %03h outside DataFlash range [0 to %03h], output XX's", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	read_out_x( S_SNGL, MEM_MM, addr, page_start, page_end );
    end
    else if (clk_valid != 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1190}, %m (@ %.3fms), Opcode %02h (MMPR) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	read_out_x( S_SNGL, MEM_MM, addr, page_start, page_end );
    end
    else
	read_out( S_SNGL, MEM_MM, addr, page_start, page_end );
end // MMPR

//==============================================================================
//  Event: Main Memory Continuous Array Read
//==============================================================================

always @(MMCAR)
begin : MMCAR_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1195}, %m (@ %.3fms), Opcode %02h (MMCAR) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMCAR_;
    end

    get_address( addr, valid );
    if (!valid) disable MMCAR_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) addr = %06h, page = %04h, page_start = %06h, page_end = %06h, inpage_addr = %03h", TIMESCALE*$time, addr, page, page_start, page_end, inpage_addr );

    get_dummy_bytes( arr_rd_dummybyte, valid );	// Skip a number of dummy bytes
    if (!valid) disable MMCAR_;

    if (skip == 1'b1) @( posedge SCK );		// Skip one SCK

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1197}, %m (@ %.3fms), Opcode %02h (MMCAR) byte address %03h outside DataFlash range [0 to %03h], output XX's", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	cont_read_out_x( S_SNGL, addr, page_start, page_end );
    end
    else if (clk_valid != 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1200}, %m (@ %.3fms), Opcode %02h (MMCAR) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	cont_read_out_x( S_SNGL, addr, page_start, page_end );
    end
    else
	cont_read_out( S_SNGL, addr, page_start, page_end );
end // MMCAR

//==============================================================================
//  Event: Dual-Output Read Array
//==============================================================================

always @(DORA)
begin : DORA_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1202}, %m (@ %.3fms), Opcode %02h (DORA) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable DORA_;
    end

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1205}, %m (@ %.3fms), Opcode %02h (DORA) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable DORA_;
    end

    get_address( addr, valid );
    if (!valid) disable DORA_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) addr = %06h, page = %04h, page_start = %06h, page_end = %06h, inpage_addr = %03h", TIMESCALE*$time, addr, page, page_start, page_end, inpage_addr );

    get_dummy_bytes( arr_rd_dummybyte, valid );	// Skip a number of dummy bytes
    if (!valid) disable DORA_;

    if (skip == 1'b1) @( posedge SCK );		// Skip one SCK

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1207}, %m (@ %.3fms), Opcode %02h (DORA) byte address %03h outside DataFlash range [0 to %03h], output XX's", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	cont_read_out_x( S_DUAL, addr, page_start, page_end );
    end
    else if (clk_valid != 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1210}, %m (@ %.3fms), Opcode %02h (DORA) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	cont_read_out_x( S_DUAL, addr, page_start, page_end );
    end
    else
	cont_read_out( S_DUAL, addr, page_start, page_end );
end // DORA

//==============================================================================
//  Event: Quad-Output Read Array
//==============================================================================

always @(QORA)
begin : QORA_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1212}, %m (@ %.3fms), Opcode %02h (QORA) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable QORA_;
    end

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1215}, %m (@ %.3fms), Opcode %02h (QORA) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable QORA_;
    end

    get_address( addr, valid );
    if (!valid) disable QORA_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) addr = %06h, page = %04h, page_start = %06h, page_end = %06h, inpage_addr = %03h", TIMESCALE*$time, addr, page, page_start, page_end, inpage_addr );

    if (QE_reg !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1217}, %m (@ %.3fms), Opcode %02h (QORA) aborted, quad-mode is disabled.", TIMESCALE*$time, opcode_reg );
	disable QORA_;
    end

    get_dummy_bytes( arr_rd_dummybyte, valid );	// Skip a number of dummy bytes
    if (!valid) disable QORA_;

    if (skip == 1'b1) @( posedge SCK );		// Skip one SCK

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1218}, %m (@ %.3fms), Opcode %02h (QORA) byte address %03h outside DataFlash range [0 to %03h], output XX's", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	cont_read_out_x( S_QUAD, addr, page_start, page_end );
    end
    else if (clk_valid != 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1220}, %m (@ %.3fms), Opcode %02h (QORA) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	cont_read_out_x( S_QUAD, addr, page_start, page_end );
    end
    else
	cont_read_out( S_QUAD, addr, page_start, page_end );
end // QORA

//==============================================================================
//  Event: Buffer 1 Read
//==============================================================================

always @(B1R)
begin : B1R_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (inuse_buffer1 == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1225}, %m (@ %.3fms), Opcode %02h (B1R) aborted, Buffer 1 is being used in a background operation.", TIMESCALE*$time, opcode_reg );
	disable B1R_;
    end

    get_address( addr, valid );
    if (!valid) disable B1R_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    get_dummy_bytes( buf_rd_dummybyte, valid );	// Skip a number of dummy bytes
    if (!valid) disable B1R_;

    if (skip == 1'b1) @( posedge SCK );		// Skip one SCK

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1227}, %m (@ %.3fms), Opcode %02h (B1R) byte address %03h outside DataFlash range [0 to %03h], output XX's", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	read_out_x( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1 );
    end
    else if (clk_valid != 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1230}, %m (@ %.3fms), Opcode %02h (B1R) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	read_out_x( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1 );
    end
    else
	read_out( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1 );
end // B1R

//==============================================================================
//  Event: Buffer 2 Read
//==============================================================================

always @(B2R)
begin : B2R_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1235}, %m (@ %.3fms), Opcode %02h (B2R) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable B2R_;
    end

    if (inuse_buffer2 == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1240}, %m (@ %.3fms), Opcode %02h (B2R) aborted, Buffer 2 is being used in a background operation.", TIMESCALE*$time, opcode_reg );
	disable B2R_;
    end

    get_address( addr, valid );
    if (!valid) disable B2R_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    get_dummy_bytes( buf_rd_dummybyte, valid );	// Skip a number of dummy bytes
    if (!valid) disable B2R_;

    if (skip == 1'b1) @( posedge SCK );		// Skip one SCK

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1242}, %m (@ %.3fms), Opcode %02h (B2R) byte address %03h outside DataFlash range [0 to %03h], output XX's", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	read_out_x( S_SNGL, MEM_B2, inpage_addr, 0, page_size-1 );
    end
    else if (clk_valid != 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1245}, %m (@ %.3fms), Opcode %02h (B2R) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	read_out_x( S_SNGL, MEM_B2, inpage_addr, 0, page_size-1 );
    end
    else
	read_out( S_SNGL, MEM_B2, inpage_addr, 0, page_size-1 );
end // B2R

//==============================================================================
//  Event: Main Memory Page To Buffer 1 Transfer
//==============================================================================

always @(MMPTB1T)
begin : MMPTB1T_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1250}, %m (@ %.3fms), Opcode %02h (MMPTB1T) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPTB1T_;
    end

    if (status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1255}, %m (@ %.3fms), Opcode %02h (MMPTB1T) is not allowed during PS1 program suspend.", TIMESCALE*$time, opcode_reg );
	disable MMPTB1T_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPTB1T_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1260}, %m (@ %.3fms), Opcode %02h (MMPTB1T) reading from a program suspended sector, corrupting buffer 1.", TIMESCALE*$time, opcode_reg );
	corrupt_buffer( MEM_B1 );
	#tXFR
	disable MMPTB1T_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1265}, %m (@ %.3fms), Opcode %02h (MMPTB1T) reading from an erase suspended sector, corrupting buffer 1.", TIMESCALE*$time, opcode_reg );
	corrupt_buffer( MEM_B1 );
	#tXFR
	disable MMPTB1T_;
    end

    @( posedge CSB );
    bg_opcode = opcode_reg;
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    updating_buffer1 = 1'b1;
    inuse_buffer1 = 1'b1;
    transfer_to_buffer( MEM_B1, page_start );
    #tXFR
    status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;	// Device is Ready
    updating_buffer1 = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPTB1T

//==============================================================================
//  Event: Main Memory Page To Buffer 2 Transfer
//==============================================================================

always @(MMPTB2T)
begin : MMPTB2T_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1270}, %m (@ %.3fms), Opcode %02h (MMPTB2T) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable MMPTB2T_;
    end

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1275}, %m (@ %.3fms), Opcode %02h (MMPTB2T) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPTB2T_;
    end

    if (status2_reg[ST_PS2] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1280}, %m (@ %.3fms), Opcode %02h (MMPTB2T) is not allowed during PS2 program suspend.", TIMESCALE*$time, opcode_reg );
	disable MMPTB2T_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPTB2T_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1285}, %m (@ %.3fms), Opcode %02h (MMPTB2T) reading from a program suspended sector, corrupting buffer 2.", TIMESCALE*$time, opcode_reg );
	corrupt_buffer( MEM_B2 );
	#tXFR
	disable MMPTB2T_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1290}, %m (@ %.3fms), Opcode %02h (MMPTB2T) reading from an erase suspended sector, corrupting buffer 2.", TIMESCALE*$time, opcode_reg );
	corrupt_buffer( MEM_B2 );
	#tXFR
	disable MMPTB2T_;
    end

    @( posedge CSB );
    bg_opcode = opcode_reg;
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    updating_buffer2 = 1'b1;
    inuse_buffer2 = 1'b1;
    transfer_to_buffer( MEM_B2, page_start );
    #tXFR
    status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;	// Device is Ready
    updating_buffer2 = 1'b0;
    inuse_buffer2 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPTB2T

//==============================================================================
//  Event: Main Memory Page To Buffer 1 Compare
//==============================================================================

always @(MMPTB1C)
begin : MMPTB1C_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1295}, %m (@ %.3fms), Opcode %02h (MMPTB1C) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPTB1C_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPTB1C_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1300}, %m (@ %.3fms), Opcode %02h (MMPTB1C) reading from a program suspended sector.", TIMESCALE*$time, opcode_reg );
	status1_reg[ST_CMP] = 1'bx;
	#tXFR
	disable MMPTB1C_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1305}, %m (@ %.3fms), Opcode %02h (MMPTB1C) reading from an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	status1_reg[ST_CMP] = 1'bx;
	#tXFR
	disable MMPTB1C_;
    end

    @( posedge CSB );
    bg_opcode = opcode_reg;
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    compare_with_buffer( MEM_B1, page_start );
    comparing = 1'b1;
    inuse_buffer1 = 1'b1;
    #tXFR
    status1_reg[ST_RDY] = 1'b1;	// Device is Ready
    comparing = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPTB1C

//==============================================================================
//  Event: Main Memory Page To Buffer 2 Compare
//==============================================================================

always @(MMPTB2C)
begin : MMPTB2C_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1310}, %m (@ %.3fms), Opcode %02h (MMPTB2C) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable MMPTB2C_;
    end

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1315}, %m (@ %.3fms), Opcode %02h (MMPTB2C) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPTB2C_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPTB2C_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1320}, %m (@ %.3fms), Opcode %02h (MMPTB2C) reading from a program suspended sector.", TIMESCALE*$time, opcode_reg );
	status1_reg[ST_CMP] = 1'bx;
	#tXFR
	disable MMPTB2C_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1325}, %m (@ %.3fms), Opcode %02h (MMPTB2C) reading from an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	status1_reg[ST_CMP] = 1'bx;
	#tXFR
	disable MMPTB2C_;
    end

    @( posedge CSB );
    bg_opcode = opcode_reg;
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    compare_with_buffer( MEM_B2, page_start );
    comparing = 1'b1;
    inuse_buffer2 = 1'b1;
    #tXFR
    status1_reg[ST_RDY] = 1'b1;	// Device is Ready
    comparing = 1'b0;
    inuse_buffer2 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPTB2C

//==============================================================================
//  Event: Buffer 1 Write
//==============================================================================

always @(B1W)
begin : B1W_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1330}, %m (@ %.3fms), Opcode %02h (B1W) is not allowed during PS1 program suspend.", TIMESCALE*$time, opcode_reg );
	disable B1W_;
    end

    get_address( addr, valid );
    if (!valid) disable B1W_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1332}, %m (@ %.3fms), Opcode %02h (B1W) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 1", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	corrupt_buffer( MEM_B1 );
    end
    else 
	write_buffer( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1, bytecount );
end // B1W

//==============================================================================
//  Event: Buffer 2 Write
//==============================================================================

always @(B2W)
begin : B2W_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1335}, %m (@ %.3fms), Opcode %02h (B2W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable B2W_;
    end

    if (status2_reg[ST_PS2] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1340}, %m (@ %.3fms), Opcode %02h (B2W) is not allowed during PS2 program suspend.", TIMESCALE*$time, opcode_reg );
	disable B2W_;
    end

    get_address( addr, valid );
    if (!valid) disable B2W_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1341}, %m (@ %.3fms), Opcode %02h (B2W) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 2", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	corrupt_buffer( MEM_B2 );
    end
    else 
	write_buffer( S_SNGL, MEM_B2, inpage_addr, 0, page_size-1, bytecount );
end // B2W

//==============================================================================
//  Event: Dual-input Buffer 1 Write
//==============================================================================

always @(DB1W)
begin : DB1W_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1342}, %m (@ %.3fms), Opcode %02h (DB1W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable DB1W_;
    end

    if (status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1345}, %m (@ %.3fms), Opcode %02h (DB1W) is not allowed during PS1 program suspend.", TIMESCALE*$time, opcode_reg );
	disable DB1W_;
    end

    get_address( addr, valid );
    if (!valid) disable DB1W_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1346}, %m (@ %.3fms), Opcode %02h (DB1W) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 1", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	corrupt_buffer( MEM_B1 );
    end
    else 
	write_buffer( S_DUAL, MEM_B1, inpage_addr, 0, page_size-1, bytecount );
end // DB1W

//==============================================================================
//  Event: Dual-input Buffer 2 Write
//==============================================================================

always @(DB2W)
begin : DB2W_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1347}, %m (@ %.3fms), Opcode %02h (DB2W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable DB2W_;
    end

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1350}, %m (@ %.3fms), Opcode %02h (DB2W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable DB2W_;
    end

    if (status2_reg[ST_PS2] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1355}, %m (@ %.3fms), Opcode %02h (DB2W) is not allowed during PS2 program suspend.", TIMESCALE*$time, opcode_reg );
	disable DB2W_;
    end

    get_address( addr, valid );
    if (!valid) disable DB2W_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1356}, %m (@ %.3fms), Opcode %02h (DB2W) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 2", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	corrupt_buffer( MEM_B2 );
    end
    else 
	write_buffer( S_DUAL, MEM_B2, inpage_addr, 0, page_size-1, bytecount );
end // DB2W

//==============================================================================
//  Event: Quad-input Buffer 1 Write
//==============================================================================

always @(QB1W)
begin : QB1W_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1357}, %m (@ %.3fms), Opcode %02h (QB1W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable QB1W_;
    end

    if (status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1360}, %m (@ %.3fms), Opcode %02h (QB1W) is not allowed during PS1 program suspend.", TIMESCALE*$time, opcode_reg );
	disable QB1W_;
    end

    get_address( addr, valid );
    if (!valid) disable QB1W_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1361}, %m (@ %.3fms), Opcode %02h (QB1W) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 1", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	corrupt_buffer( MEM_B1 );
    end
    else 
	write_buffer( S_QUAD, MEM_B1, inpage_addr, 0, page_size-1, bytecount );
end // QB1W

//==============================================================================
//  Event: Quad-input Buffer 2 Write
//==============================================================================

always @(QB2W)
begin : QB2W_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1362}, %m (@ %.3fms), Opcode %02h (QB2W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable QB2W_;
    end

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1365}, %m (@ %.3fms), Opcode %02h (QB2W) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable QB2W_;
    end

    if (status2_reg[ST_PS2] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1370}, %m (@ %.3fms), Opcode %02h (QB2W) is not allowed during PS2 program suspend.", TIMESCALE*$time, opcode_reg );
	disable QB2W_;
    end

    get_address( addr, valid );
    if (!valid) disable QB2W_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1371}, %m (@ %.3fms), Opcode %02h (QB2W) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 2", TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1 );
	corrupt_buffer( MEM_B2 );
    end
    else 
	write_buffer( S_QUAD, MEM_B2, inpage_addr, 0, page_size-1, bytecount );
end // QB2W

//==============================================================================
//  Event: Buffer 1 To Main Memory Page Prog With Built In Erase
//==============================================================================

always @(B1TMMPPWBIE)
begin : B1TMMPPWBIE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1375}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable B1TMMPPWBIE_;
    end

    get_address( addr, valid );
    if (!valid) disable B1TMMPPWBIE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    @( posedge CSB );
    bg_opcode = opcode_reg;
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1380}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWBIE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1385}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWBIE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1390}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWBIE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1394}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1395}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWBIE_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1400}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWBIE_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1405}, %m (@ %.3fms), Opcode %02h (B1TMMPPWBIE) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWBIE_;
    end

    updating_memory = 1'b1;
    inuse_buffer1 = 1'b1;

    erase_page( page );			// Erase page, set contents to FF
    write_to_memory( MEM_B1, page, page_start );

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B1;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = tEP;		// Time needed to complete operation
    inprog_pgm_epe = 1'b0;		// EPE status bit for this program operation

    #tEP
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    updating_memory = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // B1TMMPPWBIE

//==============================================================================
//  Event: Buffer 2 To Main Memory Page Prog With Built In Erase
//==============================================================================

always @(B2TMMPPWBIE)
begin : B2TMMPPWBIE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1410}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable B2TMMPPWBIE_;
    end

    get_address( addr, valid );
    if (!valid) disable B2TMMPPWBIE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1415}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1420}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1425}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1430}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1434}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1435}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1440}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1445}, %m (@ %.3fms), Opcode %02h (B2TMMPPWBIE) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWBIE_;
    end

    updating_memory = 1'b1;
    inuse_buffer2 = 1'b1;

    erase_page( page );			// Erase page, set contents to FF
    write_to_memory( MEM_B2, page, page_start );

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B2;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = tEP;		// Time needed to complete operation
    inprog_pgm_epe = 1'b0;		// EPE status bit for this program operation

    #tEP
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    updating_memory = 1'b0;
    inuse_buffer2 = 1'b0;
    bg_opcode = 8'hXX;
end

//==============================================================================
//  Event: Buffer 1 To Main Memory Page Prog WithOut Built In Erase
//==============================================================================

always @(B1TMMPPWOBIE)
begin : B1TMMPPWOBIE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    reg program_ok;		// Flag to indicate program success (1) or failure (0)

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1450}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable B1TMMPPWOBIE_;
    end

    get_address( addr, valid );
    if (!valid) disable B1TMMPPWOBIE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1455}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWOBIE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1460}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) is not allowed during program suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWOBIE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1465}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWOBIE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1469}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1470}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWOBIE_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1475}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWOBIE_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1480}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B1TMMPPWOBIE_;
    end

    if (page_status[page] == 1'b0)	// Page status: 1'b0 == Erased
	program_ok = 1'b1;
    else
    begin
	$display( "ERROR: AT45DQXXX{#1485}, %m (@ %.3fms), Opcode %02h (B1TMMPPWOBIE) failed, page %04h was not erased.", TIMESCALE*$time, opcode_reg, page );
	program_ok = 1'b0;
    end

    updating_memory = 1'b1;
    inuse_buffer1 = 1'b1;

    // Program (bit-AND) regardless of page erase status
    write_to_memory( MEM_B1, page, page_start );

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B1;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = (fast_mode ? tP : tEP);	// Time needed to complete
    inprog_pgm_epe = ~program_ok;	// EPE status bit for this program operation

    if (fast_mode == 1'b1)
	#tP;
    else
	#tEP;

    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    updating_memory = 1'b0;
    inuse_buffer1 = 1'b0;
    fast_mode = 1'b0;

    if (program_ok == 1'b1)
	status2_reg[ST_EPE] = 1'b0;	// Reset EPE status bit (no error)
    else
	status2_reg[ST_EPE] = 1'b1;	// Set EPE status bit (program verify error)

    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    bg_opcode = 8'hXX;
end // B1TMMPPWOBIE

//==============================================================================
//  Event: Buffer 2 To Main Memory Page Prog WithOut Built In Erase
//==============================================================================

always @(B2TMMPPWOBIE)
begin : B2TMMPPWOBIE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    reg program_ok;		// Flag to indicate program success (1) or failure (0)

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1490}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable B2TMMPPWOBIE_;
    end

    get_address( addr, valid );
    if (!valid) disable B2TMMPPWOBIE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1495}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1500}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1505}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) is not allowed during program suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1510}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1514}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1515}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1520}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1525}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable B2TMMPPWOBIE_;
    end

    if (page_status[page] == 1'b0)	// Page status: 1'b0 == Erased
	program_ok = 1'b1;
    else
    begin
	$display( "ERROR: AT45DQXXX{#1530}, %m (@ %.3fms), Opcode %02h (B2TMMPPWOBIE) failed, page %04h was not erased.", TIMESCALE*$time, opcode_reg, page );
	program_ok = 1'b0;
    end

    updating_memory = 1'b1;
    inuse_buffer2 = 1'b1;

    // Program (bit-AND) regardless of page erase status
    write_to_memory( MEM_B2, page, page_start );

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B2;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = (fast_mode ? tP : tEP);	// Time needed to complete
    inprog_pgm_epe = ~program_ok;	// EPE status bit for this program operation

    if (fast_mode == 1'b1)
	#tP;
    else
	#tEP;

    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    updating_memory = 1'b0;
    inuse_buffer2 = 1'b0;
    fast_mode = 1'b0;

    if (program_ok == 1'b1)
	status2_reg[ST_EPE] = 1'b0;	// Reset EPE status bit (no error)
    else
	status2_reg[ST_EPE] = 1'b1;	// Set EPE status bit (program verify error)

    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    bg_opcode = 8'hXX;
end // B2TMMPPWOBIE

//==============================================================================
//  Event: Page Erase
//==============================================================================

always @(PE)
begin : PE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1535}, %m (@ %.3fms), Opcode %02h (PE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable PE_;
    end

    get_address( addr, valid );
    if (!valid) disable PE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1540}, %m (@ %.3fms), Opcode %02h (PE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1545}, %m (@ %.3fms), Opcode %02h (PE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1550}, %m (@ %.3fms), Opcode %02h (PE) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1554}, %m (@ %.3fms), Opcode %02h (PE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1555}, %m (@ %.3fms), Opcode %02h (PE) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PE_;
    end

    erasing_page = 1'b1;
    erase_page( page );

    allow_ers_suspend = 1'b1;		// This operation may be suspended
    inprog_ers_scope = ERS_P;		// Scope of erase operation is one page
    inprog_ers_page = page;		// Memory page being programmed
    inprog_ers_start = $time;		// Start time of program operation
    inprog_ers_delay = tPE;		// Time needed to complete operation
    inprog_ers_epe = 1'b0;		// EPE status bit for this erase operation

    #tPE
    allow_ers_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_ers_scope = 0;		// 0: No erase is in progress

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    erasing_page = 1'b0;
    bg_opcode = 8'hXX;
end // PE

//==============================================================================
//  Event: Block Erase
//==============================================================================

always @(BE)
begin : BE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer block;		// Block containing specified address

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1560}, %m (@ %.3fms), Opcode %02h (BE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable BE_;
    end

    get_address( addr, valid );
    if (!valid) disable BE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );
    block = page / 8;			// Eight pages in one block

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1565}, %m (@ %.3fms), Opcode %02h (BE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1570}, %m (@ %.3fms), Opcode %02h (BE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1575}, %m (@ %.3fms), Opcode %02h (BE) aborted, block %04h is locked.", TIMESCALE*$time, opcode_reg, block );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1579}, %m (@ %.3fms), Opcode %02h (BE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1580}, %m (@ %.3fms), Opcode %02h (BE) aborted, block %04h is protected.", TIMESCALE*$time, opcode_reg, block );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BE_;
    end

    erasing_block = 1'b1;

    allow_ers_suspend = 1'b1;		// This operation may be suspended
    inprog_ers_scope = ERS_B;		// Scope of erase operation is one page
    inprog_ers_page = page;		// Memory page being programmed
    inprog_ers_start = $time;		// Start time of program operation
    inprog_ers_delay = tBE;		// Time needed to complete operation
    inprog_ers_epe = 1'b0;		// EPE status bit for this erase operation

    if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) Erasing Block %04h starting at page %04h", TIMESCALE*$time, block, 8*block );
    for (page=8*block; page<(8*(block+1)); page=page+1)
    begin
	inprog_ers_page = page;		// Memory page being erased
	erase_page( page );		// All eight pages erased simultaneously
    end

    #tBE				// Delay for one block erase
    allow_ers_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_ers_scope = 0;		// 0: No erase is in progress

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    erasing_block = 1'b0;
    bg_opcode = 8'hXX;
end // BE

//==============================================================================
//  Event: Sector Erase
//==============================================================================

always @(SE)
begin : SE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer sector;		// Sector containing specified address
    integer blockcount;		// Number of blocks (8 pages) erased

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1585}, %m (@ %.3fms), Opcode %02h (SE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable SE_;
    end

    get_address( addr, valid );
    if (!valid) disable SE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );
    sector = page / PAGES_PER_SECTOR;

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1590}, %m (@ %.3fms), Opcode %02h (SE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable SE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1595}, %m (@ %.3fms), Opcode %02h (SE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable SE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	if (page < 8)
	    $display( "ERROR: AT45DQXXX{#1600}, %m (@ %.3fms), Opcode %02h (SE) aborted, sector 0a is locked.", TIMESCALE*$time, opcode_reg );
	else if (sector == 0)
	    $display( "ERROR: AT45DQXXX{#1605}, %m (@ %.3fms), Opcode %02h (SE) aborted, sector 0b is locked.", TIMESCALE*$time, opcode_reg );
	else
	    $display( "ERROR: AT45DQXXX{#1610}, %m (@ %.3fms), Opcode %02h (SE) aborted, sector %04h is locked.", TIMESCALE*$time, opcode_reg, sector );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable SE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1614}, %m (@ %.3fms), Opcode %02h (SE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    if (page < 8)
		$display( "ERROR: AT45DQXXX{#1615}, %m (@ %.3fms), Opcode %02h (SE) aborted, sector 0a is protected.", TIMESCALE*$time, opcode_reg );
	    else if (sector == 0)
		$display( "ERROR: AT45DQXXX{#1620}, %m (@ %.3fms), Opcode %02h (SE) aborted, sector 0b is protected.", TIMESCALE*$time, opcode_reg );
	    else
		$display( "ERROR: AT45DQXXX{#1625}, %m (@ %.3fms), Opcode %02h (SE) aborted, sector %04h is protected.", TIMESCALE*$time, opcode_reg, sector );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable SE_;
    end

    erasing_sector = 1'b1;

    allow_ers_suspend = 1'b1;		// This operation may be suspended
    inprog_ers_scope = ERS_S;		// Scope of erase operation is one page
    inprog_ers_start = $time;		// Start time of erase operation
    inprog_ers_delay = tSE;		// Time needed to complete operation
    inprog_ers_epe = 1'b0;		// EPE status bit for this erase operation

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Sector Erase, start = %.3fms, delay = %.3fms", TIMESCALE*$time, TIMESCALE*inprog_ers_start, TIMESCALE*inprog_ers_delay );

    blockcount = 0;
    if (page < 8)
    begin
	if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) Erasing Sector 0a starting at page 0", TIMESCALE*$time );
	for (page=0; page<8; page=page+1)
	begin
	    inprog_ers_page = page;	// Memory page being erased
	    erase_page( page );		// Eight pages erased simultaneously
	end
	blockcount = 1;
	#tSE;
    end
    else if (sector == 0)
    begin
	if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) Erasing Sector 0b starting at page 8", TIMESCALE*$time );
	for (page=8; page<PAGES_PER_SECTOR; page=page+1)
	begin
	    inprog_ers_page = page;	// Memory page being erased
	    erase_page( page );
	    if ((page % 8) == 7)
	    begin
		blockcount = blockcount + 1;
		#tBEmin;		// Wait one assumed block erase time after each eight pages
	    end
	end
	#(tSE - blockcount*tBEmin);
    end
    else
    begin
	if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) Erasing Sector %04h starting at page %04h", TIMESCALE*$time, sector, PAGES_PER_SECTOR*sector );
	for (page=PAGES_PER_SECTOR*sector; page<(PAGES_PER_SECTOR*(sector+1)); page=page+1)
	begin
	    inprog_ers_page = page;	// Memory page being erased
	    erase_page( page );
	    if ((page % 8) == 7)
	    begin
		blockcount = blockcount + 1;
		#tBEmin;		// Wait one assumed block erase time after each eight pages
	    end
	end
	#(tSE - blockcount*tBEmin);
    end

    allow_ers_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_ers_scope = 0;		// 0: No erase is in progress

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    erasing_sector = 1'b0;
    bg_opcode = 8'hXX;
end // SE

//==============================================================================
//  Event: 4-byte Command starting with C7
//==============================================================================

always @(OPC7)
begin : OPC7_
    reg valid;			// Local flag to test serial input
    reg [7:0] byte1;		// Register to store first byte of opcode
    reg [7:0] byte2;		// Register to store second byte of opcode
    reg [7:0] byte3;		// Register to store third byte of opcode
    reg [7:0] byte4;		// Register to store fourth byte of opcode
    reg [31:0] long_opcode;	// Register to store full 4-byte opcode

    byte1 = 8'hC7;				// Set 1st byte of 4-byte opcode

    serial_in( S_SNGL, byte2, valid );		// Get 2nd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte3, valid );	// Get 3rd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte4, valid );	// Get 4th byte of 4-byte opcode

    if (!valid)
    begin
	$display( "ERROR: AT45DQXXX{#1630}, %m (@ %.3fms), Opcode %02h (OPC7) aborted.", TIMESCALE*$time, opcode_reg );
	disable OPC7_;
    end

    long_opcode[31:24] = byte1;
    long_opcode[23:16] = byte2;
    long_opcode[15:8]  = byte3;
    long_opcode[7:0]   = byte4;

    case (long_opcode)
	32'hC7_94_80_9A : -> CE;		// Chip Erase
	default :
	    begin
		$display( "ERROR: AT45DQXXX{#1635}, %m (@ %.3fms), Unrecognized 4-byte opcode %02h_%02h_%02h_%02h", TIMESCALE*$time, byte1, byte2, byte3, byte4 );
		disable   OPC7_;
	    end
    endcase
end // OPC7

//==============================================================================
//  Event: Chip Erase
//==============================================================================

always @(CE)
begin : CE_
    reg valid;			// Local flag to test serial input
    integer page;		// Page component of address
    integer sector;		// Block containing specified address
    reg skip_page;		// Flag to skip locked or protected page
    integer pagecount;		// Number of pages erased
    integer blockcount;		// Number of blocks (8 pages) erased

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1640}, %m (@ %.3fms), Opcode %02h (CE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable CE_;
    end

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1645}, %m (@ %.3fms), Opcode %02h (CE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable CE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1650}, %m (@ %.3fms), Opcode %02h (CE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable CE_;
    end

    erasing_chip = 1'b1;
    pagecount = 0;
    blockcount = 0;

    allow_ers_suspend = 1'b1;		// This operation may be suspended
    inprog_ers_scope = ERS_C;		// Scope of erase operation is one page
    inprog_ers_start = $time;		// Start time of program operation
    inprog_ers_delay = tCE;		// Time needed to complete operation
    inprog_ers_epe = 1'b0;		// EPE status bit for this erase operation

    for (page=0; page<PAGES; page=page+1)
    begin
	inprog_ers_page = page;		// Memory page being erased
	sector = page / PAGES_PER_SECTOR;
	skip_page = (check_lockdown( page ) || check_protect( page ));

	if ((page == 0  ||  page == 8  ||  (page % PAGES_PER_SECTOR) == 0)  &&  skip_page)
	begin
	    if (page == 0)
		$display( "ERROR: AT45DQXXX{#1655}, %m (@ %.3fms), Opcode %02h (CE), sector 0a is protected or locked.", TIMESCALE*$time, opcode_reg );
	    else if (page == 8)
		$display( "ERROR: AT45DQXXX{#1660}, %m (@ %.3fms), Opcode %02h (CE), sector 0b is protected or locked.", TIMESCALE*$time, opcode_reg );
	    else if ((page % PAGES_PER_SECTOR) == 0)
		$display( "ERROR: AT45DQXXX{#1665}, %m (@ %.3fms), Opcode %02h (CE), sector %04h is protected or locked.", TIMESCALE*$time, opcode_reg, sector );
	end

	if (!skip_page)
	begin
	    erase_page( page );
	    pagecount = pagecount + 1;
	    if ((page % 8) == 7)
	    begin
		blockcount = blockcount + 1;
		#tBEmin;		// Wait one assumed block erase time after each eight pages
	    end
	end
    end // for

    #(tCE - blockcount*tBEmin)
    allow_ers_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_ers_scope = 0;		// 0: No erase is in progress

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    erasing_chip = 1'b0;
    bg_opcode = 8'hXX;
end // CE

//==============================================================================
//  Event: Main Memory Page Prog Through Buffer 1 (with Erase)
//==============================================================================

always @(MMPPB1)
begin : MMPPB1_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1670}, %m (@ %.3fms), Opcode %02h (MMPPB1) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPPB1_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPPB1_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1672}, %m (@ %.3fms), Opcode %02h (MMPPB1) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 1 and page %04h",
	    TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1, page );
	corrupt_buffer( MEM_B1 );
	// Corrupted buffer will be written to main memory
	@( posedge CSB );
    end
    else write_buffer( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1, bytecount );
    // write_buffer ends when CSB goes high
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1675}, %m (@ %.3fms), Opcode %02h (MMPPB1) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1680}, %m (@ %.3fms), Opcode %02h (MMPPB1) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1685}, %m (@ %.3fms), Opcode %02h (MMPPB1) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1689}, %m (@ %.3fms), Opcode %02h (MMPPB1) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1690}, %m (@ %.3fms), Opcode %02h (MMPPB1) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1695}, %m (@ %.3fms), Opcode %02h (MMPPB1) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1700}, %m (@ %.3fms), Opcode %02h (MMPPB1) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1_;
    end

    updating_memory = 1'b1;
    inuse_buffer1 = 1'b1;

    erase_page( page );			// Erase page, set contents to FF
    write_to_memory( MEM_B1, page, page_start );

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B1;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = tEP;		// Time needed to complete operation
    inprog_pgm_epe = 1'b0;		// EPE status bit for this program operation

    #tEP
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    updating_memory = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPPB1

//==============================================================================
//  Event: Main Memory Page Prog Through Buffer 2 (with Erase)
//==============================================================================

always @(MMPPB2)
begin : MMPPB2_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1705}, %m (@ %.3fms), Opcode %02h (MMPPB2) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPPB2_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPPB2_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1707}, %m (@ %.3fms), Opcode %02h (MMPPB2) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 2 and page %04h",
	    TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1, page );
	corrupt_buffer( MEM_B2 );
	// Corrupted buffer will be written to main memory
	@( posedge CSB );
    end
    else write_buffer( S_SNGL, MEM_B2, inpage_addr, 0, page_size-1, bytecount );
    // write_buffer ends when CSB goes high
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1710}, %m (@ %.3fms), Opcode %02h (MMPPB2) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1715}, %m (@ %.3fms), Opcode %02h (MMPPB2) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1720}, %m (@ %.3fms), Opcode %02h (MMPPB2) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1725}, %m (@ %.3fms), Opcode %02h (MMPPB2) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1729}, %m (@ %.3fms), Opcode %02h (MMPPB2) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1730}, %m (@ %.3fms), Opcode %02h (MMPPB2) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1735}, %m (@ %.3fms), Opcode %02h (MMPPB2) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1740}, %m (@ %.3fms), Opcode %02h (MMPPB2) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB2_;
    end

    updating_memory = 1'b1;
    inuse_buffer2 = 1'b1;

    erase_page( page );			// Erase page, set contents to FF
    write_to_memory( MEM_B2, page, page_start );

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B2;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = tEP;		// Time needed to complete operation
    inprog_pgm_epe = 1'b0;		// EPE status bit for this program operation

    #tEP
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    updating_memory = 1'b0;
    inuse_buffer2 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPPB2

//==============================================================================
//  Event: Main Memory Page Prog Through Buffer 1 (No Erase)
//==============================================================================

always @(MMPPB1NE)
begin : MMPPB1NE_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer
    reg program_ok;		// Flag to indicate program success (1) or failure (0)

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1745}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable MMPPB1NE_;
    end

    get_address( addr, valid );
    if (!valid) disable MMPPB1NE_;
    parse_address( addr, page, page_start, page_end, inpage_addr );

    if (BADADDR_OPT  &&  (inpage_addr > page_size - 1))
    begin
	$display( "ERROR: AT45DQXXX{#1747}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 1 and page %04h",
	    TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1, page );
	corrupt_buffer( MEM_B1 );
	inpage_addr = 0;
	bytecount = page_size;
	// Corrupted buffer will be written to main memory
	@( posedge CSB );
    end
    else write_buffer( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1, bytecount );
    // write_buffer ends when CSB goes high
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1750}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1NE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1755}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) is not allowed during program suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1NE_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1760}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1NE_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1764}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1765}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1NE_;
    end

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, write  page = %04h, sector = %04h", TIMESCALE*$time, page, inprog_sector(page) );
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, inprog page = %04h, sector = %04h", TIMESCALE*$time, inprog_ers_page, inprog_sector(inprog_ers_page) );

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1770}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1NE_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1775}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable MMPPB1NE_;
    end

    // Allow partial page programming to an unerased page (no erase check)
    updating_memory = 1'b1;
    inuse_buffer1 = 1'b1;

    write_to_memory_partial( MEM_B1, page, inpage_addr, bytecount, valid );
    if (valid)
	program_ok = 1'b1;
    else
    begin
	$display( "ERROR: AT45DQXXX{#1777}, %m (@ %.3fms), Opcode %02h (MMPPB1NE) failed, EPE set, some byte locations were not in erased state.", TIMESCALE*$time, opcode_reg );
	program_ok = 1'b0;
    end

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B1;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    // inprog_pgm_delay = ((tBP*bytecount < tP) ? tBP*bytecount : tP);
    inprog_pgm_delay = tBP*bytecount;
    inprog_pgm_epe = ~program_ok;	// EPE status bit for this program operation

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, delay = %.3fms", TIMESCALE*$time, TIMESCALE*((tBP*bytecount < tP) ? tBP*bytecount : tP) );
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, tP = %.3fms", TIMESCALE*$time, TIMESCALE*tP );
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, tBP*bytecount = %.3fms", TIMESCALE*$time, TIMESCALE*tBP*bytecount );
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, bytecount = %3d", TIMESCALE*$time, bytecount );
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) Main Memory Page Prog Through Buffer 1, (tBP*bytecount < tP) = %1d", TIMESCALE*$time, (tBP*bytecount < tP) );

    // Lesser of byte programming time time bytes or one page program time
    // #((tBP*bytecount < tP) ? tBP*bytecount : tP)

    #(tBP*bytecount)
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    if (program_ok)
	status2_reg[ST_EPE] = 1'b0;	// Reset EPE status bit (no error)
    else
	status2_reg[ST_EPE] = 1'b1;	// Set EPE status bit (program verify error)

    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_memory = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // MMPPB1NE

//==============================================================================
//  Event: Auto Page Rewrite (or Read-Modify-Write) Through Buffer 1
//==============================================================================

always @(APRB1)
begin : APRB1_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer
    reg [7:0] rxdata;		// Temp register to test for additional data bytes
    reg bad_addr;		// 1 if out-of-range address supplied

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1780}, %m (@ %.3fms), Opcode %02h (APRB1) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable APRB1_;
    end

    get_address( addr, valid );
    if (!valid) disable APRB1_;
    parse_address( addr, page, page_start, page_end, inpage_addr );
    bad_addr = (inpage_addr > page_size - 1);

    if (SUPPORT_RMW)
    begin
	write_buffer( S_SNGL, MEM_B1, inpage_addr, 0, page_size-1, bytecount );
	// write_buffer ends when CSB goes high
    end
    else
    begin
	// Attempt to read one byte from the serial stream.  If there are no
	// further bits clocked in the serial_in task will end at posedge CSB
	// with rxdata==XX and valid==0.
	serial_in( S_SNGL, rxdata, valid );
	if (valid  ||  rxdata !== 8'hXX)
	    $display( "ERROR: AT45DQXXX{#1782}, %m (@ %.3fms), Opcode %02h (APRB1) does not support Read-Modify-Write in device %s, extra serial input ignored.", TIMESCALE*$time, opcode_reg, DEVICE );
	// If multiple data bytes are presented on the serial input stream,
	// wait for CSB to be de-asserted.
	if (CSB === 1'b0) @( posedge CSB );
	bytecount = 0;
    end

    if (SUPPORT_RMW  &&  BADADDR_OPT  &&  bad_addr  &&  (bytecount > 0))
    begin
	$display( "ERROR: AT45DQXXX{#1783}, %m (@ %.3fms), Opcode %02h (APRB1) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 1 and page %04h",
	    TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1, page );
	corrupt_buffer( MEM_B1 );
	inpage_addr = 0;
	bytecount = page_size;
	// Corrupted buffer will be written to main memory
    end

    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1785}, %m (@ %.3fms), Opcode %02h (APRB1) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB1_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1790}, %m (@ %.3fms), Opcode %02h (APRB1) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB1_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1795}, %m (@ %.3fms), Opcode %02h (APRB1) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB1_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1799}, %m (@ %.3fms), Opcode %02h (APRB1) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1800}, %m (@ %.3fms), Opcode %02h (APRB1) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB1_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1805}, %m (@ %.3fms), Opcode %02h (APRB1) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB1_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1810}, %m (@ %.3fms), Opcode %02h (APRB1) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB1_;
    end

    updating_buffer1 = 1'b1;
    updating_memory = 1'b1;
    inuse_buffer1 = 1'b1;

    if (SUPPORT_RMW  &&  bytecount > 0)
    begin
	// Transfer remaining bytes (those that were not updated) from the main
	// memory page to the buffer, then write back to the main memory page
	transfer_to_buffer_partial( MEM_B1, page, inpage_addr + bytecount, page_size - bytecount );
	erase_page( page );		// Erase page, set contents to FF
	write_to_memory( MEM_B1, page, page_start );
    end
    else
    begin
	// We don't need to rewrite the data back into main-memory, as the
	// data is already available in the main-memory
	transfer_to_buffer( MEM_B1, page_start );
    end

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B1;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = tEP;		// Time needed to complete operation
    inprog_pgm_epe = 1'b0;		// EPE status bit for this program operation

    #tEP
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    updating_buffer1 = 1'b0;
    updating_memory = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // APRB1

//==============================================================================
//  Event: Auto Page Rewrite (or Read-Modify-Write) Through Buffer 2
//==============================================================================

always @(APRB2)
begin : APRB2_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer bytecount;		// Number of bytes written to buffer
    reg [7:0] rxdata;		// Temp register to test for additional data bytes
    reg bad_addr;		// 1 if out-of-range address supplied

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1815}, %m (@ %.3fms), Opcode %02h (APRB2) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable APRB2_;
    end

    get_address( addr, valid );
    if (!valid) disable APRB2_;
    parse_address( addr, page, page_start, page_end, inpage_addr );
    bad_addr = (inpage_addr > page_size - 1);

    if (SUPPORT_RMW)
    begin
	write_buffer( S_SNGL, MEM_B2, inpage_addr, 0, page_size-1, bytecount );
	// write_buffer ends when CSB goes high
    end
    else
    begin
	// Attempt to read one byte from the serial stream.  If there are no
	// further bits clocked in the serial_in task will end at posedge CSB
	// with rxdata==XX and valid==0.
	serial_in( S_SNGL, rxdata, valid );
	if (valid  ||  rxdata !== 8'hXX)
	    $display( "ERROR: AT45DQXXX{#1817}, %m (@ %.3fms), Opcode %02h (APRB2) does not support Read-Modify-Write in device %s, extra serial input ignored.", TIMESCALE*$time, opcode_reg, DEVICE );
	// If multiple data bytes are presented on the serial input stream,
	// wait for CSB to be de-asserted.
	if (CSB === 1'b0) @( posedge CSB );
	bytecount = 0;
    end

    if (SUPPORT_RMW  &&  BADADDR_OPT  &&  bad_addr  &&  (bytecount > 0))
    begin
	$display( "ERROR: AT45DQXXX{#1818}, %m (@ %.3fms), Opcode %02h (APRB2) byte address %03h outside DataFlash range [0 to %03h], corrupting buffer 2 and page %04h",
	    TIMESCALE*$time, opcode_reg, inpage_addr, page_size-1, page );
	corrupt_buffer( MEM_B2 );
	inpage_addr = 0;
	bytecount = page_size;
	// Corrupted buffer will be written to main memory
    end

    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (BUFFERS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#1820}, %m (@ %.3fms), Opcode %02h (APRB2) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1825}, %m (@ %.3fms), Opcode %02h (APRB2) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1830}, %m (@ %.3fms), Opcode %02h (APRB2) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end

    if (check_lockdown( page ) == 1)	// Check if sector is locked
    begin
	$display( "ERROR: AT45DQXXX{#1835}, %m (@ %.3fms), Opcode %02h (APRB2) aborted, page %04h is locked.", TIMESCALE*$time, opcode_reg, page );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end

    if (check_protect( page ) == 1)	// Check if sector is protected
    begin
	if (hw_protect_active == 1'b1)
	    $display( "ERROR: AT45DQXXX{#1839}, %m (@ %.3fms), Opcode %02h (APRB2) aborted, WPB asserted.", TIMESCALE*$time, opcode_reg );
	else
	begin
	    $display( "ERROR: AT45DQXXX{#1840}, %m (@ %.3fms), Opcode %02h (APRB2) aborted, page %04h is protected.", TIMESCALE*$time, opcode_reg, page );
	    #tABORT;
	end
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1845}, %m (@ %.3fms), Opcode %02h (APRB2) aborted, writing to a program suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
    begin
	$display( "ERROR: AT45DQXXX{#1850}, %m (@ %.3fms), Opcode %02h (APRB2) aborted, writing to an erase suspended sector.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable APRB2_;
    end

    updating_buffer2 = 1'b1;
    updating_memory = 1'b1;
    inuse_buffer2 = 1'b1;

    if (SUPPORT_RMW  &&  bytecount > 0)
    begin
	// Transfer remaining bytes (those that were not updated) from the main
	// memory page to the buffer, then write back to the main memory page
	transfer_to_buffer_partial( MEM_B2, page, inpage_addr + bytecount, page_size - bytecount );
	erase_page( page );		// Erase page, set contents to FF
	write_to_memory( MEM_B2, page, page_start );
    end
    else
    begin
	// We don't need to rewrite the data back into main-memory, as the
	// data is already available in the main-memory
	transfer_to_buffer( MEM_B2, page_start );
    end

    allow_pgm_suspend = 1'b1;		// This operation may be suspended
    inprog_pgm_buffer = MEM_B2;		// Source buffer for program operation
    inprog_pgm_page = page;		// Memory page being programmed
    inprog_pgm_start = $time;		// Start time of program operation
    inprog_pgm_delay = tEP;		// Time needed to complete operation
    inprog_pgm_epe = 1'b0;		// EPE status bit for this program operation

    #tEP
    allow_pgm_suspend = 1'b0;		// Operation is done; nothing to suspend
    inprog_pgm_buffer = 0;		// No buffer being written to main memory

    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    updating_buffer2 = 1'b0;
    updating_memory = 1'b0;
    inuse_buffer2 = 1'b0;
    bg_opcode = 8'hXX;
end // APRB2

//==============================================================================
//  Event: Status Register Read
//==============================================================================

always @(SR)
begin : SR_
    reg valid;			// Local flag to test serial output

    if (skip == 1'b1)
	@( posedge SCK );	// Skip one SCK

    while (CSB === 1'b0)
    begin
	serial_out( S_SNGL, status1_reg, valid );
	if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) read out status1_reg = %02h (%08b)", TIMESCALE*$time, status1_reg, status1_reg );
	if(CSB === 1'b0)
	begin
	    serial_out( S_SNGL, status2_reg, valid );
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) read out status2_reg = %02h (%08b)", TIMESCALE*$time, status2_reg, status2_reg );
	end
    end
end // SR

//==============================================================================
//  Event: Configuration Register Read
//==============================================================================

always @(CSR)
begin : CSR_
    reg valid;			// Local flag to test serial output
    reg [7:0] data_out;		// Temporary registor for serial output

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1855}, %m (@ %.3fms), Opcode %02h (CSR) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable CSR_;
    end

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#1857}, %m (@ %.3fms), Opcode %02h (CSR) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable CSR_;
    end

    if (skip == 1'b1)
	@( posedge SCK );	// Skip one SCK

    if (clk_valid == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1860}, %m (@ %.3fms), Opcode %02h (CSR) serial clock is too fast.", TIMESCALE*$time, opcode_reg );
	// Output one byte of unknown if clock is too fast
	data_out = 8'hXX;
	serial_out( S_SNGL, data_out, valid );
	if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) read out config_reg = %02h (%08b)", TIMESCALE*$time, data_out, data_out );
    end
    while (CSB === 1'b0)
    begin
	serial_out( S_SNGL, config_reg, valid );
	if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) read out config_reg = %02h (%08b)", TIMESCALE*$time, config_reg, config_reg );
    end
end // CSR

//==============================================================================
//  Event: Manufacturing ID Read
//==============================================================================

always @(MIR)
begin : MIR_
    reg valid;			// Local flag to test serial output
    reg [7:0] data_out;		// Temporary registor for serial output
    integer i;

    if (skip == 1'b1)
	@( posedge SCK );	// Skip one SCK

    for (i=5; i>0; i=i-1)
    begin
	case (i)
	    5 : data_out = MAN_ID[39:32];
	    4 : data_out = MAN_ID[31:24];
	    3 : data_out = MAN_ID[23:16];
	    2 : data_out = MAN_ID[15:8];
	    1 : data_out = MAN_ID[7:0];
	endcase
	if(CSB === 1'b0)
	begin
	    serial_out( S_SNGL, data_out, valid );
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) read out Manufacturing_ID[%1d] = %02h", TIMESCALE*$time, i, data_out );
	end
    end

    // Take SO output to high-Z on the next clock.  The positive edge of CSB will
    // shutdown the MIR event if the next clock never comes.
    @( negedge SCK or posedge CSB )
    highz_outputs;
end // MIR

//==============================================================================
//  Event: 4-byte Command starting with 9B
//==============================================================================

always @(OP9B)
begin : OP9B_
    reg valid;			// Local flag to test serial input
    reg [7:0] byte1;		// Register to store first byte of opcode
    reg [7:0] byte2;		// Register to store second byte of opcode
    reg [7:0] byte3;		// Register to store third byte of opcode
    reg [7:0] byte4;		// Register to store fourth byte of opcode
    reg [31:0] long_opcode;	// Register to store full 4-byte opcode

    byte1 = 8'h9B;				// Set 1st byte of 4-byte opcode

    serial_in( S_SNGL, byte2, valid );		// Get 2nd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte3, valid );	// Get 3rd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte4, valid );	// Get 4th byte of 4-byte opcode

    if (!valid)
    begin
	$display( "ERROR: AT45DQXXX{#1865}, %m (@ %.3fms), Opcode %02h (OP9B) aborted.", TIMESCALE*$time, opcode_reg );
	disable OP9B_;
    end

    long_opcode[31:24] = byte1;
    long_opcode[23:16] = byte2;
    long_opcode[15:8]  = byte3;
    long_opcode[7:0]   = byte4;

    case (long_opcode)
	32'h9B_00_00_00 : -> SRP;		// Security Register Program
	default :
	    begin
		$display( "ERROR: AT45DQXXX{#1870}, %m (@ %.3fms), Unrecognized 4-byte opcode %02h_%02h_%02h_%02h", TIMESCALE*$time, byte1, byte2, byte3, byte4 );
		disable   OP9B_;
	    end
    endcase
end // OP9B

//==============================================================================
//  Event: Security Register Program
//==============================================================================

always @(SRP)
begin : SRP_
    integer addr;		// 24-bit address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer bytecount;		// Number of bytes written to buffer
    integer i;

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1875}, %m (@ %.3fms), Opcode %02h (SRP) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable SRP_;
    end

    addr = 0;
    page_start = 0;
    page_end = 63;
    write_buffer( S_SNGL, MEM_B1, addr, page_start, page_end, bytecount );
    // write_buffer ends when CSB goes high
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1880}, %m (@ %.3fms), Opcode %02h (SRP) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable SRP_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1885}, %m (@ %.3fms), Opcode %02h (SRP) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable SRP_;
    end

    inuse_buffer1 = 1'b1;
    updating_sec_reg = 1'b1;
    if (security_flag == 0)		// Security Register may only be programmed once
    begin
	for (i=0; i < 64; i = i+1)
	    security_reg[i] = buffer1[i];
	security_flag = 1'b1;		// Update security_flag;
	#tOTPP
	status2_reg[ST_EPE] = 1'b0;	// Reset EPE status bit (no error)
	status1_reg[ST_CMP] = 1'b0;	// Reset compare status bit
	status1_reg[ST_RDY] = 1'b1;	// Device is Ready
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Security register programmed", TIMESCALE*$time );
    end
    else
    begin
	#tBP
	status1_reg[ST_RDY] = 1'b1;	// Device is Ready
	$display( "ERROR: AT45DQXXX{#1887}, %m (@ %.3fms), Opcode %02h (SRP) aborted, may program Security OTP Register only once.", TIMESCALE*$time, opcode_reg );
    end
    updating_sec_reg = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // SRP

//==============================================================================
//  Event: Security Register Read
//==============================================================================

always @(SRR)
begin : SRR_
    reg valid;				// Local flag to test serial input

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1890}, %m (@ %.3fms), Opcode %02h (SRR) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable SRR_;
    end

    get_dummy_bytes( 3, valid );	// Skip three dummy bytes
    if (!valid) disable SRR_;

    if (skip == 1'b1) @( posedge SCK );	// Skip one SCK

    read_out_reg( R_SEC, 0, 63 );
end // SRR

//==============================================================================
//  Event: 4-byte command starting with 3D
//==============================================================================

always @(OP3D)
begin : OP3D_
    reg valid;			// Local flag to test serial input
    reg [7:0] byte1;		// Register to store first byte of opcode
    reg [7:0] byte2;		// Register to store second byte of opcode
    reg [7:0] byte3;		// Register to store third byte of opcode
    reg [7:0] byte4;		// Register to store fourth byte of opcode
    reg [31:0] long_opcode;	// Register to store full 4-byte opcode

    byte1 = 8'h3D;				// Set 1st byte of 4-byte opcode

    serial_in( S_SNGL, byte2, valid );		// Get 2nd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte3, valid );	// Get 3rd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte4, valid );	// Get 4th byte of 4-byte opcode

    if (!valid)
    begin
	$display( "ERROR: AT45DQXXX{#1895}, %m (@ %.3fms), Opcode %02h (OP3D) aborted.", TIMESCALE*$time, opcode_reg );
	disable OP3D_;
    end

    long_opcode[31:24] = byte1;
    long_opcode[23:16] = byte2;
    long_opcode[15:8]  = byte3;
    long_opcode[7:0]   = byte4;

    case (long_opcode)
	32'h3D_2A_7F_A9 : -> SPE;		// Sector Protection Enable
	32'h3D_2A_7F_9A : -> SPD;		// Sector Protection Disable
	32'h3D_2A_7F_CF : -> PRE;		// Protection Register Erase
	32'h3D_2A_7F_FC : -> PRP;		// Protection Register Program
	32'h3D_2A_7F_30 : -> LRP;		// Lock-down Register Program
	32'h3D_2A_80_A6 : -> BPS;		// Binary page size command
	32'h3D_2A_80_A7 : -> DPS;		// DataFlash page size command
	32'h3D_2A_81_66 : -> QEN;		// Quad mode enable
	32'h3D_2A_81_67 : -> QDIS;		// Quad mode disable
	default :
	begin
	    $display("ERROR: AT45DQXXX{#1900}, %m (@ %.3fms), Unrecognized 4-byte opcode %02h_%02h_%02h_%02h", TIMESCALE*$time, byte1, byte2, byte3, byte4 );
	    disable   OP3D_;
	end
    endcase
end // OP3D

//==============================================================================
//  Event: Sector Protection Enable
//==============================================================================

always @(SPE)
begin : SPE_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1905}, %m (@ %.3fms), Opcode %02h (SPE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable SPE_;
    end

    @( posedge CSB );

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1910}, %m (@ %.3fms), Opcode %02h (SPE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	disable SPE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1915}, %m (@ %.3fms), Opcode %02h (SPE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	disable SPE_;
    end

    status1_reg[ST_PRT] = 1'b1;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Sector Protection Enabled", TIMESCALE*$time );
end // SPE

//==============================================================================
//  Event: Sector Protection Disable
//==============================================================================

always @(SPD)
begin : SPD_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1920}, %m (@ %.3fms), Opcode %02h (SPD) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable SPD_;
    end

    @( posedge CSB );

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1925}, %m (@ %.3fms), Opcode %02h (SPD) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	disable SPD_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1930}, %m (@ %.3fms), Opcode %02h (SPD) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	disable SPD_;
    end

    // Added the WPB===1'bz case on account of the test bench, which left WPB floating for this op-code.
    // WPB has an internal pull-up, which is not modeled (see module ports list).
    if (WPB === 1'b1 || WPB === 1'bz || QE_reg === 1'b1)
    begin
	status1_reg[ST_PRT] = 1'b0;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Sector Protection Disabled", TIMESCALE*$time );
    end
    else
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) WPB is asserted, Sector Protection not disabled", TIMESCALE*$time );
end // SPD

//==============================================================================
//  Event: Protection Register Erase
//==============================================================================

always @(PRE)
begin : PRE_
    integer i;

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1935}, %m (@ %.3fms), Opcode %02h (PRE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable PRE_;
    end

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1940}, %m (@ %.3fms), Opcode %02h (PRE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PRE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1945}, %m (@ %.3fms), Opcode %02h (PRE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PRE_;
    end

    updating_prot_reg = 1'b1;
    prot_reg[0] = 8'hFF;
    prot_status[0] = 1;
    prot_status[1] = 1;
    for (i=1; i<SECTORS; i=i+1)
    begin
	prot_reg[i] = 8'hFF;
	prot_status[i+1] = 1;
    end
    #tPE
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Protection Register Erased", TIMESCALE*$time );
    updating_prot_reg = 1'b0;
    bg_opcode = 8'hXX;
end // PRE

//==============================================================================
//  Event: Protection Register Program
//==============================================================================

always @(PRP)
begin : PRP_
    integer addr;		// 24-bit address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer bytecount;		// Number of bytes written to buffer
    integer i;
    reg [7:0] temp_reg;		// Temporary store
    reg program_ok;		// Flag to indicate program success (1) or failure (0)

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1950}, %m (@ %.3fms), Opcode %02h (PRP) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable PRP_;
    end

    addr = 0;
    page_start = 0;
    page_end = SECTORS-1;
    write_buffer( S_SNGL, MEM_B1, addr, page_start, page_end, bytecount );
    // write_buffer ends when CSB goes high
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1955}, %m (@ %.3fms), Opcode %02h (PRP) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PRP_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1960}, %m (@ %.3fms), Opcode %02h (PRP) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable PRP_;
    end

    inuse_buffer1 = 1'b1;
    updating_prot_reg = 1'b1;
    // Program prot_reg
    program_ok = 1'b1;
    for (i=0; i < SECTORS; i = i+1)
    begin
	prot_reg[i] = (prot_reg[i] & buffer1[i]);
	if (prot_reg[i] !== buffer1[i]) program_ok = 1'b0;
    end
    if (program_ok == 1'b0)
	$display( "ERROR: AT45DQXXX{#1962}, %m (@ %.3fms), Opcode %02h (PRP) Protection Register programming failed, some altered bytes were not erased.", TIMESCALE*$time, opcode_reg );
    // Update prot_status;
    temp_reg[7:0] = prot_reg[0];
    prot_status[0] = &(temp_reg[7:6]);
    prot_status[1] = &(temp_reg[5:4]);
    for (i=1; i< SECTORS; i= i+1)
	prot_status[i+1] = &(prot_reg[i]);
    #tP
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    status1_reg[ST_CMP] = 1'b0;		// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Protection Register Programmed", TIMESCALE*$time );
    updating_prot_reg = 1'b0;
    inuse_buffer1 = 1'b0;
    bg_opcode = 8'hXX;
end // PRP

//==============================================================================
//  Event: Lock Down Register Program
//==============================================================================

always @(LRP)
begin : LRP_
    reg valid;			// Local flag to test serial input
    integer addr;		// 24-bit address
    integer page;		// Page component of address
    integer page_start;		// Start index of addressed page in array
    integer page_end;		// End index of addressed page in array
    integer inpage_addr;	// Address offset within page
    integer sector;		// Sector number derived from address

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1965}, %m (@ %.3fms), Opcode %02h (LRP) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable LRP_;
    end

    get_address( addr, valid );
    if (!valid) disable LRP_;
    parse_address( addr, page, page_start, page_end, inpage_addr );
    sector = page / PAGES_PER_SECTOR;

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;		// Device is Busy
    bg_opcode = opcode_reg;

    if (status2_reg[ST_SLE] == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1970}, %m (@ %.3fms), Opcode %02h (LRP) disabled, lockdown state is frozen.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable LRP_;
    end

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#1975}, %m (@ %.3fms), Opcode %02h (LRP) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable LRP_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1980}, %m (@ %.3fms), Opcode %02h (LRP) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable LRP_;
    end

    updating_lock_reg = 1'b1;
    if (page < 8)
    begin
	case (lock_status[1])
	    1'b0 : lock_reg[0] = 8'hC0; // Sector 0a locked, 0b unlocked
	    1'b1 : lock_reg[0] = 8'hF0; // Sectors 0a, 0b locked
	    default : $display( "ERROR: AT45DQXXX{#1985}, %m (@ %.3fms), Opcode %02h (LRP) unexpected lock_status[1] = %1b.", TIMESCALE*$time, opcode_reg, lock_status[1] );
	endcase
	lock_status[0] = 1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Lockdown Register Program: sector 0a locked", TIMESCALE*$time );
    end
    else if (sector == 0)
    begin
	case (lock_status[0])
	    1'b0 : lock_reg[0] = 8'h30; // Sector 0a unlocked, 0b locked
	    1'b1 : lock_reg[0] = 8'hF0; // Sectors 0a, 0b locked
	    default : $display( "ERROR: AT45DQXXX{#1990}, %m (@ %.3fms), Opcode %02h (LRP) unexpected lock_status[0] = %1b.", TIMESCALE*$time, opcode_reg, lock_status[0] );
	endcase
	lock_status[1] = 1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Lockdown Register Program: sector 0b locked", TIMESCALE*$time );
    end
    else
    begin
	lock_reg[sector] = 8'hFF;
	lock_status[sector+1] = 1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Lockdown Register Program: sector %04h locked", TIMESCALE*$time, sector );
    end
    #tP
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_lock_reg = 1'b0;
    bg_opcode = 8'hXX;
end // LRP

//==============================================================================
//  Event: Binary Page Size Program
//==============================================================================

always @(BPS)
begin : BPS_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#1995}, %m (@ %.3fms), Opcode %02h (BPS) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable BPS_;
    end

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2000}, %m (@ %.3fms), Opcode %02h (BPS) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BPS_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2005}, %m (@ %.3fms), Opcode %02h (BPS) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BPS_;
    end

    if (status1_reg[ST_BIN] === 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2007}, %m (@ %.3fms), Opcode %02h (BPS) ignored, Binary page size already selected.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable BPS_;
    end

    updating_pagesize = 1'b1;
    #tEP
    status1_reg[ST_BIN] = 1'b1;
    binary_page = 1'b1;
    page_size = DF_PAGESIZE - DF_EXTRA_BYTES_PER_PAGE;
    memory_top = DF_PAGESIZE*(PAGES - 1) + page_size - 1;
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_pagesize = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Binary Page Size set, page_size = %5d", TIMESCALE*$time, page_size );
end // BPS

//==============================================================================
//  Event: DataFlash Page Size Program
//==============================================================================

always @(DPS)
begin : DPS_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2010}, %m (@ %.3fms), Opcode %02h (DPS) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable DPS_;
    end

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2015}, %m (@ %.3fms), Opcode %02h (DPS) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable DPS_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2020}, %m (@ %.3fms), Opcode %02h (DPS) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable DPS_;
    end

    if (status1_reg[ST_BIN] === 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2022}, %m (@ %.3fms), Opcode %02h (DPS) ignored, DataFlash page size already selected.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable DPS_;
    end

    updating_pagesize = 1'b1;
    #tEP
    status1_reg[ST_BIN] = 1'b0;
    binary_page = 1'b0;
    page_size = DF_PAGESIZE;
    memory_top = DF_PAGESIZE*(PAGES - 1) + page_size - 1;
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_pagesize = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) DataFlash Page Size set, page_size = %5d", TIMESCALE*$time, page_size );
end // DPS

//==============================================================================
//  Event: Quad Mode Enable
//==============================================================================

always @(QEN)
begin : QEN_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2025}, %m (@ %.3fms), Opcode %02h (QEN) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable QEN_;
    end

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#2027}, %m (@ %.3fms), Opcode %02h (QEN) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable QEN_;
    end

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Prior to CSB rising edge, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );
    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Post CSB rising edge, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2030}, %m (@ %.3fms), Opcode %02h (QEN) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QEN_;
    end

    if (QUADEN_DIS == 1)
    begin
	$display( "ERROR: AT45DQXXX{#2035}, %m (@ %.3fms), Opcode %02h (QEN) is not allowed when model option QUADEN_DIS is set.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QEN_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2040}, %m (@ %.3fms), Opcode %02h (QEN) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QEN_;
    end

    if (QE_reg === 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2042}, %m (@ %.3fms), Opcode %02h (QEN) ignored, Quad-bit mode is already enabled.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QEN_;
    end

    updating_cfg_reg = 1'b1;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Delay before QE bit update, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );
    #tWRCR
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Do QE bit update, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );
    config_reg[7] = 1'b1;		// Set quad-mode enable bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_cfg_reg = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Quad-bit read/write mode enabled", TIMESCALE*$time );
end // QEN

//==============================================================================
//  Event: Quad Mode Disable
//==============================================================================

always @(QDIS)
begin : QDIS_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2045}, %m (@ %.3fms), Opcode %02h (QDIS) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable QDIS_;
    end

    if (SUPPORT_DQBIT == 0)
    begin
	$display( "ERROR: AT45DQXXX{#2047}, %m (@ %.3fms), Opcode %02h (QDIS) is not offered in device %s.", TIMESCALE*$time, opcode_reg, DEVICE );
	disable QDIS_;
    end

    if (WPB === 1'b0  &&  QE_reg === 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2048}, %m (@ %.3fms), Opcode %02h (QDIS), disabling Quad-bit mode while WPB is asserted, hardware sector protection enabled (inadvertantly?).", TIMESCALE*$time, opcode_reg );
	hw_protect_pending = 1'b1;
    end

    if (RESETB === 1'b0)
	$display( "ERROR: AT45DQXXX{#2049}, %m (@ %.3fms), Opcode %02h (QDIS), disabling Quad-bit mode while RESETB is asserted will result in device reset.", TIMESCALE*$time, opcode_reg );

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Prior to CSB rising edge, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );
    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Post CSB rising edge, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2050}, %m (@ %.3fms), Opcode %02h (QDIS) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QDIS_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2055}, %m (@ %.3fms), Opcode %02h (QDIS) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QDIS_;
    end

    if (QE_reg === 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2057}, %m (@ %.3fms), Opcode %02h (QDIS) ignored, Quad-bit mode is already disabled.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable QDIS_;
    end

    updating_cfg_reg = 1'b1;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Delay before QE bit update, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );

    if (hw_protect_pending == 1'b1)
    begin
	#tWPE
	if (WPB === 1'b0)
	begin
	    hw_protect_active = 1'b1;
	    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) WPB asserted", TIMESCALE*$time );
	end
	hw_protect_pending = 1'b0;
	#(tWRCR - tWPE);
    end
    else #tWRCR;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) opcode = %02h, Do QE bit update, ready bit is %1d", TIMESCALE*$time, opcode_reg, status1_reg[ST_RDY] );
    config_reg[7] = 1'b0;		// Reset quad-mode enable bit
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_cfg_reg = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Quad-bit read/write mode disabled", TIMESCALE*$time );
end // QDIS

//==============================================================================
//  Event: Protection Register Read
//==============================================================================

always @(PRR)
begin : PRR_
    reg valid;			// Local flag to test serial input

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2060}, %m (@ %.3fms), Opcode %02h (PRR) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable PRR_;
    end

    get_dummy_bytes( 3, valid );	// Skip three dummy bytes
    if (!valid) disable PRR_;

    if (skip == 1'b1)
	@( posedge SCK ); // skip one SCK
    read_out_reg( R_PRT, 0, SECTORS-1 );
end // PRR

//==============================================================================
//  Event: Lock-down Register Read
//==============================================================================

always @(LRR)
begin : LRR_
    reg valid;			// Local flag to test serial input

    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2065}, %m (@ %.3fms), Opcode %02h (LRR) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable LRR_;
    end

    get_dummy_bytes( 3, valid );	// Skip three dummy bytes
    if (!valid) disable LRR_;

    if (skip == 1'b1)
	@( posedge SCK ); // skip one SCK
    read_out_reg( R_LCK, 0, SECTORS-1 );
end // LRR

//==============================================================================
//  Event: 4-byte command starting with 34
//==============================================================================

always @(OP34)
begin : OP34_
    reg valid;			// Local flag to test serial input
    reg [7:0] byte1;		// Register to store first byte of opcode
    reg [7:0] byte2;		// Register to store second byte of opcode
    reg [7:0] byte3;		// Register to store third byte of opcode
    reg [7:0] byte4;		// Register to store fourth byte of opcode
    reg [31:0] long_opcode;	// Register to store full 4-byte opcode

    byte1 = 8'h34;				// Set 1st byte of 4-byte opcode

    serial_in( S_SNGL, byte2, valid );		// Get 2nd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte3, valid );	// Get 3rd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte4, valid );	// Get 4th byte of 4-byte opcode

    if (!valid)
    begin
	$display( "ERROR: AT45DQXXX{#2070}, %m (@ %.3fms), Opcode %02h (OP34) aborted.", TIMESCALE*$time, opcode_reg );
	disable OP34_;
    end

    long_opcode[31:24] = byte1;
    long_opcode[23:16] = byte2;
    long_opcode[15:8]  = byte3;
    long_opcode[7:0]   = byte4;

    case (long_opcode)
	32'h34_55_AA_40 : -> FREEZE;		// Freeze Sector Lockdown
	default :
	begin
	    $display("ERROR: AT45DQXXX{#2075}, %m (@ %.3fms), Unrecognized 4-byte opcode %02h_%02h_%02h_%02h", TIMESCALE*$time, byte1, byte2, byte3, byte4 );
	    disable   OP34_;
	end
    endcase
end // OP34

//==============================================================================
//  Event: Freeze Sector Lockdown
//==============================================================================

always @(FREEZE)
begin : FREEZE_
    if (status1_reg[ST_RDY] !== 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2080}, %m (@ %.3fms), Opcode %02h (FREEZE) is not allowed while device is busy.", TIMESCALE*$time, opcode_reg );
	disable FREEZE_;
    end

    @( posedge CSB );
    status1_reg[ST_RDY] = 1'b0;	// Device is Busy
    bg_opcode = opcode_reg;

    if (background_op_enable == 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2085}, %m (@ %.3fms), Opcode %02h (FREEZE) is not allowed before time %.3fms.", TIMESCALE*$time, opcode_reg, TIMESCALE*tPUW );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable FREEZE_;
    end

    if (status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_ES] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2090}, %m (@ %.3fms), Opcode %02h (FREEZE) is not allowed during program or erase suspend.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable FREEZE_;
    end

    if (status2_reg[ST_SLE] === 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2092}, %m (@ %.3fms), Opcode %02h (FREEZE) ignored, sector lockdown is already disabled.", TIMESCALE*$time, opcode_reg );
	#tABORT
	status1_reg[ST_RDY] = 1'b1;
	bg_opcode = 8'hXX;
	disable FREEZE_;
    end

    updating_freeze = 1'b1;
    #tLOCK
    status2_reg[ST_SLE] = 1'b0;		// Sector lockdown command disabled
    status2_reg[ST_EPE] = 1'b0;		// Reset EPE status bit (no error)
    status1_reg[ST_RDY] = 1'b1;		// Device is Ready
    updating_freeze = 1'b0;
    bg_opcode = 8'hXX;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Lockdown state is frozen", TIMESCALE*$time );
end // FREEZE

//==============================================================================
//  Event: 4-byte command starting with F0
//==============================================================================

always @(OPF0)
begin : OPF0_
    reg valid;			// Local flag to test serial input
    reg [7:0] byte1;		// Register to store first byte of opcode
    reg [7:0] byte2;		// Register to store second byte of opcode
    reg [7:0] byte3;		// Register to store third byte of opcode
    reg [7:0] byte4;		// Register to store fourth byte of opcode
    reg [31:0] long_opcode;	// Register to store full 4-byte opcode

    byte1 = 8'hF0;				// Set 1st byte of 4-byte opcode

    serial_in( S_SNGL, byte2, valid );		// Get 2nd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte3, valid );	// Get 3rd byte of 4-byte opcode

    if (valid)
        serial_in( S_SNGL, byte4, valid );	// Get 4th byte of 4-byte opcode

    if (!valid)
    begin
	$display( "ERROR: AT45DQXXX{#2095}, %m (@ %.3fms), Opcode %02h (OPF0) aborted.", TIMESCALE*$time, opcode_reg );
	disable OPF0_;
    end

    long_opcode[31:24] = byte1;
    long_opcode[23:16] = byte2;
    long_opcode[15:8]  = byte3;
    long_opcode[7:0]   = byte4;

    case (long_opcode)
	32'hF0_00_00_00 : -> SOFTRESET;		// Software Reset
	default :
	begin
	    $display("ERROR: AT45DQXXX{#2100}, %m (@ %.3fms), Unrecognized 4-byte opcode %02h_%02h_%02h_%02h", TIMESCALE*$time, byte1, byte2, byte3, byte4 );
	    disable   OPF0_;
	end
    endcase
end // OPF0

//==============================================================================
//  Event: Software Reset
//==============================================================================

always @(SOFTRESET)
begin : SOFTRESET_
    @( posedge CSB );

    if (status1_reg[ST_RDY] !== 1'b0  &&  status2_reg[ST_PS2] === 1'b0  &&  status2_reg[ST_PS1] === 1'b0  &&  status2_reg[ST_ES] === 1'b0)
    begin
	$display( "ERROR: AT45DQXXX{#2102}, %m (@ %.3fms), Opcode %02h (SOFTRESET) ignored, device is not busy or suspended.", TIMESCALE*$time, opcode_reg );
	disable SOFTRESET_;
    end

    if (updating_sec_reg == 1'b1)
    begin
	$display("ERROR: AT45DQXXX{#2105}, %m (@ %.3fms), Software Reset is not allowed during Security Register (OTP) programming", TIMESCALE*$time );
	disable   SOFTRESET_;
    end

    reset_active = 1'b1;
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Software Reset", TIMESCALE*$time );
    reset_chip;
    #tSWRST
    status1_reg[ST_RDY] = 1'b1;
    highz_outputs;		// All outputs are High-impedance
    reset_active = 1'b0;
end // SOFTRESET

//==============================================================================
//  Posedge CSB. Stop all reading, receiving commands/addresses etc.
//==============================================================================

always @( posedge CSB )
begin : POSEDGE_CSB_
    if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) CSB positive edge - end of foreground operation", TIMESCALE*$time );

    // The event handlers below will end themselves when CSB goes high

    // disable MMPR_; // MMPR will stop, if CSB goes high
    // disable MMCAR_; // MMCAR will stop, if CSB goes high
    // disable DORA_; // DORA will stop, if CSB goes high
    // disable QORA_; // QORA will stop, if CSB goes high

    // disable B1R_; // B1R will stop, if CSB goes high
    // disable B2R_; // B2R will stop, if CSB goes high
    // disable B1W_; // B1W will stop, if CSB goes high
    // disable B2W_; // B2W will stop, if CSB goes high
    // disable DB1W_; // DB1W will stop, if CSB goes high
    // disable DB2W_; // DB2W will stop, if CSB goes high
    // disable QB1W_; // QB1W will stop, if CSB goes high
    // disable QB2W_; // QB2W will stop, if CSB goes high

    // disable SR_; // Status reading should stop.
    // disable CSR_; // Configuration Register reading should stop.
    // disable MIR_; // MIR will stop, if CSB goes high

    // disable OP3D_; // 3D command sequence will stop, if CSB goes high
    // disable OP9B_; // 9B command sequence will stop, if CSB goes high
    // disable OPC7_; // C7 command sequence will stop, if CSB goes high
    // disable OP34_; // 34 command sequence will stop, if CSB goes high
    // disable OPF0_; // F0 command sequence will stop, if CSB goes high

    // disable PRR_; // PRR will stop, if CSB goes high
    // disable LRR_; // LRR will stop, if CSB goes high
    // disable SRR_; // SRR will stop, if CSB goes high

    // These tasks will end themselves when CSB goes high

    // disable serial_in;
    // disable serial_out;
    // disable read_out;
    // disable read_out_x;
    // disable cont_read_out;
    // disable cont_read_out_x;
    // disable write_buffer;

    highz_outputs;		// All outputs return to High-impedance state

    #1
    if (DEBUG_OPT5)
    begin
	$display( "DBG: %m (@ %.3fms) opcode = %02h, background opcode = %02h, RDY/BSY = %1d, PS1 = %1d, PS2 = %1d, ES = %1d",
	    TIMESCALE*$time, opcode_reg, bg_opcode, status1_reg[ST_RDY], status2_reg[ST_PS1], status2_reg[ST_PS2], status2_reg[ST_ES] );
	$display( "DBG: %m (@ %.3fms) allow_pgm_suspend = %1d, inprog_pgm_buffer = %1d, inprog_pgm_page = %04h, inprog_pgm_start = %.3fms, inprog_pgm_delay = %.3fms",
	    TIMESCALE*$time, allow_pgm_suspend, inprog_pgm_buffer, inprog_pgm_page, TIMESCALE*inprog_pgm_start, TIMESCALE*inprog_pgm_delay );
	$display( "DBG: %m (@ %.3fms) allow_ers_suspend = %1d, inprog_ers_scope = %1d, inprog_ers_page = %04h, inprog_ers_start = %.3fms, inprog_ers_delay = %.3fms",
	    TIMESCALE*$time, allow_ers_suspend, inprog_ers_scope, inprog_ers_page, TIMESCALE*inprog_ers_start, TIMESCALE*inprog_ers_delay );
	$display( "DBG: %m (@ %.3fms) updating_memory = %1d, erasing_page = %1d, erasing_block = %1d, erasing_sector = %1d, erasing_chip = %1d, suspending = %1d, resuming = %1d",
	    TIMESCALE*$time, updating_memory, erasing_page, erasing_block, erasing_sector, erasing_chip, suspending, resuming );
    end
end // POSEDGE_CSB

//==============================================================================
//  RESETB asserted
//==============================================================================

always @( RESETB )
begin
    if (RESETB === 1'b0 && QE_reg === 1'b0)
    begin
	reset_active = 1'b1;
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) RESETB asserted", TIMESCALE*$time );
	reset_chip;
	highz_outputs;		// All outputs are High-impedance
    end
    else if (RESETB === 1'b1 && reset_active == 1'b1)
    begin
	#tREC
	status1_reg[ST_RDY] = 1'b1;
	reset_active = 1'b0;	// Reset de-asserted after reset recovery time
    end
end // RESETB

//==============================================================================
//  WPB asserted - hardware write protect
//==============================================================================

always @( negedge WPB )
begin : WPB_ASSERT_
    if (WPB === 1'b0  &&  QE_reg === 1'b0)
    begin
	if (hw_protect_pending == 1'b1)
	begin
	    $display("ERROR: AT45DQXXX{#2107}, %m (@ %.3fms), Write protect re-asserted before tWPD (%.3fus).", TIMESCALE*$time, 1.0e-3*tWPD );
	    disable WPB_CANCEL_;
	end
	hw_protect_pending = 1'b1;
	#tWPE
	if (WPB === 1'b0)
	begin
	    hw_protect_active = 1'b1;
	    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) WPB asserted", TIMESCALE*$time );
	end
	hw_protect_pending = 1'b0;
    end
end // WPB_ASSERT

always @( posedge WPB )
begin : WPB_CANCEL_
    // Include floating WPB case.  WPB has a pull-up, that is not included in this model.
    if ((WPB === 1'b1  ||  WPB === 1'bz)  &&  QE_reg === 1'b0)
    begin
	if (hw_protect_pending == 1'b1)
	begin
	    $display("ERROR: AT45DQXXX{#2108}, %m (@ %.3fms), Write protect assertion cancelled before tWPE (%.3fus).", TIMESCALE*$time, 1.0e-3*tWPE );
	    disable WPB_ASSERT_;
	end
	if (hw_protect_active == 1'b1)
	begin
	    hw_protect_pending = 1'b1;
	    #tWPD
	    if (WPB === 1'b1  ||  WPB === 1'bz)
	    begin
		hw_protect_active = 1'b0;
		if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) WPB de-asserted", TIMESCALE*$time );
	    end
	    hw_protect_pending = 1'b0;
	end
    end
end // WPB_CANCEL

//==============================================================================
//  Posedge SCK, for clock period determination
//==============================================================================

always @( posedge SCK )
begin
    t_clk_period = $time - t_clk_prev_edge;
    t_clk_prev_edge = $time;
    sck_CAR1_valid = (t_clk_period >= tSCK_CAR1) ? 1'b1 : 1'b0;
    sck_CAR2_valid = (t_clk_period >= tSCK_CAR2) ? 1'b1 : 1'b0;
    sck_CAR3_valid = (t_clk_period >= tSCK_CAR3) ? 1'b1 : 1'b0;
    sck_SCK_valid  = (t_clk_period >= tSCK_SCK)  ? 1'b1 : 1'b0;
    sck_MAX_valid  = (t_clk_period >= tSCK_MAX)  ? 1'b1 : 1'b0;
end // posedge SCK

//==============================================================================
//  RDY/nBUSY status2_reg[7] mirrors status1_reg[7]
//==============================================================================

always @( status1_reg[ST_RDY] )
begin
    status2_reg[ST_RDY] = status1_reg[ST_RDY];
end

//##############################################################################
//  TASKS / FUNCTIONS
//##############################################################################

//==============================================================================
//  Task: serial_in (mode, >data, >valid)
//==============================================================================
//  Read byte of incoming data from the serial input stream
//------------------------------------------------------------------------------

task serial_in;

input	mode;	// 1: single wire input via SI; 2: dual-input; 4: quad-input
output	data;	// One byte of serial input data
output	valid;	// Successful read of one full byte if true

integer	mode;
reg [7:0] data;
reg	valid;

integer	i;	// Local counter for bit number

begin
    // S_SNGL=1, S_DUAL=2, S_QUAD=4
    if (!(mode == 1  ||  mode == 2  ||  mode == 4))
	$display( "ERROR: AT45DQXXX{#2110}, %m (@ %.3fms), Bad mode (%1d) for task serial_in.", TIMESCALE*$time, mode );

    valid = 1'b0;
    i = 8 - mode;
    data = 8'hXX;
    while (i >= 0  &&  CSB === 1'b0)
    begin : serial_in_LOOP
    	@( posedge SCK or posedge CSB )

	if (DEBUG_OPT7)
	    $display( "DBG: %m (@ %.3fms) serial_in SCK = %1b, CSB = %1b, opcode = %02h, mode = %1d, (i == 8 - mode) = %s, i = %1d, data = %02h (%08b), SI = %1b, valid = %1b",
		TIMESCALE*$time, SCK, CSB, opcode_reg, mode, (i == 8 - mode) ? "yes" : "no", i, data, data, SI, valid );

	// Serial input only while CSB is asserted
	if (CSB === 1'b0)
	begin
	    // Read inputs at SCK rising edge
	    case (mode)
		1 :	begin
			data[i]   = SI;
		    end
		2 :	begin
			data[i+1] = SO;
			data[i]   = SI;
		    end
		4 :	if (QE_reg === 1'b1)
		    begin
			data[i+3] = RESETB;
			data[i+2] = WPB;
			data[i+1] = SO;
			data[i]   = SI;
		    end
		    else $display( "ERROR: AT45DQXXX{#2115}, %m (@ %.3fms), Set QE config bit for quad-input.", TIMESCALE*$time );
	    endcase
	    i = i - mode;
	end // if CSB === 1'b0
    end // while

    if (i == -mode)
    begin
	debug_reg = data;
	-> BYTEBOUND;
	valid = 1'b1;	// valid is true if all eight bits read
    end
    else valid = 1'b0;

    if (DEBUG_OPT7)
	$display( "DBG: %m (@ %.3fms) serial_in SCK = %1b, CSB = %1b, opcode = %02h, mode = %1d, (i == -mode) = %s, i = %1d, data = %02h (%08b), valid = %1b",
	    TIMESCALE*$time, SCK, CSB, opcode_reg, mode, (i == -mode) ? "yes" : "no", i, data, data, valid );
end

endtask // serial_in

//==============================================================================
//  Task: serial_out (mode, data, >valid)
//==============================================================================
//  Write byte of data onto the serial output stream
//------------------------------------------------------------------------------

task serial_out;

input	mode;	// 1: single wire output via SO; 2: dual-bit; 4: quad-bit
input	data;	// One byte of serial output data
output	valid;	// Successful output of one full byte if true

integer	mode;
reg [7:0] data;
reg	valid;

integer	i;	// Local counter for bit number

begin
    // mode: S_SNGL=1, S_DUAL=2, S_QUAD=4
    case (mode)
    	S_SNGL	: -> DEBUG_SO;
    	S_DUAL	: -> DEBUG_DO;
    	S_QUAD	: -> DEBUG_QO;
	default	: $display( "ERROR: AT45DQXXX{#2120}, %m (@ %.3fms), Bad mode (%1d) for task serial_out.", TIMESCALE*$time, mode );
    endcase

    valid = 1'b0;
    i = 8 - mode;
    while (i >= 0  &&  CSB === 1'b0)
    begin : serial_out_LOOP
    	@( negedge SCK or posedge CSB )

	if (DEBUG_OPT7)
	    $display( "DBG: %m (@ %.3fms) serial_out SCK = %1b, CSB = %1b, opcode = %02h, mode = %1d, (i == 8 - mode) = %s, i = %1d, data = %02h (%08b), data[%1d] = %1b, valid = %1b",
		TIMESCALE*$time, SCK, CSB, opcode_reg, mode, (i == 8 - mode) ? "yes" : "no", i, data, data, i, data[i], valid );

	// Serial output only while CSB is asserted
	if (CSB === 1'b0)
	begin
	    #tHO
	    // Previous output cease to be valid tHO (output hold time) after falling SCK
	    // (Not modeled, to match behavior of structural netlist)
	    /*---
	    case (mode)
		1 :	begin
			SO_reg = 1'bx;
		    end
		2 :	begin
			SO_reg = 1'bx;
			SI_reg = 1'bx;
		    end
		4 :	begin
			RESETB_reg	= 1'bx;
			WPB_reg	= 1'bx;
			SO_reg	= 1'bx;
			SI_reg	= 1'bx;
		    end
	    endcase
	    ---*/

	    // Turn on output drivers at tV (output valid time) after falling SCK
	    #( tV - tHO )
	    if (i == 8 - mode  &&  CSB === 1'b0)
		case (mode)
		    1 :	begin
			    SO_on = 1'b1;
			end
		    2 :	begin
			    SO_on = 1'b1;
			    SI_on = 1'b1;
			end
		    4 :	begin
			    SO_on = 1'b1;
			    SI_on = 1'b1;
			    if (QE_reg === 1'b1)
			    begin
				RESETB_on = 1'b1;
				WPB_on    = 1'b1;
			    end
			    else $display( "ERROR: AT45DQXXX{#2125}, %m (@ %.3fms), Set QE config bit for quad-output.", TIMESCALE*$time );
			end
		endcase

	    // Output are valid tV (output valid time) after SCK falling edge
	    case (mode)
		1 :	begin
			SO_reg = data[i];
		    end
		2 :	begin
			SO_reg = data[i+1];
			SI_reg = data[i];
		    end
		4 :	begin
			RESETB_reg	= data[i+3];
			WPB_reg	= data[i+2];
			SO_reg	= data[i+1];
			SI_reg	= data[i];
		    end
	    endcase
	    i = i - mode;
	end // if CSB === 1'b0
    end // while

    // Depending on the opcode and mode0/mode3, the output will be clocked out
    // on the SCK negative edge just following the SCK positive edge for the
    // last bit clocked in, or the output will be clocked out once cycle later.
    // We can account for both cases here:
    // valid = (i == 8 - mode) || (i == 8 - 2*mode) ? 1'b1 : 1'b0;
    // Or we can use some global variables to see if a SCK cycle was skipped:
    case (spi_mode)
	0 : valid = ((i == -mode) || (i == 8 - 2*mode)) ? 1'b1 : 1'b0;
	3 : valid = ((i == -mode) || (i == 8 -   mode)) ? 1'b1 : 1'b0;
    endcase
    // Checking whether the serial output ends on a byte boundry is really just
    // a load of rubbish.  The customer is free to halt the serial output at any
    // time and the model has no business commenting on it.

    if (valid)
    begin
	debug_reg = data;
	-> BYTEBOUND;
    end

    if (DEBUG_OPT7)
	$display( "DBG: %m (@ %.3fms) serial_out SCK = %1b, CSB = %1b, opcode = %02h, skip = %1d, spi_mode = %1d, mode = %1d, (i == -mode) = %s, i = %1d, data = %02h (%08b), valid = %1b",
	    TIMESCALE*$time, SCK, CSB, opcode_reg, skip, spi_mode, mode, (i == -mode) ? "yes" : "no", i, data, data, valid );
end

endtask // serial_out

//==============================================================================
//  Task: highz_outputs
//==============================================================================
//  Set all serial outputs to high-Z
//------------------------------------------------------------------------------

task highz_outputs;

begin
    #tDIS
    SI_on = 1'b0;
    SO_on = 1'b0;
    WPB_on    = 1'b0;
    RESETB_on = 1'b0;
    if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) Outputs tri-stated", TIMESCALE*$time );
end

endtask // highz_outputs

//==============================================================================
//  Task: read_out (mode, mem_sel, addr, low_addr, high_addr)
//==============================================================================
//  Read main memory or buffer contents out on the serial interface
//------------------------------------------------------------------------------

task read_out;

input	mode;		// S_SNGL for output on SO, S_DUAL for dual-bit, S_QUAD for quad-bit
input	mem_sel;	// Memory selector, MEM_MM for mail memory, MEM_B1/MEM_B2 for buffers
input	addr;		// Start address
input	low_addr;	// Low address for roll-over
input	high_addr;	// High address for roll-over

integer	mode;
integer	mem_sel;
integer	addr;
integer	low_addr;
integer	high_addr;

reg [7:0] txdata;	// Local register to hold a temporary copy of the memory/buffer entry
reg	valid;		// Local flag to accept return status from serial_out task
integer bytecount;	// Local counter for bytes read out
integer	page;		// Page number for low_addr (current page from input address)
integer byte_addr;	// In-page byte address
integer	addr_bits;	// Number of bits for in-page byte address
reg [23:0] addr_mask;	// Address mask for in-page byte address

begin
    if (!(mem_sel == MEM_MM  ||  mem_sel == MEM_B1  || mem_sel == MEM_B2))
	$display( "ERROR: AT45DQXXX{#2130}, %m (@ %.3fms), Bad mem_sel (%1d) for task read_out.", TIMESCALE*$time, mem_sel );

    page = low_addr / DF_PAGESIZE;
    bytecount = 0;
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) read_out %4d addr = %06h, low_addr = %06h, high_addr = %06h", TIMESCALE*$time, bytecount, addr, low_addr, high_addr );

    if (addr > high_addr)
    begin
	addr_bits = DF_INPAGE_ADDR_BITS - binary_page;
	addr_mask = (24'h000001 << addr_bits) - 1;
	byte_addr = addr - low_addr;
	case (mem_sel)
	    MEM_B1 : $display( "ERROR: AT45DQXXX{#2135}, %m (@ %.3fms), Reading buffer1[%03h], outside the byte address range [%03h to %03h].", TIMESCALE*$time, addr, low_addr, high_addr );
	    MEM_B2 : $display( "ERROR: AT45DQXXX{#2140}, %m (@ %.3fms), Reading buffer2[%03h], outside the byte address range [%03h to %03h].", TIMESCALE*$time, addr, low_addr, high_addr );
	    MEM_MM : $display( "ERROR: AT45DQXXX{#2145}, %m (@ %.3fms), Reading memory[%06h], outside the byte address range [%06h to %06h].",
			TIMESCALE*$time, (page << addr_bits) + byte_addr, (page << addr_bits), (page << addr_bits) + page_size - 1 );
	endcase
	if (DEBUG_OPT5)
	    $display( "DBG: %m (@ %.3fms) read_out %4d addr = %06h, low_addr = %06h, byte_addr = %06h, addr_mask = %06h, (addr & addr_mask) = %06h -- %s",
		TIMESCALE*$time, bytecount, addr, low_addr, byte_addr, addr_mask, (byte_addr & addr_mask), (((byte_addr & addr_mask) != 24'h0) ? "continue" : "end") );
	// Output bytes of unknowns until inpage byte address rolls over, to match structural netlist behavior
	while ((byte_addr & addr_mask) != 24'h0)
	begin
	    txdata = 8'hXX;
	    serial_out( mode, txdata, valid );
	    bytecount = bytecount + 1;
	    if (DEBUG_OPT1)
		case (mem_sel)
		    MEM_B1 : $display( "DBG: %m (@ %.3fms) read_out %4d buffer1[%03h] = %02h (out-of-range)", TIMESCALE*$time, bytecount, addr, txdata );
		    MEM_B2 : $display( "DBG: %m (@ %.3fms) read_out %4d buffer2[%03h] = %02h (out-of-range)", TIMESCALE*$time, bytecount, addr, txdata );
		    MEM_MM : $display( "DBG: %m (@ %.3fms) read_out %4d memory[%06h] = %02h (out-of-range)", TIMESCALE*$time,  bytecount, addr, txdata );
		endcase
	    addr = addr + 1;
	    byte_addr = addr - low_addr;
	    if (DEBUG_OPT5)
		$display( "DBG: %m (@ %.3fms) read_out %4d addr = %06h, low_addr = %06h, byte_addr = %06h, addr_mask = %06h, (addr & addr_mask) = %06h -- %s",
		    TIMESCALE*$time, bytecount, addr, low_addr, byte_addr, addr_mask, (byte_addr & addr_mask), (((byte_addr & addr_mask) != 24'h0) ? "continue" : "end") );
	end
	addr = low_addr;
    end

    // Continue transmitting bytes while CSB is asserted (active low)
    while (CSB === 1'b0)
    begin : read_out_LOOP
	if (addr > high_addr)
	begin
	    addr = low_addr;	// Page wrap-around
	    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) read_out %4d addr = %06h, low_addr = %06h, high_addr = %06h --page wrap-around", TIMESCALE*$time, bytecount, addr, low_addr, high_addr );
	end

	case (mem_sel)
	    MEM_B1 : txdata = buffer1[addr];
	    MEM_B2 : txdata = buffer2[addr];
	    MEM_MM : txdata =  memory[addr];
	endcase

	serial_out( mode, txdata, valid );
	if (WARN_SOUT  &&  valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2150}, %m (@ %.3fms), Incomplete byte in serial output.", TIMESCALE*$time );

	if (CSB !== 1'b0) disable read_out_LOOP;

	bytecount = bytecount + 1;
	if (DEBUG_OPT1)
	    case (mem_sel)
		MEM_B1 : $display( "DBG: %m (@ %.3fms) read_out %4d buffer1[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
		MEM_B2 : $display( "DBG: %m (@ %.3fms) read_out %4d buffer2[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
		MEM_MM : $display( "DBG: %m (@ %.3fms) read_out %4d memory[%06h] = %02h", TIMESCALE*$time,  bytecount, addr, txdata );
	    endcase

	addr = addr + 1;	// Next byte
    end // while
end

endtask // read_out

//==============================================================================
//  Task: read_out_x (mode, mem_sel, addr, low_addr, high_addr)
//==============================================================================
//  Read main memory or buffer out on the serial interface as X's
//------------------------------------------------------------------------------

task read_out_x;

input	mode;		// S_SNGL for output on SO, S_DUAL for dual-bit, S_QUAD for quad-bit
input	mem_sel;	// Memory selector, MEM_MM for mail memory, MEM_B1/MEM_B2 for buffers
input	addr;		// Start address
input	low_addr;	// Low address for roll-over
input	high_addr;	// High address for roll-over

integer	mode;
integer	mem_sel;
integer	addr;
integer	low_addr;
integer	high_addr;

reg [7:0] txdata;	// Local register to hold a temporary copy of the memory/buffer entry
reg	valid;		// Local flag to accept return status from serial_out task
integer bytecount;	// Local counter for bytes read out

begin
    if (!(mem_sel == MEM_MM  ||  mem_sel == MEM_B1  || mem_sel == MEM_B2))
	$display( "ERROR: AT45DQXXX{#2155}, %m (@ %.3fms), Bad mem_sel (%1d) for task read_out_x.", TIMESCALE*$time, mem_sel );

    bytecount = 0;
    // Continue transmitting bytes while CSB is asserted (active low)
    while (CSB === 1'b0)
    begin : read_out_x_LOOP
	if (addr > high_addr)
	    addr = low_addr;	// Page wrap-around

	txdata = 8'hXX;

	serial_out( mode, txdata, valid );
	if (WARN_SOUT  &&  valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2160}, %m (@ %.3fms), Incomplete byte in serial output.", TIMESCALE*$time );

	if (CSB !== 1'b0) disable read_out_x_LOOP;

	bytecount = bytecount + 1;
	if (DEBUG_OPT1)
	    case (mem_sel)
		MEM_B1 : $display( "DBG: %m (@ %.3fms) read_out_x %4d buffer1[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
		MEM_B2 : $display( "DBG: %m (@ %.3fms) read_out_x %4d buffer2[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
		MEM_MM : $display( "DBG: %m (@ %.3fms) read_out_x %4d memory[%06h] = %02h", TIMESCALE*$time,  bytecount, addr, txdata );
	    endcase

	addr = addr + 1;	// Next byte
    end
end

endtask // read_out_x

//==============================================================================
//  Task: cont_read_out (mode, *addr, *low_addr, *high_addr)
//==============================================================================
//  Read out main memory contents, either in Continuous Mode or in Burst Mode
//  Will modify addr, low_addr and high_addr arguments to reflect last address read out
//------------------------------------------------------------------------------

task cont_read_out;

input	mode;		// S_SNGL for output on SO, S_DUAL for dual-bit, S_QUAD for quad-bit
inout	addr;		// Start address
inout	low_addr;	// Low address for roll-over
inout	high_addr;	// High address for roll-over

integer	mode;
integer	addr;
integer	low_addr;
integer	high_addr;

reg [7:0] txdata;	// Local register to hold a temporary copy of the memory/buffer entry
reg	valid;		// Local flag to accept return status from serial_out task
integer bytecount;	// Local counter for bytes read out
integer	page;		// Page number for low_addr (current page from input address)
integer byte_addr;	// In-page byte address
integer	addr_bits;	// Number of bits for in-page byte address
reg [23:0] addr_mask;	// Address mask for in-page byte address

begin
    bytecount = 0;
    page = low_addr / DF_PAGESIZE;
    byte_addr = addr - low_addr;
    addr_bits = DF_INPAGE_ADDR_BITS - binary_page;
    addr_mask = (24'h000001 << addr_bits) - 1;

    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
	$display( "ERROR: AT45DQXXX{#2165}, %m (@ %.3fms), Reading out page %04h from program suspended sector.", TIMESCALE*$time, page );
    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
	$display( "ERROR: AT45DQXXX{#2170}, %m (@ %.3fms), Reading out page %04h from erase suspended sector.", TIMESCALE*$time, page );

    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) cont_read_out %4d addr = %06h, page = %04h, low_addr = %06h, high_addr = %06h", TIMESCALE*$time, bytecount, addr, page, low_addr, high_addr );

    if (addr > memory_top)
	$display( "ERROR: AT45DQXXX{#2175}, %m (@ %.3fms), Reading memory[%06h], outside the memory address range [0 to %06h].",
	    TIMESCALE*$time, (page << addr_bits) + byte_addr, ((PAGES - 1) << addr_bits) + page_size - 1 );

    if( addr > high_addr )
    begin
	$display( "ERROR: AT45DQXXX{#2172}, %m (@ %.3fms), Reading memory[%06h], outside the DataFlash byte address range [%06h to %06h].",
	    TIMESCALE*$time, (page << addr_bits) + byte_addr, (page << addr_bits), (page << addr_bits) + page_size - 1 );
	if (DEBUG_OPT5)
	    $display( "DBG: %m (@ %.3fms) cont_read_out %4d addr = %06h, low_addr = %06h, byte_addr = %06h, addr_mask = %06h, (addr & addr_mask) = %06h -- %s",
		TIMESCALE*$time, bytecount, addr, low_addr, byte_addr, addr_mask, (byte_addr & addr_mask), (((byte_addr & addr_mask) != 24'h0) ? "continue" : "end") );
	// Output bytes of unknowns until inpage byte address rolls over, to match structural netlist behavior
	while ((byte_addr & addr_mask) != 24'h0)
	begin
	    txdata = 8'hXX;
	    serial_out( mode, txdata, valid );
	    bytecount = bytecount + 1;
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) cont_read_out %4d memory[%06h] = %02h (out-of-range)", TIMESCALE*$time, bytecount, addr, txdata );
	    addr = addr + 1;
	    byte_addr = addr - low_addr;
	    if (DEBUG_OPT5)
		$display( "DBG: %m (@ %.3fms) cont_read_out %4d addr = %06h, low_addr = %06h, byte_addr = %06h, addr_mask = %06h, (addr & addr_mask) = %06h -- %s",
		    TIMESCALE*$time, bytecount, addr, low_addr, byte_addr, addr_mask, (byte_addr & addr_mask), (((byte_addr & addr_mask) != 24'h0) ? "continue" : "end") );
	end
	addr = low_addr;
    end

    // Continue transmitting bytes while CSB is asserted (active low)
    while (CSB === 1'b0)
    begin : cont_read_out_LOOP
	// Note that rollover occurs at end of memory, not at the end of the page
	if (addr > memory_top)
	begin
	    low_addr = 0;
	    high_addr = page_size - 1;
	    page = 0;
	    addr = 0;

	    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) cont_read_out %4d addr = %06h, page = %04h, low_addr = %06h, high_addr = %06h", TIMESCALE*$time, bytecount, addr, page, low_addr, high_addr );

	    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
		$display( "ERROR: AT45DQXXX{#2185}, %m (@ %.3fms), Reading out page %04h from program suspended sector.", TIMESCALE*$time, page );
	    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
		$display( "ERROR: AT45DQXXX{#2190}, %m (@ %.3fms), Reading out page %04h from erase suspended sector.", TIMESCALE*$time, page );
	end
	else if (addr > high_addr)
	begin
	    // Next page
	    low_addr = low_addr + DF_PAGESIZE;
	    high_addr = high_addr + DF_PAGESIZE;
	    page = low_addr / DF_PAGESIZE;
	    addr = low_addr;

	    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) cont_read_out %4d addr = %06h, page = %04h, low_addr = %06h, high_addr = %06h", TIMESCALE*$time, bytecount, addr, page, low_addr, high_addr );

	    if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
		$display( "ERROR: AT45DQXXX{#2195}, %m (@ %.3fms), Reading out page %04h from program suspended sector.", TIMESCALE*$time, page );
	    else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
		$display( "ERROR: AT45DQXXX{#2200}, %m (@ %.3fms), Reading out page %04h from erase suspended sector.", TIMESCALE*$time, page );
	end

	if ((status2_reg[ST_PS2] == 1'b1  ||  status2_reg[ST_PS1] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_pgm_page)))
	    txdata = 8'hXX;	// Reading from a sector with a suspended program operation
	else if ((status2_reg[ST_ES] == 1'b1) && (inprog_sector(page) == inprog_sector(inprog_ers_page)))
	    txdata = 8'hXX;	// Reading from a sector with a suspended erase operation
	else
	    txdata = memory[addr];

	serial_out( mode, txdata, valid );
	if (WARN_SOUT  &&  valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2180}, %m (@ %.3fms), Incomplete byte in serial output.", TIMESCALE*$time );

	if (CSB !== 1'b0) disable cont_read_out_LOOP;

	bytecount = bytecount + 1;
	if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) cont_read_out %4d memory[%06h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );

	addr = addr + 1;	// Next byte
    end // while
end

endtask // cont_read_out

//==============================================================================
//  Task: cont_read_out_x (mode, *addr, *low_addr, *high_addr)
//==============================================================================
//  Read out main memory contents as X's, either in Continuous Mode or in Burst Mode
//  Will modify addr, low_addr and high_addr arguments to reflect last address read out
//------------------------------------------------------------------------------

task cont_read_out_x;

input	mode;		// S_SNGL for output on SO, S_DUAL for dual-bit, S_QUAD for quad-bit
inout	addr;		// Start address
inout	low_addr;	// Low address for roll-over
inout	high_addr;	// High address for roll-over

integer	mode;
integer	addr;
integer	low_addr;
integer	high_addr;

reg [7:0] txdata;	// Local register to hold a temporary copy of the memory/buffer entry
reg	valid;		// Local flag to accept return status from serial_out task
integer bytecount;	// Local counter for bytes read out

begin
    bytecount = 0;

    // Continue transmitting bytes while CSB is asserted (active low)
    while (CSB === 1'b0)
    begin : cont_read_out_x_LOOP
	// Note that rollover occurs at end of memory, not at the end of the page
	if (addr > memory_top)
	begin
	    addr = 0;
	    low_addr = 0;
	    high_addr = page_size - 1;
	end
	else if (addr > high_addr)
	begin
	    // Next page
	    low_addr = low_addr + DF_PAGESIZE;
	    high_addr = high_addr + DF_PAGESIZE;
	    addr = low_addr;
	end

	txdata = 8'hXX;

	serial_out( mode, txdata, valid );
	if (WARN_SOUT  &&  valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2205}, %m (@ %.3fms), Incomplete byte in serial output.", TIMESCALE*$time );

	if (CSB !== 1'b0) disable cont_read_out_x_LOOP;

	bytecount = bytecount + 1;
	if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) cont_read_out_x %4d memory[%06h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );

	addr = addr + 1;	// Next byte
    end // while
end

endtask // cont_read_out_x

//==============================================================================
//  Task: read_out_reg (reg_sel, addr, high_addr)
//==============================================================================
//  Read out the contents of the Protection, Lockdown or Security Register
//------------------------------------------------------------------------------

task read_out_reg;

input	reg_sel;	// R_PRT: Protection; R_LCK: Lockdown; R_SEC: Security Register
input	addr;		// Start address
input	high_addr;	// Highest valid address

integer	reg_sel;
integer	addr;
integer	high_addr;

reg [7:0] txdata;	// Local register to hold a temporary copy of the register entry
reg	valid;		// Local flag to accept return status from serial_out task
integer bytecount;	// Local counter for bytes read out

begin
    if (!(reg_sel == R_PRT  ||  reg_sel == R_LCK  || reg_sel == R_SEC))
	$display( "ERROR: AT45DQXXX{#2210}, %m (@ %.3fms), Bad reg_sel (%1d) for task read_out_reg.", TIMESCALE*$time, reg_sel );

    bytecount = 0;
    // Continue transmitting bytes while CSB is asserted (active low)
    while (CSB === 1'b0)
    begin : read_out_reg_LOOP
	if (addr > high_addr)
	begin
	    txdata = 8'hXX;
	end
	else
	    case (reg_sel)
		R_PRT : txdata = prot_reg[addr];
		R_LCK : txdata = lock_reg[addr];
		R_SEC : if (addr < 64) txdata = security_reg[addr];
			else           txdata = factory_reg[addr-64];
	    endcase

	serial_out( S_SNGL, txdata, valid );
	if (WARN_SOUT  &&  valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2235}, %m (@ %.3fms), Incomplete byte in serial output.", TIMESCALE*$time );

	if (CSB !== 1'b0) disable read_out_reg_LOOP;

	if (addr > high_addr)
	begin
	    case (reg_sel)
		R_PRT : $display( "ERROR: AT45DQXXX{#2215}, %m (@ %.3fms), Protection_Register[%03h], reading outside the address range [0 to %03h].", TIMESCALE*$time, addr, high_addr );
		R_LCK : $display( "ERROR: AT45DQXXX{#2220}, %m (@ %.3fms), Lockdown_Register[%03h], reading outside the address range [0 to %03h].", TIMESCALE*$time, addr, high_addr );
		R_SEC : if (addr < 64)
			    $display( "ERROR: AT45DQXXX{#2225}, %m (@ %.3fms), Security_Register[%03h], reading outside the byte address range [0 to %03h].", TIMESCALE*$time, addr, high_addr );
			else
			    $display( "ERROR: AT45DQXXX{#2230}, %m (@ %.3fms), Factory_Register[%03h], reading outside the byte address range [0 to %03h].", TIMESCALE*$time, addr, high_addr );
	    endcase
	end

	bytecount = bytecount + 1;
	if (DEBUG_OPT1)
	    case (reg_sel)
		R_PRT : $display( "DBG: %m (@ %.3fms) read_out_reg %4d prot_reg[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
		R_LCK : $display( "DBG: %m (@ %.3fms) read_out_reg %4d lock_reg[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
		R_SEC : if (addr < 64)
			    $display( "DBG: %m (@ %.3fms) read_out_reg %4d security_reg[%03h] = %02h", TIMESCALE*$time, bytecount, addr, txdata );
			else
			    $display( "DBG: %m (@ %.3fms) read_out_reg %4d factory_reg[%03h] = %02h", TIMESCALE*$time, bytecount, addr - 64, txdata );
	    endcase

	addr = addr + 1;	// Next byte
    end // while
end

endtask // read_out_reg

//==============================================================================
//  Task: write_buffer (mode, buf_sel, *addr, low_addr, high_addr, >bytecount)
//==============================================================================
//  Write data from the serial input to one of the page buffers.  Will output
//  bytecount argument.  Will update addr if out-of-range addr was supplied.
//------------------------------------------------------------------------------

task write_buffer;

input	mode;		// S_SNGL for output on SO, S_DUAL for dual-bit, S_QUAD for quad-bit
input	buf_sel;	// MEM_B1 for Buffer 1, MEM_B2 for Buffer 2
inout	addr;		// Start address
input	low_addr;	// Low address for roll-over
input	high_addr;	// High address for roll-over
output	bytecount;	// Counter for bytes read in

integer mode;
integer buf_sel;
integer addr;
integer low_addr;
integer high_addr;
integer bytecount;

reg [7:0] rxdata;	// Local register to hold a temporary copy of the incoming byte
reg	valid;		// Local flag to accept return status from task serial_in
integer byte_addr;	// In-page byte address
integer	addr_bits;	// Number of bits for in-page byte address
reg [23:0] addr_mask;	// Address mask for in-page byte address

begin
    if (!(buf_sel == MEM_B1  || buf_sel == MEM_B2))
	$display( "ERROR: AT45DQXXX{#2240}, %m (@ %.3fms), Bad buf_sel (%1d) for task write_buffer.", TIMESCALE*$time, buf_sel );

    bytecount = 0;
    byte_addr = addr - low_addr;

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) write_buffer (-0-) bytecount = %1d, addr = %06h, low_addr = %06h, high_addr = %06h, byte_addr = %06h", TIMESCALE*$time, bytecount, addr, low_addr, high_addr, byte_addr );

    if (addr > high_addr)
    begin
	$display( "ERROR: AT45DQXXX{#2250}, %m (@ %.3fms), Writing to buffer%1d[%03h], outside the byte address range [%03h to %03h].", TIMESCALE*$time, buf_sel, addr, low_addr, high_addr );
	// Read and discard bytes until inpage byte address rolls over, to match structural netlist behavior
	addr_bits = DF_INPAGE_ADDR_BITS - binary_page;
	addr_mask = (24'h000001 << addr_bits) - 1;
	if (DEBUG_OPT5)
	    $display( "DBG: %m (@ %.3fms) write_buffer %4d addr = %06h, low_addr = %06h, byte_addr = %06h, addr_mask = %06h, (addr & addr_mask) = %06h -- %s",
		TIMESCALE*$time, bytecount, addr, low_addr, byte_addr, addr_mask, (byte_addr & addr_mask), (((byte_addr & addr_mask) != 24'h0) ? "continue" : "end") );
	while ((byte_addr & addr_mask) != 24'h0)
	begin
	    serial_in( mode, rxdata, valid );
	    if (valid != 1'b1) disable write_buffer;
	    bytecount = bytecount + 1;
	    if (DEBUG_OPT1)
		$display( "DBG: %m (@ %.3fms) write_buffer %4d buffer%1d[%03h] := %02h (out-of-range)", TIMESCALE*$time, bytecount, buf_sel, addr, rxdata );
	    addr = addr + 1;
	    byte_addr = addr - low_addr;
	    if (DEBUG_OPT5)
		$display( "DBG: %m (@ %.3fms) write_buffer %4d addr = %06h, low_addr = %06h, byte_addr = %06h, addr_mask = %06h, (addr & addr_mask) = %06h -- %s",
		    TIMESCALE*$time, bytecount, addr, low_addr, byte_addr, addr_mask, (byte_addr & addr_mask), (((byte_addr & addr_mask) != 24'h0) ? "continue" : "end") );
	end
	addr = low_addr;
	byte_addr = 0;
    end

    // Continue receiving bytes while CSB is asserted (active low)
    while (CSB === 1'b0)
    begin
	serial_in( mode, rxdata, valid );
	// The write_buffer task will continue to call the serial_in task while CSB is low.
	// CSB may still be low after the last byte is clocked in, so serial_in will be
	// called again, but will return with valid==0 and rxdata=XX once CSB goes high.
	if (valid == 1'b0  &&  rxdata !== 8'hXX)
	    $display( "ERROR: AT45DQXXX{#2245}, %m (@ %.3fms), Incomplete byte in serial input: %02h.", TIMESCALE*$time, rxdata );

	if (valid == 1'b1)
	begin
	    case (buf_sel)
	      MEM_B1 :  buffer1[addr] = rxdata;
	      MEM_B2 :  buffer2[addr] = rxdata;
	    endcase

	    bytecount = bytecount + 1;
	    if (DEBUG_OPT1)
		$display( "DBG: %m (@ %.3fms) write_buffer %4d buffer%1d[%03h] := %02h", TIMESCALE*$time, bytecount, buf_sel, addr, rxdata );

	    addr = addr + 1;		// Next byte
	    if (addr > high_addr)
		addr = low_addr;	// Page roll-over
	end
    end // while

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) write_buffer (-1-) bytecount = %1d, addr = %06h, low_addr = %06h, high_addr = %06h, byte_addr = %06h", TIMESCALE*$time, bytecount, addr, low_addr, high_addr, byte_addr );

    addr = addr - bytecount;
    if (addr < low_addr)
	addr = addr + page_size;

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) write_buffer (end) bytecount = %1d, addr = %06h, low_addr = %06h, high_addr = %06h, byte_addr = %06h", TIMESCALE*$time, bytecount, addr, low_addr, high_addr, byte_addr );
end

endtask // write_buffer

//==============================================================================
//  Task: write_to_memory (buf_sel, page, low_addr)
//==============================================================================
//  Transfer data from a buffer into a page of main memory
//------------------------------------------------------------------------------

task write_to_memory;

input	buf_sel;	// MEM_B1 for Buffer 1, MEM_B2 for Buffer 2
input	page;		// Page number
input	page_start;	// Byte index of start of page

integer buf_sel;
integer page;
integer page_start;

integer i;

begin
    if (!(buf_sel == MEM_B1  || buf_sel == MEM_B2))
	$display( "ERROR: AT45DQXXX{#2255}, %m (@ %.3fms), Bad buf_sel (%1d) for task write_to_memory.", TIMESCALE*$time, buf_sel );

    // Do bit-AND instead of direct assignment, on account of flash program operation
    // when bits can only transition from a one to a zero, not the other way round.
    if (buf_sel == MEM_B1)
	for (i=0; i < page_size; i = i+1)
	begin
	    memory[page_start+i] = (memory[page_start+i] & buffer1[i]);	// bit-AND
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) write_to_memory %4d memory[%06h] := buffer1[%03h] = %02h", TIMESCALE*$time, i, page_start+i, i, buffer1[i] );
	end
    else if (buf_sel == MEM_B2)
	for (i=0; i < page_size; i = i+1)
	begin
	    memory[page_start+i] = (memory[page_start+i] & buffer2[i]);	// bit-AND
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) write_to_memory %4d memory[%06h] := buffer2[%03h] = %02h", TIMESCALE*$time, i, page_start+i, i, buffer2[i] );
	end

    page_status[page] = 1'b1;	// Not erased
end

endtask // write_to_memory

//==============================================================================
//  Task: write_to_memory_partial (buf_sel, page, low_addr, num_bytes, valid)
//==============================================================================
//  Transfer a number of bytes from a buffer into a page of main memory
//------------------------------------------------------------------------------

task write_to_memory_partial;

input	buf_sel;	// MEM_B1 for Buffer 1, MEM_B2 for Buffer 2
input	page;		// Page number
input	low_addr;	// Initial byte index from start of page to start writing
input	num_bytes;	// Number of bytes to write
output	valid;		// 1: Write Ok; 0: Write failed, some byte locations not erased

integer buf_sel;
integer page;
integer low_addr;
integer num_bytes;
reg	valid;

integer i;

begin
    if (!(buf_sel == MEM_B1  || buf_sel == MEM_B2))
	$display( "ERROR: AT45DQXXX{#2260}, %m (@ %.3fms), Bad buf_sel (%1d) for task write_to_memory_partial.", TIMESCALE*$time, buf_sel );

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) write_to_memory_partial (-0-) buffer%1d, page = %04h, low_addr = %06h, num_bytes = %1d, valid = %1b", TIMESCALE*$time, buf_sel, page, low_addr, num_bytes, valid );

    // Do bit-AND instead of direct assignment, on account of flash program operation
    // when bits can only transition from a one to a zero, not the other way round.
    valid = 1'b1;
    if (buf_sel == MEM_B1)
	for (i=0; i < num_bytes; i = i+1)
	begin
	    if (low_addr+i >= page_size) low_addr = low_addr - page_size;	// Page wrap-around
	    if (memory[(page*DF_PAGESIZE)+low_addr+i] !== 8'hFF) valid = 1'b0;
	    memory[(page*DF_PAGESIZE)+low_addr+i] = (memory[(page*DF_PAGESIZE)+low_addr+i] & buffer1[low_addr+i]);	// bit-AND
	    if (DEBUG_OPT1)
	    	$display( "DBG: %m (@ %.3fms) write_to_memory_partial %4d memory[%06h] := buffer1[%03h] = %02h", TIMESCALE*$time, i, (page*DF_PAGESIZE)+low_addr+i, low_addr+i, buffer1[low_addr+i] );
	end
    else if (buf_sel == MEM_B2)
	for (i=0; i < num_bytes; i = i+1)
	begin
	    if (low_addr+i >= page_size) low_addr = low_addr - page_size;	// Page wrap-around
	    if (memory[(page*DF_PAGESIZE)+low_addr+i] !== 8'hFF) valid = 1'b0;
	    memory[(page*DF_PAGESIZE)+low_addr+i] = (memory[(page*DF_PAGESIZE)+low_addr+i] & buffer2[low_addr+i]);	// bit-AND
	    if (DEBUG_OPT1)
	    	$display( "DBG: %m (@ %.3fms) write_to_memory_partial %4d memory[%06h] := buffer2[%03h] = %02h", TIMESCALE*$time, i, (page*DF_PAGESIZE)+low_addr+i, low_addr+i, buffer2[low_addr+i] );
	end

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) write_to_memory_partial (end) buffer%1d, page = %04h, low_addr = %06h, num_bytes = %1d, low_addr + num_bytes = %06h, valid = %1b",
	    TIMESCALE*$time, buf_sel, page, low_addr, num_bytes, low_addr + num_bytes, valid );

    page_status[page] = 1'b1;	// Not erased
end

endtask // write_to_memory_partial

//==============================================================================
//  Task: get_address (>addr, >valid)
//==============================================================================
//  Get a three byte address from the serial input stream
//------------------------------------------------------------------------------

task get_address;

output	addr;		// Full range address from start of memory (24-bit)
output	valid;		// True if three address bytes received

reg [23:0] addr;
reg	valid;

reg [7:0] rxdata;
integer	bytecount;

begin
    addr = 24'bx;
    bytecount = 0;
    valid = 1'b1;
    // Continue receiving bytes while CSB is asserted (active low)
    while (CSB === 1'b0  &&  bytecount < 3  &&  valid == 1'b1)
    begin
	serial_in( S_SNGL, rxdata, valid );
	if (valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2265}, %m (@ %.3fms), Incomplete byte in serial input: %02h.", TIMESCALE*$time, rxdata );

	addr = (addr << 8) | rxdata;
	bytecount = bytecount + 1;
	if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) get_address - byte %1d rxdata %02h addr %06h (%024b)", TIMESCALE*$time, bytecount, rxdata, addr, addr );
    end // while
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) get_address - CSB %b, bytecount %1d, valid %b, addr %06h (%024b)", TIMESCALE*$time, CSB, bytecount, valid, addr, addr );
    if (bytecount == 3  &&  valid)
    begin
	if (DEBUG_OPT2) $display( "DBG: %m (@ %.3fms) get_address, addr                     = %06h (%024b) %8d", TIMESCALE*$time, addr, addr, addr );
    end
    else
    begin
	$display( "ERROR: AT45DQXXX{#2270}, %m (@ %.3fms), Incomplete address.", TIMESCALE*$time );
	valid = 1'b0;
    end
end

endtask // get_address

//==============================================================================
//  Task: parse_address (*addr, >page, >page_start, >page_end, >inpage_addr)
//==============================================================================
//  Return page number, page address boundaries and local in-page address for
//  a 24-bit memory address.  Also updates input addr for storage arrangement
//  of memory array.
//------------------------------------------------------------------------------

task parse_address;

inout	addr;		// Address from start of memory (24-bit)
output	page;		// Page number
output	page_start;	// Byte index of start of page
output	page_end;	// Byte index of end of page
output	inpage_addr;	// Local address within the page

reg [23:0] addr;
integer	page;
integer page_start;
integer page_end;
integer inpage_addr;

integer	inpage_addr_bits;
reg [23:0] mask;
integer	i;
integer sector;

begin
    if (DEBUG_OPT5) $display( "DBG: %m (@ %.3fms) binary_page = %1b, page_size = %5d, memory_top = %06h", TIMESCALE*$time, binary_page, page_size, memory_top );
    // Do some math stuff to extract the different components of the address
    inpage_addr_bits = DF_INPAGE_ADDR_BITS - binary_page;
    // mask = (2 ** inpage_addr_bits) - 1;		// Verilog 2001
    mask = (24'h000001 << inpage_addr_bits) - 1;	// Verilog 1999
    inpage_addr = addr & mask;
    if (inpage_addr >= page_size)
	$display( "ERROR: AT45DQXXX{#2275}, %m (@ %.3fms), Byte address %03h is outside %s range [0 to %03h].", TIMESCALE*$time, inpage_addr, (binary_page ? "Binary" : "DataFlash"), page_size-1 );
    page = (addr & (~mask)) >> inpage_addr_bits;
    if (page >= PAGES)
    begin
	$display( "ERROR: AT45DQXXX{#2280}, %m (@ %.3fms), Page address %04h more than device %s number of pages (%1d), wrapping around to %04h.", TIMESCALE*$time, page, DEVICE, PAGES, page%PAGES );
	page = page % PAGES;
    end
    // Note here for a more compact representation of the flash memory array we're
    // storing DataFlash size pages one after the other.  We'll use the same representation
    // for Binary size pages, since the device can switch between DataFlash and Binary
    // size while in operation.
    page_start = page * DF_PAGESIZE;
    page_end = page_start + page_size - 1;	// page_size is global, updated by DPS and BPS events
    sector = page / PAGES_PER_SECTOR;
    if (DEBUG_OPT2)
    begin
	$display( "DBG: %m (@ %.3fms) parse_address, inpage_addr_bits   = %3d", TIMESCALE*$time, inpage_addr_bits );
	$display( "DBG: %m (@ %.3fms) parse_address, mask               = %06h (%024b) %8d", TIMESCALE*$time, mask, mask, mask );
	$display( "DBG: %m (@ %.3fms) parse_address, addr               = %06h (%024b) %8d", TIMESCALE*$time, addr, addr, addr );
	$display( "DBG: %m (@ %.3fms) parse_address, sector             = %06h (%024b) %8d", TIMESCALE*$time, sector, sector, sector );
	$display( "DBG: %m (@ %.3fms) parse_address, page               = %06h (%024b) %8d", TIMESCALE*$time, page, page, page );
	$display( "DBG: %m (@ %.3fms) parse_address, page_start         = %06h (%024b) %8d", TIMESCALE*$time, page_start, page_start, page_start );
	$display( "DBG: %m (@ %.3fms) parse_address, page_end           = %06h (%024b) %8d", TIMESCALE*$time, page_end, page_end, page_end );
	$display( "DBG: %m (@ %.3fms) parse_address, inpage_addr        = %06h (%024b) %8d", TIMESCALE*$time, inpage_addr, inpage_addr, inpage_addr );
    end
    // On account of compacting the pages into a single memory array we need to modify
    // the incoming address to point to the right entry in the memory array.
    addr = page_start + inpage_addr;
    if (DEBUG_OPT2)
	$display( "DBG: %m (@ %.3fms) parse_address, new addr           = %06h (%024b) %8d", TIMESCALE*$time, addr, addr, addr );
end

endtask // parse_address

//==============================================================================
//  Task: get_dummy_bytes (dummy_bytes, >valid)
//==============================================================================
//  Get a number of dummy bytes from the serial input stream
//------------------------------------------------------------------------------

task get_dummy_bytes;

input	dummy_bytes;	// Number of dummy bytes to read and ignore
output	valid;		// True if expected number of dummy bytes received

integer	dummy_bytes;
reg	valid;

reg [7:0] rxdata;
integer	bytecount;

begin
    bytecount = 0;
    valid = 1'b1;
    // Continue receiving bytes while CSB is asserted (active low)
    while (CSB === 1'b0  &&  bytecount < dummy_bytes  &&  valid == 1'b1)
    begin
	serial_in( S_SNGL, rxdata, valid );
	if (valid != 1'b1)
	    $display( "ERROR: AT45DQXXX{#2285}, %m (@ %.3fms), Incomplete byte in serial input: %02h.", TIMESCALE*$time, rxdata );

	bytecount = bytecount + 1;
	if (DEBUG_OPT2) $display( "DBG: %m (@ %.3fms) get_dummy_bytes - byte %1d rxdata %02h", TIMESCALE*$time, bytecount, rxdata );
    end // while
    if (bytecount != dummy_bytes  ||  !valid)
    begin
	$display( "ERROR: AT45DQXXX{#2290}, %m (@ %.3fms), Incomplete dummy bytes.", TIMESCALE*$time );
	valid = 1'b0;
    end
end

endtask // get_dummy_bytes

//==============================================================================
//  Task: transfer_to_buffer_partial (buf_sel, page, low_addr, num_bytes)
//==============================================================================
//  Transfer a number of bytes from a page of main memory into a buffer
//------------------------------------------------------------------------------

task transfer_to_buffer_partial;

input	buf_sel;	// MEM_B1 for Buffer 1, MEM_B2 for Buffer 2
input	page;		// Page number
input	low_addr;	// Initial byte index from start of page to start transferring
input	num_bytes;	// Number of bytes to transfer

integer buf_sel;
integer page;
integer low_addr;
integer num_bytes;

integer i;

begin
    if (!(buf_sel == MEM_B1  || buf_sel == MEM_B2))
	$display( "ERROR: AT45DQXXX{#2292}, %m (@ %.3fms), Bad buf_sel (%1d) for task transfer_to_buffer_partial.", TIMESCALE*$time, buf_sel );

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) transfer_to_buffer_partial (-0-) buffer%1d, page = %04h, low_addr = %06h, num_bytes = %1d", TIMESCALE*$time, buf_sel, page, low_addr, num_bytes );

    if (buf_sel == MEM_B1)
	for (i=0; i < num_bytes; i = i+1)
	begin
	    if (low_addr+i >= page_size) low_addr = low_addr - page_size;	// Page wrap-around
	    buffer1[low_addr+i] = memory[(page*DF_PAGESIZE)+low_addr+i];
	    if (DEBUG_OPT1)
	    	$display( "DBG: %m (@ %.3fms) transfer_to_buffer_partial %4d buffer1[%03h] := memory[%06h] = %02h", TIMESCALE*$time, i, low_addr+i, (page*DF_PAGESIZE)+low_addr+i, buffer1[low_addr+i] );
	end
    else if (buf_sel == MEM_B2)
	for (i=0; i < num_bytes; i = i+1)
	begin
	    if (low_addr+i >= page_size) low_addr = low_addr - page_size;	// Page wrap-around
	    buffer2[low_addr+i] = memory[(page*DF_PAGESIZE)+low_addr+i];
	    if (DEBUG_OPT1)
	    	$display( "DBG: %m (@ %.3fms) transfer_to_buffer_partial %4d buffer2[%03h] := memory[%06h] = %02h", TIMESCALE*$time, i, low_addr+i, (page*DF_PAGESIZE)+low_addr+i, buffer2[low_addr+i] );
	end

    if (DEBUG_OPT5)
	$display( "DBG: %m (@ %.3fms) transfer_to_buffer_partial (end) buffer%1d, page = %04h, low_addr = %06h, num_bytes = %1d, low_addr + num_bytes = %06h",
	    TIMESCALE*$time, buf_sel, page, low_addr, num_bytes, low_addr + num_bytes );
end

endtask // transfer_to_buffer_partial

//==============================================================================
//  Task: transfer_to_buffer (buf_sel, page_start)
//==============================================================================
//  Transfer data from a page of main memory into a buffer
//------------------------------------------------------------------------------

task transfer_to_buffer;

input	buf_sel;	// MEM_B1 for Buffer 1, MEM_B2 for Buffer 2
input	page_start;	// Byte index of start of page

integer buf_sel;
integer page_start;

integer i;

begin
    if (buf_sel == MEM_B1)
	for (i=0; i < page_size; i = i+1)
	begin
	    buffer1[i] = memory[page_start+i];
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) Transfer buffer%1d[%03h] := memory[%06h] = %02h", TIMESCALE*$time, buf_sel, i, page_start+i, memory[page_start+i] );
	end
    else if (buf_sel == MEM_B2)
	for (i=0; i < page_size; i = i+1)
	begin
	    buffer2[i] = memory[page_start+i];
	    if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) Transfer buffer%1d[%03h] := memory[%06h] = %02h", TIMESCALE*$time, buf_sel, i, page_start+i, memory[page_start+i] );
	end
    else
	$display( "ERROR: AT45DQXXX{#2295}, %m (@ %.3fms), Bad buf_sel (%1d) for task transfer_to_buffer.", TIMESCALE*$time, buf_sel );
end

endtask // transfer_to_buffer

//==============================================================================
//  Task: compare_with_buffer (buf_sel, page_start)
//==============================================================================
//  Compare data into a buffer against a page of main memory
//------------------------------------------------------------------------------

task compare_with_buffer;

input	buf_sel;	// MEM_B1 for Buffer 1, MEM_B2 for Buffer 2
input	page_start;	// Byte index of start of page

integer buf_sel;
integer page_start;

integer i;

begin
    status1_reg[ST_CMP] = 1'b0;

    if (buf_sel == MEM_B1)
	for (i=0; i < page_size; i = i+1)
	begin : COMPARE_LOOP1
	    if (memory[page_start+i] !== buffer1[i])
	    begin
		if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) Compare buffer1[%03h] (%02h) != memory[%06h] (%02h)", TIMESCALE*$time, i, buffer1[i], page_start+i, memory[page_start+i] );
		status1_reg[ST_CMP] = 1'b1;	// Mismatch found
		disable COMPARE_LOOP1;
	    end
	    else
		if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) Compare buffer1[%03h] (%02h) == memory[%06h] (%02h)", TIMESCALE*$time, i, buffer1[i], page_start+i, memory[page_start+i] );
	end // COMPARE_LOOP1
    else if (buf_sel == MEM_B2)
	for (i=0; i < page_size; i = i+1)
	begin : COMPARE_LOOP2
	    if (memory[page_start+i] !== buffer2[i])
	    begin
		if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) Compare buffer2[%03h] (%02h) != memory[%06h] (%02h)", TIMESCALE*$time, i, buffer2[i], page_start+i, memory[page_start+i] );
		status1_reg[ST_CMP] = 1'b1;	// Mismatch found
		disable COMPARE_LOOP2;
	    end
	    else
		if (DEBUG_OPT1) $display( "DBG: %m (@ %.3fms) Compare buffer2[%03h] (%02h) == memory[%06h] (%02h)", TIMESCALE*$time, i, buffer2[i], page_start+i, memory[page_start+i] );
	end // COMPARE_LOOP2
    else
	$display( "ERROR: AT45DQXXX{#2300}, %m (@ %.3fms), Bad buf_sel (%1d) for task compare_with_buffer.", TIMESCALE*$time, buf_sel );
    if (DEBUG_OPT0  &&  status1_reg[ST_CMP] === 1'b0) $display( "DBG: %m (@ %.3fms) Compare buffer %1d to memory page %04h match", TIMESCALE*$time, buf_sel, page_start / DF_PAGESIZE );
    if (DEBUG_OPT0  &&  status1_reg[ST_CMP] === 1'b1) $display( "DBG: %m (@ %.3fms) Compare buffer %1d to memory page %04h mismatch", TIMESCALE*$time, buf_sel, page_start / DF_PAGESIZE );
end

endtask // compare_with_buffer

//==============================================================================
//  Task: erase_page (page)
//==============================================================================
//  Erase a page of main memory
//------------------------------------------------------------------------------

task erase_page;

input	page;		// Page number

integer page;

integer page_start;
integer i;

begin
    page_start = page * DF_PAGESIZE;

    // Erase entire DataFlash size page
    for (i=0; i < DF_PAGESIZE; i = i+1)
	memory[page_start+i] = 8'hFF;

    page_status[page] = 1'b0;	// Erased
    if (DEBUG_OPT3) $display( "DBG: %m (@ %.3fms) Erasing memory page %04h", TIMESCALE*$time, page );
end

endtask // erase_page

//==============================================================================
//  Task: corrupt_buffer (buf_sel)
//==============================================================================
//  Corrupt one entire SRAM buffer
//------------------------------------------------------------------------------

task corrupt_buffer;

input buf_sel;

integer buf_sel;

integer i;

begin
    if (buf_sel == MEM_B1)
	for (i=0; i < DF_PAGESIZE; i = i+1)
	    buffer1[i] = 8'hXX;
    else if (buf_sel == MEM_B2)
	for (i=0; i < DF_PAGESIZE; i = i+1)
	    buffer2[i] = 8'hXX;
    else
	$display( "ERROR: AT45DQXXX{#2305}, %m (@ %.3fms), Bad buf_sel (%1d) for task corrupt_buffer.", TIMESCALE*$time, buf_sel );
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting buffer %1d", TIMESCALE*$time, buf_sel );
end

endtask // corrupt_buffer

//==============================================================================
//  Task: corrupt_page (page)
//==============================================================================
//  Corrupt a page of main memory
//------------------------------------------------------------------------------

task corrupt_page;

input	page;		// Page number

integer page;

integer page_start;
integer i;

begin
    page_start = page * DF_PAGESIZE;

    // Corrupt entire DataFlash size page
    for (i=0; i < DF_PAGESIZE; i = i+1)
	memory[page_start+i] = 8'hXX;

    page_status[page] = 1'b1;	// Not erased
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting page %04h", TIMESCALE*$time, page );
end

endtask // corrupt_page

//==============================================================================
//  Task: corrupt_block (page)
//==============================================================================
//  Corrupt a block (8 pages) of memory
//------------------------------------------------------------------------------

task corrupt_block;

input	page;

integer	i;

begin
    page = (page / 8) * 8;		// Block starts at eight page inteval
    for (i=page; i < (page+8); i = i+1)
	corrupt_page( i );

    for (i=page; i < (page+8); i = i+1)
	page_status[i+1] = 1'b1;	// Set page to un-erased state
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting block %04h", TIMESCALE*$time, page / 8 );
end

endtask // corrupt_block

//==============================================================================
//  Task: corrupt_sector (page)
//==============================================================================
//  Corrupt a sector (8 pages/248 pages/256 pages) of memory
//------------------------------------------------------------------------------

task corrupt_sector;

input	page;

integer i;
integer	sector;

begin
    sector = page / PAGES_PER_SECTOR;
    if (page < 8)
    begin
	for (i=0; i < 8; i = i+1)
	    corrupt_page( i );
	for (i=0; i < 8; i = i+1)
	    page_status[i+1] = 1'b1;	// Set page to un-erased state
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting sector 0a", TIMESCALE*$time );
    end
    else if (sector == 0)
    begin
	for (i=8; i < PAGES_PER_SECTOR; i = i+1)
	    corrupt_page( i );
	for (i=8; i < PAGES_PER_SECTOR; i = i+1)
	    page_status[i+1] = 1'b1;	// Set page to un-erased state
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting sector 0b", TIMESCALE*$time );
    end
    else
    begin
	page = sector * PAGES_PER_SECTOR;
	for (i=page; i < (page+PAGES_PER_SECTOR); i = i+1)
	    corrupt_page( i );
	for (i=page; i < (page+PAGES_PER_SECTOR); i = i+1)
	    page_status[i+1] = 1'b1;	// Set page to un-erased state
	if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting sector %04h", TIMESCALE*$time, sector );
    end
end

endtask // corrupt_sector

//==============================================================================
//  Task: corrupt_chip
//==============================================================================
//  Corrupt the entire main memory
//------------------------------------------------------------------------------

task corrupt_chip;

integer i;

begin
    for (i=0; i < PAGES; i = i+1)
	corrupt_page( i );
    for (i=0; i < PAGES; i = i+1)
	page_status[i+1] = 1'b1;	// Set page to un-erased state
    if (DEBUG_OPT0) $display( "DBG: %m (@ %.3fms) Corrupting chip", TIMESCALE*$time );
end

endtask // corrupt_chip

//==============================================================================
//  Task: reset_chip
//==============================================================================
//  Reset the device, from issuing a Software Reset or asserting RESETB
//------------------------------------------------------------------------------

task reset_chip;

integer j;

begin
    // -------------------------------------------------------------------------
    // Kill all running event handlers
    // -------------------------------------------------------------------------
    // disable INITIALIZE_;	// Exception, needed to setup initial state
    disable GET_OPCODE_;
    disable SUSPEND_;
    disable RESUME_;
    disable EDPD_;
    disable RDPD_;
    disable UDPD_;
    // disable XUDPD_;		// Exception, triggered by CSB pulse
    disable MMPR_;
    disable MMCAR_;
    disable DORA_;
    disable QORA_;
    disable B1R_;
    disable B2R_;
    disable MMPTB1T_;
    disable MMPTB2T_;
    disable MMPTB1C_;
    disable MMPTB2C_;
    disable B1W_;
    disable B2W_;
    disable DB1W_;
    disable DB2W_;
    disable QB1W_;
    disable QB2W_;
    disable B1TMMPPWBIE_;
    disable B2TMMPPWBIE_;
    disable B1TMMPPWOBIE_;
    disable B2TMMPPWOBIE_;
    disable PE_;
    disable BE_;
    disable SE_;
    disable OPC7_;
    disable CE_;
    disable MMPPB1_;
    disable MMPPB2_;
    disable MMPPB1NE_;
    disable APRB1_;
    disable APRB2_;
    disable SR_;
    disable CSR_;
    disable MIR_;
    disable OP9B_;
    disable SRP_;
    disable SRR_;
    disable OP3D_;
    disable SPE_;
    disable SPD_;
    disable PRE_;
    disable PRP_;
    disable LRP_;
    disable BPS_;
    disable DPS_;
    disable QEN_;
    disable QDIS_;
    disable PRR_;
    disable LRR_;
    disable OP34_;
    disable FREEZE_;
    disable OPF0_;

    // -------------------------------------------------------------------------
    // Kill all running tasks
    // -------------------------------------------------------------------------
    disable serial_in;
    disable serial_out;
    // disable highz_outputs;		// Exception, used in this task
    disable read_out;
    disable read_out_x;
    disable cont_read_out;
    disable cont_read_out_x;
    disable read_out_reg;
    disable write_buffer;
    disable write_to_memory;
    disable write_to_memory_partial;
    disable get_address;
    disable parse_address;
    disable get_dummy_bytes;
    disable transfer_to_buffer;
    disable compare_with_buffer;
    disable erase_page;
    // disable corrupt_buffer;		// Exception, used in this task
    // disable corrupt_page;		// Exception, used in this task
    // disable corrupt_block;		// Exception, used in this task
    // disable corrupt_sector;		// Exception, used in this task
    // disable corrupt_chip;		// Exception, used in this task
    // disable reset_chip;		// Exception, this task

    // -------------------------------------------------------------------------
    // If some operation was in progress, destroy the contents of potentially affected stores
    // -------------------------------------------------------------------------
    if (updating_buffer1 == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2310}, %m (@ %.3fms), Reset while updating SRAM Buffer 1. Corrupting Buffer 1.", TIMESCALE*$time );
	corrupt_buffer( MEM_B1 );
    end

    if (updating_buffer2 == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2315}, %m (@ %.3fms), Reset while updating SRAM Buffer 2. Corrupting Buffer 2.", TIMESCALE*$time );
	corrupt_buffer( MEM_B2 );
    end

    if (comparing == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2320}, %m (@ %.3fms), Reset while comparing. Corrupting status register bit 6.", TIMESCALE*$time );
	status1_reg[ST_CMP] = 1'bx;
    end

    if (updating_memory == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2325}, %m (@ %.3fms), Reset while updating memory. Corrupting Memory Page.", TIMESCALE*$time );
	corrupt_page( inprog_pgm_page );
    end

    if (erasing_page == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2330}, %m (@ %.3fms), Reset while erasing page. Corrupting Memory Page.", TIMESCALE*$time );
	corrupt_page( inprog_ers_page );
    end

    if (erasing_block == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2335}, %m (@ %.3fms), Reset while erasing block. Corrupting Memory Block.", TIMESCALE*$time );
	corrupt_block( inprog_ers_page );
    end

    if (erasing_sector == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2340}, %m (@ %.3fms), Reset while erasing sector. Corrupting Memory Sector.", TIMESCALE*$time );
	corrupt_block( inprog_ers_page );	// Erase operates on one block at a time
    end

    if (erasing_chip == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2345}, %m (@ %.3fms), Reset while erasing chip. Corrupting Memory.", TIMESCALE*$time );
	corrupt_block( inprog_ers_page );	// Erase operates on one block at a time
    end

    if (updating_pagesize == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2350}, %m (@ %.3fms), Reset while updating page size.", TIMESCALE*$time );
	status1_reg[ST_BIN] = 1'bx;
    end

    if (updating_cfg_reg == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2355}, %m (@ %.3fms), Reset while updating config register.", TIMESCALE*$time );
	config_reg = 8'hXX;
    end

    if (updating_sec_reg == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2360}, %m (@ %.3fms), Reset while updating security register.  Corrupting and locking the security register.", TIMESCALE*$time );
	for (j=0; j<64; j=j+1)
	    security_reg[j] = 8'hXX;
	security_flag = 1'b1;
    end

    // There is no operation to update the factory_reg, so we don't need to handle it here
    // (updating_fac_reg)

    if (updating_lock_reg == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2365}, %m (@ %.3fms), Reset while updating lockdown register.  Assume everything is locked.", TIMESCALE*$time );
	for (j=0; j<SECTORS; j=j+1)
	begin
	    lock_reg[j] = 8'hFF;
	    lock_status[j+1] = 1'b1;
	end
	lock_status[0] = 1'b1;
    end

    if (updating_prot_reg == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2370}, %m (@ %.3fms), Reset while updating protection register.  Assume everything is protected.", TIMESCALE*$time );
	for (j=0; j<SECTORS; j=j+1)
	begin
	    prot_reg[j] = 8'hFF;
	    prot_status[j+1] = 1'b1;
	end
	prot_status[0] = 1'b1;
    end

    if (updating_freeze == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2375}, %m (@ %.3fms), Reset while freezing lockdown state.  Assumed frozen.", TIMESCALE*$time );
	status2_reg[ST_SLE] = 1'b0;
    end

    if (status2_reg[ST_PS1] == 1'b1  ||  status2_reg[ST_PS2] == 1'b1)
    begin
	$display( "ERROR: AT45DQXXX{#2380}, %m (@ %.3fms), Reset while a program operation is suspended.  Aborting suspended operation and corrupting page %04h.", TIMESCALE*$time, inprog_pgm_page );
	corrupt_page( inprog_pgm_page );
    end

    if (status2_reg[ST_ES] == 1'b1)
    begin
	case (inprog_ers_scope)
	    ERS_P   :	begin
			    $display( "ERROR: AT45DQXXX{#2385}, %m (@ %.3fms), Reset while an erase operation is suspended.  Aborting suspended operation and corrupting page %04h.",
				TIMESCALE*$time, inprog_ers_page );
			    corrupt_page( inprog_ers_page );
	    		end
	    default :	begin
			    $display( "ERROR: AT45DQXXX{#2390}, %m (@ %.3fms), Reset while an erase operation is suspended.  Aborting suspended operation and corrupting block %04h.",
				TIMESCALE*$time, inprog_ers_page/8 );
			    corrupt_block( inprog_ers_page );
	    		end
	endcase
    end

    // -------------------------------------------------------------------------
    // Reset all status flags.  Nothing is in progress.
    // -------------------------------------------------------------------------
    deep_power_down = 1'b0;
    ultra_power_down = 1'b0;
    updating_buffer1 = 1'b0;
    updating_buffer2 = 1'b0;
    inuse_buffer1 = 1'b0;
    inuse_buffer2 = 1'b0;
    updating_memory = 1'b0;
    updating_pagesize = 1'b0;
    updating_cfg_reg = 1'b0;
    updating_sec_reg = 1'b0;
    updating_fac_reg = 1'b0;
    updating_lock_reg = 1'b0;
    updating_prot_reg = 1'b0;
    updating_freeze = 1'b0;
    comparing = 1'b0;
    erasing_page = 1'b0;
    erasing_block = 1'b0;
    erasing_sector = 1'b0;
    erasing_chip = 1'b0;

    opcode_reg = 8'hXX;
    bg_opcode = 8'hXX;

    // Reset the program/erase suspend status bits
    status2_reg[ST_PS1] = 1'b0;
    status2_reg[ST_PS2] = 1'b0;
    status2_reg[ST_ES] = 1'b0;

    status2_reg[ST_EPE] = 1'b0;		// Reset the Erase/Program Error bit
    // status1_reg[ST_CMP] = 1'b0;	// Reset Compare status bit

    allow_ers_suspend = 1'b0;
    inprog_pgm_buffer = 0;
    allow_ers_suspend = 1'b0;
    inprog_ers_scope = 0;
end

endtask // reset_chip

//==============================================================================
//  Function: check_lockdown (page)
//==============================================================================
//  Return the lockdown status of the sector containing page.
//  1: Sector is locked down, 0 otherwise
//------------------------------------------------------------------------------

function integer check_lockdown;

input	page;

integer	page;
integer	sector;

begin
    sector = page / PAGES_PER_SECTOR;

    if (page < 8)
	check_lockdown = lock_status[0];
    else if (sector == 0)
	check_lockdown = lock_status[1];
    else
	check_lockdown = lock_status[sector+1];
end

endfunction // check_lockdown

//==============================================================================
//  Function: check_protect (page)
//==============================================================================
//  Return the protection status of the sector containing page.
//  1: Sector is protected, 0 otherwise
//------------------------------------------------------------------------------

function integer check_protect;

input	page;

integer	page;
integer	sector;

begin
    sector = page / PAGES_PER_SECTOR;

    if (hw_protect_active == 1'b1)
	    check_protect = 1'b1;
    else if (status1_reg[ST_PRT] == 1'b1)
    begin
	// Software Sector Protection is Enabled
	if (page < 8)
	    check_protect = prot_status[0];
	else if (sector == 0)
	    check_protect = prot_status[1];
	else
	    check_protect = prot_status[sector+1];
    end
    else    check_protect = 1'b0;
end

endfunction // check_protect

//==============================================================================
//  Function: inprog_sector (page)
//==============================================================================
//  Return the sector number containing page for suspended operation checks
//------------------------------------------------------------------------------

function integer inprog_sector;

input	page;

integer	page;

begin
    if (page < 8)
	inprog_sector = 0;
    else if (page < PAGES_PER_SECTOR)
	inprog_sector = 1;
    else
	inprog_sector = (page / PAGES_PER_SECTOR)+1;
end

endfunction // inprog_sector

//==============================================================================

`endprotect
endmodule

